Require Import VST.floyd.proofauto.
Require Export ASI_surelyMalloc.
Require Import ASI_DFA.
Require Import ASI_SingleDFA.
Require Import ASI_StaticDFA.
Require Import ASI_ComplementDFA.
Require Import ASI_ProductDFA.
Require Import ASI_IntersectionDFA.
Require Import WeekdayDFA.
Require Import TrueFalseDFA.
Require Import ASI_EDFA.
Require Import ASI_T_Weekdays.
Require Import ASI_Twice_Weekdays.

   Notation CHrepW := (fun (n:ch)(v:val) =>
             (v = Vint (Int.repr (proj1_sig n)))).
   Notation CHrepTF := (fun (b:bool)(v:val) =>
             (b=true <-> (v = Vint (Int.repr 48)))).

Definition BoolBinopFunspec (final: bool -> bool -> bool):funspec :=
  WITH bl:bool, br:bool
  PRE [ tbool, tbool ]
    PROP ()
    PARAMS (Val.of_bool bl; Val.of_bool br)
    SEP ()
  POST [ tbool ]
    PROP ()
    RETURN (Val.of_bool (final bl br))
    SEP ().

  Definition runStaticDFAs_spec (StaticW_APD: DFA_APD _ WeekdayDFA) (StaticTF_APD: DFA_APD _ TrueFalseDFA) (SGL_APD: SingleDFA_APD _ TrueFalseDFA) (ClientGV:globals -> mpred) cs staticops i f s : funspec :=
   WITH binop:val, bop:bool -> bool -> bool, gv: globals
   PRE [ tptr (Tfunction (Tcons tbool (Tcons tbool Tnil)) tbool cc_default) ]
     PROP () PARAMS(binop) GLOBALS (gv)
     SEP ((*S*)DFA_GV _ _ StaticW_APD gv; (*S*)DFA_GV _ _ StaticTF_APD gv; SingleGV _ _ SGL_APD false gv;
          ClientGV gv; func_ptr' (BoolBinopFunspec bop) binop)
   POST [ tbool ]
     EX p: val, EX q:val,
        PROP () RETURN (Val.of_bool (bop false (bop true true)))
        SEP (SDFA _ CHrepW WeekdayDFA StaticW_APD staticops i f s cs ((Tuesday,gv), p);
             SDFA _ CHrepTF TrueFalseDFA StaticTF_APD staticops i f s cs ((true,gv), q);
             ClientGV gv; SingleGV _ _ SGL_APD true gv).

Section ClientSpecs.
  Variable M:MallocFreeAPD.
  Variable W_APD: DFA_APD _ WeekdayDFA.
  Variable TF_APD: DFA_APD _ TrueFalseDFA.

  Variable state:type.
  Variable dfaops:type.
  Variable ops i f s:ident.
  Variable cs:compspecs.

  Variable C_APD: DFA_APD _ (ComplementDFA WeekdayDFA).

  Definition runClientComplement_spec : funspec :=
   WITH gv: globals
   PRE [ ]
     PROP (WA+WORD < 100) PARAMS() GLOBALS (gv)
     SEP (mem_mgr M gv; DFA_GV _ _ W_APD gv; DFA_GV _ _ C_APD gv)
   POST [ tbool ]
     EX p: val, PROP () RETURN (Vtrue)
        SEP (mem_mgr M gv; DFA_GV _ _ W_APD gv; DFA_GV _ _ C_APD gv;
             DFA _ CHrepW (ComplementDFA WeekdayDFA)  C_APD  M state dfaops ops i f s cs (Tuesday, gv, p)).

  Notation Prod := (ProductDFA (*ch ch*) WeekdayDFA (ComplementDFA (*ch*) WeekdayDFA)).
  Variable Prod_APD: DFA_APD _ Prod.
  Notation Prod_GV:= (DFA_GV _ _ Prod_APD).

  Definition Inter := (IntersectionDFA (*ch ch*) WeekdayDFA Prod).
  Variable Inter_APD: DFA_APD _ Inter.
  Notation Inter_GV:= (DFA_GV _ _ Inter_APD).

(*  Variable InterInherit_APD: DFA_APD _ Inter.
  Notation InterInherit_GV:= (DFA_GV _ _ InterInherit_APD).*)

(*OLD
  Notation Prod := (ProductDFA (*ch ch*) WeekdayDFA (ComplementDFA (*ch*) WeekdayDFA)).
  Variable Prod_APD: DFA_APD _ Prod.

  Notation Inter := (IntersectionDFA (*ch ch*) WeekdayDFA (ComplementDFA (*ch*) WeekdayDFA)).
  Variable Inter_APD: DFA_APD _ Inter.
  Variable InterInh_APD: DFA_APD _ Inter.*)

  Notation ProdCHrep:= (fun (c: ch * ch) (v:val) => CHrepW (fst c) v /\ CHrepW (snd c) v).
  Notation InterCHrep:= (fun (c: ch * (ch * ch)) (v:val) => CHrepW (fst c) v /\ ProdCHrep (snd c) v).
(*
  Definition runClientEither_spec : funspec :=
   WITH b:bool, gv: globals
   PRE [ tbool ]
     PROP (WA+WORD < 100) PARAMS(Val.of_bool b) GLOBALS (gv)
     SEP (mem_mgr M gv; DFA_GV _ _ W_APD gv; DFA_GV _ _ C_APD gv;
          DFA_GV _ _ Prod_APD gv; DFA_GV _ _ Inter_APD gv; DFA_GV _ _ InterInherit_APD gv)
   POST [ tbool ]
     EX p: val, PROP () RETURN ((*Val.of_bool b*)Vfalse)
        SEP (mem_mgr M gv; DFA_GV _ _ W_APD gv; DFA_GV _ _ C_APD gv;
             DFA_GV _ _ Prod_APD gv; DFA_GV _ _ Inter_APD gv; DFA_GV _ _ InterInherit_APD gv;
             (*swallowd DFA _ ProdCHrep _ Prod_APD  M state dfaops ops i f s cs ((Monday,Monday), gv, p)*)
             if b then DFA _ InterCHrep _ InterInherit_APD  M state dfaops ops i f s cs (Monday,(Monday,Monday), gv, p)
                  else DFA _ InterCHrep _ Inter_APD  M state dfaops ops i f s cs (Monday,(Monday,Monday), gv, p)).*)

  Definition runClientBoth_spec (ClientGV:globals -> mpred): funspec :=
   WITH binop:val, bop:bool -> bool -> bool, gv: globals
   PRE [ tptr (Tfunction (Tcons tbool (Tcons tbool Tnil)) tbool cc_default) ]
     PROP (WA+WORD < 100) PARAMS(binop) GLOBALS (gv)
     SEP (mem_mgr M gv; DFA_GV _ _ W_APD gv; DFA_GV _ _ C_APD gv;
          DFA_GV _ _ Prod_APD gv; DFA_GV _ _ Inter_APD gv; (*DFA_GV _ _ InterInherit_APD gv;*) ClientGV gv;
          func_ptr' (BoolBinopFunspec bop) binop)
   POST [ tbool ]
     EX p: val, (*EX q:val,*)
        PROP () RETURN (Val.of_bool (bop true false))
        SEP (mem_mgr M gv; DFA_GV _ _ W_APD gv; DFA_GV _ _ C_APD gv;
             DFA_GV _ _ Prod_APD gv; DFA_GV _ _ Inter_APD gv; (*DFA_GV _ _ InterInherit_APD gv; *)ClientGV gv;
             DFA _ InterCHrep _ Inter_APD  M state dfaops ops i f s cs (Monday,(Monday,Monday), gv, p)(*;
             DFA _ InterCHrep _ InterInherit_APD  M state dfaops ops i f s cs (Monday,(Monday,Monday), gv, q)*)).

  Variable TWeekday_APD : EDFA_APD TWeekdayDFA.
  Definition runTClient_spec tdfaops t: funspec :=
   WITH gv: globals
   PRE [ ]
     PROP (WA+WORD < 100) PARAMS() GLOBALS (gv)
     SEP (mem_mgr M gv; EDFA_GV _ TWeekday_APD gv)
   POST [ tbool ]
     EX p: val, PROP () RETURN (Vfalse)
        SEP (mem_mgr M gv; EDFA_GV _ TWeekday_APD gv;
             EDFA _ CHrepW TWeekdayDFA TWeekday_APD M state tdfaops ops i f s t cs (Wednesday, gv, p)).
 
  Variable TwiceWeekday_APD : EDFA_APD TwiceWeekdayDFA.
  Definition runTwiceClient_spec tdfaops t: funspec :=
   WITH gv: globals
   PRE [ ]
     PROP (WA+WORD < 100) PARAMS() GLOBALS (gv)
     SEP (mem_mgr M gv; EDFA_GV _ TwiceWeekday_APD gv)
   POST [ tbool ]
     EX p: val, PROP () RETURN (Vfalse)
        SEP (mem_mgr M gv; EDFA_GV _ TwiceWeekday_APD gv;
             EDFA _ CHrepW TwiceWeekdayDFA TwiceWeekday_APD M state tdfaops ops i f s t cs (Monday, gv, p)). (*twice stepping X yields Monday*)

  Definition createWeekdayDFA_spec: funspec :=
   WITH b:bool, gv: globals
   PRE [ tbool ]
     PROP (WA+WORD < 100) PARAMS(Val.of_bool b) GLOBALS (gv)
     SEP (mem_mgr M gv; EDFA_GV _ TWeekday_APD gv; DFA_GV _ _ W_APD gv)
   POST [ tptr state ]
     EX p: val, PROP () RETURN (p)
        SEP (mem_mgr M gv; EDFA_GV _ TWeekday_APD gv; DFA_GV _ _ W_APD gv;
             EX APD, DFA _ CHrepW WeekdayDFA APD M state dfaops ops i f s cs (Monday, gv, p)).

  Definition runClient_spec (ClientGV:globals -> mpred): funspec :=
   WITH gv: globals
   PRE [ ]
     PROP (WA+WORD < 100) PARAMS() GLOBALS (gv)
     SEP (mem_mgr M gv; DFA_GV _ _ TF_APD gv; EDFA_GV _ TWeekday_APD gv; DFA_GV _ _ W_APD gv; ClientGV gv)
   POST [ tbool ]
     EX w tf APD, PROP () RETURN (Vfalse)
        SEP (mem_mgr M gv; DFA_GV _ _ TF_APD gv; EDFA_GV _ TWeekday_APD gv; DFA_GV _ _ W_APD gv; ClientGV gv;
             DFA _ CHrepW WeekdayDFA (*W_APD*)APD M state dfaops ops i f s cs (Tuesday, gv, w);
             DFA bool CHrep TrueFalseDFA TF_APD M state dfaops ops i f s cs (true, gv, tf)).

End ClientSpecs.