Require Import VST.floyd.proofauto.
Require Export ASI_surelyMalloc.
(*Should use this: Require Export ASI_malloc.*)
Require Import ObjectBase.
Require Export DFA.
Require Export DFAbase.

Section DFA_specs.
Variable state:type. (*ie. Tstruct _state noattr))*)
Variable ch:Type.
Variable CHrep: ch -> val -> Prop.

Variable A: @Dfa ch.

Definition init_spec (P: GPred (Dfa_state A)):funspec := 
    WITH s:Dfa_state A, gv:globals, v: val
    PRE [ tptr state ]
     PROP()
     PARAMS (v) GLOBALS (gv)
     SEP(P (s, gv, v))
    POST[ tvoid ]
      PROP()
      RETURN ()
      SEP(P (Dfa_s A, gv, v)).

Definition final_spec (P: GPred (Dfa_state A)):funspec := 
    WITH s:Dfa_state A, gv:globals, v: val
    PRE [ tptr state ]
     PROP()
     PARAMS (v) GLOBALS (gv)
     SEP(P (s, gv, v))
    POST[ tbool ]
      PROP()
      RETURN (if Dfa_fin A s then Vtrue else Vfalse)
      SEP(P (s, gv, v)).

Definition step_spec (P: GPred (Dfa_state A)):funspec := 
    WITH s:Dfa_state A, gv:globals, v: val, c:val, d:ch
    PRE [ tptr state, tuchar ]
     PROP(CHrep d c)
     PARAMS (v; c) GLOBALS (gv)
     SEP(P (s, gv, v))
    POST[ tvoid ]
      PROP()
      RETURN ()
      SEP(P (@Dfa_trans ch A s d, gv, v)).

Definition DFA_ops_specs:list (GPred (Dfa_state A) -> funspec):= [init_spec; final_spec; step_spec].

End DFA_specs.

Section MkDFAspec.

Variable char: Type.
Variable CHrep: char -> val -> Prop.

Variable A: @Dfa char.
Variable D:DFA_APD _ A.
Variable M:MallocFreeAPD.

Variable state:type.
Variable dfaops:type.
Variable ops i f s:ident.
Variable cs: compspecs.

Definition DFA: GPred (Dfa_state A) :=
  (*EX fptrs, *)dynobj_rep cs (Dfa_state A) (DFA_Inv char A D) M state ops dfaops [i; f; s] (DFA_ops_specs state char CHrep A) (*fptrs*).

Lemma DFA_fold_unfold:
  DFA = fun x => EX fptrs cso, (*EX P:Pred (dfa_state A), *)
                        DFA_Inv char A D (*P*) x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt (Dfa_state A) state ops m x *
                                     @funcptrs cs (Dfa_state A) dfaops [i; f; s] (DFA_ops_specs state char CHrep A) fptrs m (*P*)(DFA_Inv char A D)).
Proof. reflexivity. Qed.

Lemma DFA_validpointer tgv v: DFA (tgv,v) |-- valid_pointer v.
Proof. rewrite DFA_fold_unfold. Intros fptrs cso (*P*) m. entailer!. Qed.

Lemma DFA_isptr tgv v: DFA (tgv,v) |-- !!(isptr v).
Proof. rewrite DFA_fold_unfold. Intros fptrs cso (*P*) m. entailer!. apply H. Qed.

Definition makeDFA_spec : funspec :=
 WITH gv: globals
 PRE [ ]
    PROP (WA+WORD < 100) PARAMS() GLOBALS (gv) (*The condition WA+WORD  100 is true for 64 and 32 bit (and conservative...)*)
    SEP (mem_mgr M gv; DFA_GV _ _ D gv)
 POST [ tptr state ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; DFA_GV _ _ D gv;
          DFA (Dfa_s A, gv, p)). (*allocation yields an in-memory representation of A, in the initial state*) 

End MkDFAspec.

Lemma DFA_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) char CHrep A D M state dfaops ops i f s
      (Hdfaops: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) dfaops = true)
      (Hstate: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) state = true)
      (Hidents: forall i0 : ident, @In ident i0 [i; f; s] ->
                @reptype cs1 (@nested_field_type cs1 dfaops (DOT i0)) =
                @reptype cs2 (@nested_field_type cs2 dfaops (DOT i0)))
      (Hnestedfields: @reptype cs1 (@nested_field_type cs1 state (DOT ops)) =
                      @reptype cs2 (@nested_field_type cs2 state (DOT ops))):
      DFA char CHrep A D M state dfaops ops i f s cs1 = DFA char CHrep A D M state dfaops ops i f s cs2.
Proof. apply (obj_rep_change_compspecs _ _ CCE12); trivial. Qed.

Section BehSub.
  Variable Sigma Sigma':Type.
  Variable A: @Dfa Sigma.
  Variable A':@Dfa Sigma'.
  Variable MOR: DfaMorphism A A'.

  Variable CHrep: Sigma -> val -> Prop.
  Variable CHrep': Sigma' -> val -> Prop.

  (*APD construction is contravariant in MOR*)
  Variable BehSub_APD': DFA_APD Sigma' A'.
  Definition BehSub_APD: DFA_APD Sigma A :=
    {| DFA_GV := DFA_GV Sigma' A' BehSub_APD';
       DFA_Inv := fun '(a, gv, v) => DFA_Inv Sigma' A' BehSub_APD' (DfaMorph_state A A' MOR a, gv, v) |}.

  Variable M:MallocFreeAPD.

  Variable state:type.
  Variable dfaops:type.
  Variable ops i f s:ident.
  Variable cs: compspecs.

  Variable CHAxiom: forall d c, CHrep d c -> CHrep' (DfaMorph_Sigma A A' MOR d) c.

  Lemma init_sub: funspec_sub
    (init_spec state Sigma A (fun '(a, gv, v0) => DFA_Inv Sigma' A' BehSub_APD' (DfaMorph_state A A' MOR a, gv, v0)))
    (init_spec state Sigma' A' (DFA_Inv Sigma' A' BehSub_APD')).
  Proof. do_funspec_sub. destruct w as [[a' gv] v]. Abort.

  Lemma init_sub: funspec_sub
    (init_spec state Sigma' A' (DFA_Inv Sigma' A' BehSub_APD'))
    (init_spec state Sigma A (fun '(a, gv, v) => DFA_Inv Sigma' A' BehSub_APD' (DfaMorph_state A A' MOR a, gv, v))).
  Proof. do_funspec_sub. destruct w as [[a gv] v]. Intros.
    Exists (DfaMorph_state _ _ MOR a, gv, v) emp. entailer!.
    intros. rewrite DfaMorph_s. trivial.
  Qed.
  Lemma final_sub: funspec_sub
    (final_spec state Sigma' A' (DFA_Inv Sigma' A' BehSub_APD'))
    (final_spec state Sigma A (fun '(a, gv, v) => DFA_Inv Sigma' A' BehSub_APD' (DfaMorph_state A A' MOR a, gv, v))).
  Proof. do_funspec_sub. destruct w as [[a gv] v]. Intros.
    Exists (DfaMorph_state _ _ MOR a, gv, v) emp. entailer!.
    intros. rewrite <- DfaMorph_fin in H3, H4. entailer!.
  Qed.
  Lemma step_sub: funspec_sub
    (step_spec state Sigma' CHrep' A' (DFA_Inv Sigma' A' BehSub_APD'))
    (step_spec state Sigma CHrep A (fun '(a, gv, v) => DFA_Inv Sigma' A' BehSub_APD' (DfaMorph_state A A' MOR a, gv, v))).
  Proof. do_funspec_sub. destruct w as [[[[a gv] v] c] d]. Intros.
    Exists (DfaMorph_state _ _ MOR a, gv, v, c, DfaMorph_Sigma _ _ MOR d) emp. entailer!.
    intros. rewrite DfaMorph_trans. trivial.
  Qed.
  Lemma DFA_Morph a gv v:
      DFA Sigma' CHrep' A' BehSub_APD' M state dfaops ops i f s cs (DfaMorph_state _ _ MOR a, gv, v)
  |-- DFA Sigma CHrep A BehSub_APD M state dfaops ops i f s cs (a, gv, v).
  Proof. rewrite 2 DFA_fold_unfold. Intros fptrs cso m. Exists fptrs cso m.
    entailer!.
    apply sepcon_derives; [ apply derives_refl | simpl]. 
    clear - CHAxiom. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer! |].
    { apply func_ptr'_mono. apply init_sub. }
    clear - CHAxiom. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer! |].
    { apply func_ptr'_mono. apply final_sub. }
    clear - CHAxiom. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer! | trivial].
    apply func_ptr'_mono. apply step_sub.
  Qed.

  Lemma DFA_BehSub:
        funspec_sub (makeDFA_spec Sigma' CHrep' A' BehSub_APD' M state dfaops ops i f s cs)
                    (makeDFA_spec Sigma CHrep A BehSub_APD M state dfaops ops i f s cs).
  Proof.
    do_funspec_sub. Exists w emp; rename w into gv; entailer!.
    intros rho; intros. Exists (eval_id ret_temp rho). entailer!.
    specialize (DFA_Morph (Dfa_s A) gv (eval_id ret_temp rho)).
    rewrite (DfaMorph_s _ _ MOR). trivial.
  Qed.
End BehSub.
