Require Import VST.floyd.proofauto.
Require Export ASI_surelyMalloc.
(*Should use this: Require Export ASI_malloc.*)
Require Import ObjectBase.
Require Import ASI_DFA.

Record EDfa {Sigma} :=
  { Dfa_ :> @Dfa Sigma;
    Dfa_t : Dfa_state Dfa_ -> Sigma -> Sigma -> Dfa_state Dfa_ }.

Section EDFA_specs.
Variable state:type. (*ie. Tstruct _state noattr))*)
Variable Sigma:Type.
Variable CHrep: Sigma -> val -> Prop.

Variable A: @EDfa Sigma.

Definition twosteps_spec (P: GPred (Dfa_state A)):funspec := 
    WITH s:Dfa_state A, gv:globals, v: val, c1:val, c2:val, d1:Sigma, d2:Sigma
    PRE [ tptr state, tuchar, tuchar ]
     PROP(CHrep d1 c1; CHrep d2 c2)
     PARAMS (v; c1; c2) GLOBALS (gv)
     SEP(P (s, gv, v))
    POST[ tvoid ]
      PROP()
      RETURN ()
      SEP(P (Dfa_t A s d1 d2, gv, v)).

Definition EDFA_ops_specs :list (GPred (Dfa_state A) -> funspec):= 
           DFA_ops_specs state Sigma CHrep A ++ [twosteps_spec].

End EDFA_specs.

Record EDFA_APD {Sigma} (A: @EDfa Sigma):= {
  EDFA_GV: globals -> mpred;
  EDFA_Inv: GPred (Dfa_state A);
}.

Section MkEDFAspec.

Variable Sigma: Type.
Variable CHrep: Sigma -> val -> Prop.

Variable A: @EDfa Sigma.

Variable D:EDFA_APD A.

Variable M:MallocFreeAPD.

Variable state:type.
Variable dfaops:type.
Variable ops i f s t:ident.
Variable cs: compspecs.

Definition EDFA: GPred (Dfa_state A) :=
(*  EX fptrs,*) dynobj_rep cs (Dfa_state A) (EDFA_Inv A D) M state ops dfaops [i; f; s; t] 
            (EDFA_ops_specs state Sigma CHrep A).

Lemma EDFA_fold_unfold:
  EDFA = fun x => EX fptrs cso, (*EX P:GPred (Dfa_state A), *)
                        EDFA_Inv A D (*P*) x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt (Dfa_state A) state ops m x *
                                     @funcptrs cs (Dfa_state A) dfaops [i; f; s; t] 
                                     (EDFA_ops_specs state Sigma CHrep A) fptrs m (*P*)(EDFA_Inv A D)).
Proof. reflexivity. Qed.

Lemma EDFA_validpointer tgv v: EDFA (tgv,v) |-- valid_pointer v.
Proof. rewrite EDFA_fold_unfold. Intros fptrs cso (*P*) m. entailer!. Qed.

Lemma EDFA_isptr tgv v: EDFA (tgv,v) |-- !!(isptr v).
Proof. rewrite EDFA_fold_unfold. Intros fptrs cso (*P*) m. entailer!. apply H. Qed.

Definition makeEDFA_spec: funspec :=
 WITH gv: globals
 PRE [ ]
    PROP (WA+WORD < 100) PARAMS() GLOBALS (gv) (*The condition WA+WORD  100 is true for 64 and 32 bit (and conservative...)*)
    SEP (mem_mgr M gv; EDFA_GV A D gv)
 POST [ tptr state ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; EDFA_GV A D gv;
          EDFA (Dfa_s A, gv, p)).
End MkEDFAspec.

Lemma EDFA_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) Sigma CHrep A D M state dfaops ops i f s t
      (Hdfaops: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) dfaops = true)
      (Hstate: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) state = true)
      (Hidents: forall i0 : ident, @In ident i0 [i; f; s; t] ->
                @reptype cs1 (@nested_field_type cs1 dfaops (DOT i0)) =
                @reptype cs2 (@nested_field_type cs2 dfaops (DOT i0)))
      (Hnestedfields: @reptype cs1 (@nested_field_type cs1 state (DOT ops)) =
                      @reptype cs2 (@nested_field_type cs2 state (DOT ops))):
      EDFA Sigma CHrep A D M state dfaops ops i f s t cs1 = EDFA Sigma CHrep A D M state dfaops ops i f s t cs2.
Proof. apply (obj_rep_change_compspecs _ _ CCE12); trivial. Qed.

Definition EApd2DApd {Sigma A} (T:@EDFA_APD Sigma A): DFA_APD Sigma A :=
 {| DFA_GV := EDFA_GV A T;  DFA_Inv := fun x => EX F, EDFA_Inv A T x * F |}.

Lemma EDFA2DFA Sigma CHrep A TA M state dfaops ops i f s t cs x (*b*):
  EDFA Sigma CHrep A TA M state dfaops ops i f s t cs x
|-- DFA Sigma CHrep A (EApd2DApd TA) M state dfaops ops i f s cs x.
Proof. destruct x as [[a gv] v].
  rewrite DFA_fold_unfold, EDFA_fold_unfold. Intros fptrs cso m.
  Exists (firstn 3 fptrs) cso m.
  unfold funcptrs. destruct fptrs.
  { simpl. apply derives_trans with FF. rewrite <- sepcon_assoc.  apply sepcon_FF_derives'; trivial.
    apply FF_left. }
  simpl. Intros ish iv. destruct fptrs.
  { simpl. apply derives_trans with FF. rewrite <- !  sepcon_assoc. apply sepcon_FF_derives'; trivial.
    apply FF_left. }
  simpl. Intros fsh fv. destruct fptrs.
  { simpl. apply derives_trans with FF. rewrite <- !  sepcon_assoc. apply sepcon_FF_derives'; trivial.
    apply FF_left. }
  simpl. Intros ssh sv. destruct fptrs.
  { simpl. apply derives_trans with FF. rewrite <- !  sepcon_assoc. apply sepcon_FF_derives'; trivial.
    apply FF_left. }
  simpl. Intros tsh tv. destruct fptrs.
  2:{ simpl. clear. apply derives_trans with FF. rewrite <-! sepcon_assoc. apply sepcon_FF_derives'; trivial.
    apply FF_left. }
  Exists (field_at tsh dfaops (DOT t) tv m).
  Exists ish iv fsh fv ssh sv. 
  sep_apply (func_ptr'_emp (twosteps_spec state Sigma CHrep A (EDFA_Inv A TA)) v3). entailer!.
  repeat apply sepcon_derives; apply func_ptr'_mono; clear.
  + do_funspec_sub. Exists w; destruct w as [[s gv] v]. Intros F; Exists F. 
    entailer!. intros. Exists F. cancel.
  + do_funspec_sub. Exists w; destruct w as [[s gv] v]. Intros F; Exists F. 
    entailer!. intros. Exists F. entailer!.
  + do_funspec_sub. Exists w; destruct w as [[[[s gv] v] c] d]. Intros F; Exists F.
    entailer!. intros. Exists F. cancel.
Qed. 