Require Import VST.floyd.proofauto.
Require Export ASI_surelyMalloc.
Require Import ASI_DFA.

Section ProductDFA_ASI.
  Variable SigmaL SigmaR: Type.
  Variable ParentL: @Dfa SigmaL.
  Variable ParentR: @Dfa SigmaR.

  Notation ProductDFA := (ProductDFA ParentL ParentR).

  Variable CHrepL: SigmaL -> val -> Prop.
  Variable CHrepR: SigmaR -> val -> Prop.

  Variable DParentL: DFA_APD SigmaL ParentL.
  Variable DParentR: DFA_APD SigmaR ParentR.

  Variable M:MallocFreeAPD.

  Variable state:type.
  Variable dfaops:type.
  Variable ops i f s:ident.
  Variable cs: compspecs.

  Variable DProd: DFA_APD _ ProductDFA.

  Notation ProductGV:= (DFA_GV _ _ DProd). (*location predicate for globals*)
 
  (*note that C code (and this spec) allow invocation of product irrespective of
    whether the parent DFAs are in their initial states*)
  Definition makeProduct_spec : funspec :=
   WITH l:val, r:val, stL:_, stR:_, gv: globals
   PRE [ tptr state, tptr state ]
     PROP (WA+WORD < 100) PARAMS(l;r) GLOBALS (gv)
     SEP (mem_mgr M gv; ProductGV gv;
          DFA SigmaL CHrepL ParentL DParentL M state dfaops ops i f s cs (stL, gv, l);
          DFA SigmaR CHrepR ParentR DParentR M state dfaops ops i f s cs (stR, gv, r))
   POST [ tptr state ]
     EX p: val, PROP () RETURN (p)
        SEP (mem_mgr M gv; ProductGV gv;
             DFA (SigmaL * SigmaR)%type (fun c v => CHrepL (fst c) v /\ CHrepR (snd c) v) 
                 ProductDFA DProd M state dfaops ops i f s cs ((stL, stR), gv, p)).
End ProductDFA_ASI.

Section Pair_ASI.

  Variable Sigma: Type.
  Variable ParentL: @Dfa Sigma.
  Variable ParentR: @Dfa Sigma.

  Notation PairDFA := (PairDFA ParentL ParentR).

  Variable CHrep: Sigma -> val -> Prop.

  Variable DParentL: DFA_APD Sigma ParentL.
  Variable DParentR: DFA_APD Sigma ParentR.

  Variable M:MallocFreeAPD.

  Variable state:type.
  Variable dfaops:type.
  Variable ops i f s:ident.
  Variable cs: compspecs.

  Variable DPair: DFA_APD _ PairDFA.
  Notation PairGV:= (DFA_GV _ _ DPair). (*location predicate for globals*)
  
  (*note that C code (and this spec) allow invocation of product irrespective of
    whether the parent DFAs are in their initial states*)
  Definition makePair_spec : funspec :=
   WITH l:val, r:val, stL:_, stR:_, gv: globals
   PRE [ tptr state, tptr state ]
     PROP (WA+WORD < 100) PARAMS(l;r) GLOBALS (gv)
     SEP (mem_mgr M gv; PairGV gv;
          DFA Sigma CHrep ParentL DParentL M state dfaops ops i f s cs (stL, gv, l);
          DFA Sigma CHrep ParentR DParentR M state dfaops ops i f s cs (stR, gv, r))
   POST [ tptr state ]
     EX p: val, PROP () RETURN (p)
        SEP (mem_mgr M gv; PairGV gv;
             DFA Sigma CHrep
                 PairDFA DPair M state dfaops ops i f s cs ((stL, stR), gv, p)).

End Pair_ASI.

Section Prod2Pair.
  Variable Sigma: Type.
  Variable ParentL: @Dfa Sigma.
  Variable ParentR: @Dfa Sigma.

  Notation ProdDFA := (ProductDFA ParentL ParentR).
  Notation PairDFA := (PairDFA ParentL ParentR).

  Variable CHrep: Sigma -> val -> Prop.

  Variable DParentL: DFA_APD Sigma ParentL.
  Variable DParentR: DFA_APD Sigma ParentR.

  Variable M:MallocFreeAPD.

  Variable state:type.
  Variable dfaops:type.
  Variable ops i f s:ident.
  Variable cs: compspecs.

  Variable DProd: DFA_APD _ ProdDFA.

  Notation DPair(*: DFA_APD _ PairDFA*) :=
             (Build_DFA_APD Sigma PairDFA (DFA_GV _ _ DProd) (DFA_Inv _ _ DProd)).

  Lemma step_funspec_sub: funspec_sub
    (step_spec state (Sigma * Sigma)
       (fun c v =>  CHrep (fst c) v /\ CHrep (snd c) v)
       (ProductDFA ParentL ParentR)
       (DFA_Inv (Sigma * Sigma) (ProductDFA ParentL ParentR) DProd))
    (step_spec state Sigma CHrep PairDFA
       (DFA_Inv Sigma PairDFA DPair)).
  Proof.
    do_funspec_sub. destruct w as [[[[st gv] v] c] d]; Intros.
    Exists (st, gv, v, c, (d, d)) emp. entailer!.
  Qed.

  Lemma DFA_ProductPair x:
    DFA (Sigma * Sigma) (fun c v => CHrep (fst c) v /\ CHrep (snd c) v)
        (ProductDFA ParentL ParentR) DProd M state dfaops ops i f s cs x
    |-- DFA Sigma CHrep PairDFA DPair M state dfaops ops i f s cs x.
  Proof.
    rewrite 2 DFA_fold_unfold. Intros fptrs cso m. Exists fptrs cso m.
    destruct x as [[st gv] v]; entailer!.
    Transparent ObjectBase.funcptrs. simpl. Opaque ObjectBase.funcptrs.
(*    apply sepcon_derives; [ apply derives_refl | simpl]. *)
    clear. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer! | trivial ].
    apply func_ptr'_mono. apply step_funspec_sub.
  Qed.

  Lemma makeProduct_makePair:
        funspec_sub (@makeProduct_spec _ _ ParentL ParentR CHrep CHrep DParentL DParentR M state dfaops ops i f s cs DProd)
                    (@makePair_spec _ ParentL ParentR CHrep  DParentL DParentR M state dfaops ops i f s cs DPair).
  Proof. do_funspec_sub. Exists w emp.
    destruct w as [[[[l r] stL] stR] gv]. entailer!.
    intros; Exists (eval_id ret_temp x); entailer!.
    apply DFA_ProductPair.
  Qed.

(*Here's a proof that the Lemma DFA_ProductPair above can also be proven by
  instantating using DFA_Morph*)
  Program Definition PairProdMorphism:=
    Build_DfaMorphism _ _ PairDFA ProdDFA (fun d => (d,d)) id _ _ _.

  Lemma DFA_ProductPairUsingBehSub x:
    DFA (Sigma * Sigma) (fun c v => CHrep (fst c) v /\ CHrep (snd c) v)
        (ProductDFA ParentL ParentR) DProd M state dfaops ops i f s cs x
    |-- DFA Sigma CHrep PairDFA DPair M state dfaops ops i f s cs x.
  Proof. destruct x as [[a gv] v].
    eapply derives_trans. apply (DFA_Morph _ _ _ _ PairProdMorphism CHrep).
    { simpl; intros; split; trivial. }
    apply derives_refl'. f_equal. destruct DProd; unfold BehSub_APD; simpl.
    f_equal. extensionality x; destruct x as [[? ?] ?]; trivial.
  Qed.

(*The inverse constructor specification subsumption requires an extra axiom on CHrep*)
  Lemma step_funspec_sub_inv (H: forall d1 d2 c, CHrep d1 c -> CHrep d2 c -> d1=d2):
    funspec_sub (step_spec state Sigma CHrep PairDFA (DFA_Inv Sigma PairDFA DPair))
                (step_spec state (Sigma * Sigma)
                                 (fun c v =>  CHrep (fst c) v /\ CHrep (snd c) v)
                                 (ProductDFA ParentL ParentR)
                                 (DFA_Inv (Sigma * Sigma) (ProductDFA ParentL ParentR) DProd)).
  Proof. do_funspec_sub. destruct w as [[[[st gv] v] c] [d1 d2]]; Intros.
    apply (H  _ _ _ H1) in H2. simpl in *; subst d2.
    Exists (st, gv, v, c, d1) emp. entailer!.
  Qed.

  Lemma DFA_PairProduct (H: forall d1 d2 c, CHrep d1 c -> CHrep d2 c -> d1=d2) x:
    DFA Sigma CHrep PairDFA DPair M state dfaops ops i f s cs x |--
    DFA (Sigma * Sigma) (fun c v => CHrep (fst c) v /\ CHrep (snd c) v)
        (ProductDFA ParentL ParentR) DProd M state dfaops ops i f s cs x.
  Proof.
    rewrite 2 DFA_fold_unfold. Intros fptrs cso m. Exists fptrs cso m.
    destruct x as [[st gv] v]; entailer!.
    apply sepcon_derives; [ apply derives_refl | simpl]. 
    Transparent ObjectBase.funcptrs. simpl. Opaque ObjectBase.funcptrs.
    clear - H. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear - H. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear - H. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer! | trivial ].
    apply func_ptr'_mono. apply (step_funspec_sub_inv H).
  Qed.

  Lemma makePair_makeProduct (H: forall d1 d2 c, CHrep d1 c -> CHrep d2 c -> d1=d2):
        funspec_sub (@makePair_spec _ ParentL ParentR CHrep DParentL DParentR M state dfaops ops i f s cs DPair)
                    (@makeProduct_spec _ _ ParentL ParentR CHrep CHrep DParentL DParentR M state dfaops ops i f s cs DProd).
  Proof. do_funspec_sub. Exists w emp.
    destruct w as [[[[l r] stL] stR] gv].
    (*entailer! takes 10secs...*) Intros; apply andp_right; [apply prop_right; intuition; entailer! | cancel].
    intros; Exists (eval_id ret_temp rho'); entailer!.
    apply (DFA_PairProduct H).
  Qed.
End Prod2Pair.

Section Pair2Prod.
  Variable Sigma: Type.
  Variable ParentL: @Dfa Sigma.
  Variable ParentR: @Dfa Sigma.

  Notation ProdDFA := (ProductDFA ParentL ParentR).
  Notation PairDFA := (PairDFA ParentL ParentR).

  Variable CHrep: Sigma -> val -> Prop.

  Variable DParentL: DFA_APD Sigma ParentL.
  Variable DParentR: DFA_APD Sigma ParentR.

  Variable M:MallocFreeAPD.

  Variable state:type.
  Variable dfaops:type.
  Variable ops i f s:ident.
  Variable cs: compspecs.

  Variable DPair: DFA_APD _ PairDFA.

  Notation DProd(*: DFA_APD _ ProdDFA*) :=
             (Build_DFA_APD (Sigma * Sigma) ProdDFA (DFA_GV _ _ DPair) (DFA_Inv _ _ DPair)).

  Lemma step_funspec_sub': funspec_sub
    (step_spec state (Sigma * Sigma)
       (fun c v =>  CHrep (fst c) v /\ CHrep (snd c) v)
       (ProductDFA ParentL ParentR)
       (DFA_Inv (Sigma * Sigma) (ProductDFA ParentL ParentR) DProd))
    (step_spec state Sigma CHrep PairDFA
       (DFA_Inv Sigma PairDFA DPair)).
  Proof.
    do_funspec_sub. destruct w as [[[[st gv] v] c] d]; Intros.
    Exists (st, gv, v, c, (d, d)) emp. entailer!.
  Qed.

  Lemma DFA_ProductPair' x:
    DFA (Sigma * Sigma) (fun c v => CHrep (fst c) v /\ CHrep (snd c) v)
        (ProductDFA ParentL ParentR) DProd M state dfaops ops i f s cs x
    |-- DFA Sigma CHrep PairDFA DPair M state dfaops ops i f s cs x.
  Proof.
    rewrite 2 DFA_fold_unfold. Intros fptrs cso m. Exists fptrs cso m.
    destruct x as [[st gv] v];  entailer!.
    apply sepcon_derives; [ apply derives_refl | simpl]. 
    clear. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer! | trivial ].
    apply func_ptr'_mono. apply step_funspec_sub'.
  Qed.

  Lemma makeProduct_makePair':
        funspec_sub (@makeProduct_spec _ _ ParentL ParentR CHrep CHrep DParentL DParentR M state dfaops ops i f s cs DProd)
                    (@makePair_spec _ ParentL ParentR CHrep  DParentL DParentR M state dfaops ops i f s cs DPair).
  Proof. do_funspec_sub. Exists w emp.
    destruct w as [[[[l r] stL] stR] gv]. 
    (*entailer! takes 10secs...*) Intros; apply andp_right; [apply prop_right; intuition; entailer! | cancel].
    Exists (eval_id ret_temp rho'); entailer!.
    apply DFA_ProductPair'.
  Qed.

  Lemma step_funspec_sub_inv' (H: forall d1 d2 c, CHrep d1 c -> CHrep d2 c -> d1=d2):
    funspec_sub (step_spec state Sigma CHrep PairDFA (DFA_Inv Sigma PairDFA DPair))
                (step_spec state (Sigma * Sigma)
                                 (fun c v =>  CHrep (fst c) v /\ CHrep (snd c) v)
                                 (ProductDFA ParentL ParentR)
                                 (DFA_Inv (Sigma * Sigma) (ProductDFA ParentL ParentR) DProd)).
  Proof. do_funspec_sub. destruct w as [[[[st gv] v] c] [d1 d2]]; Intros.
    apply (H  _ _ _ H1) in H2. simpl in *; subst d2.
    Exists (st, gv, v, c, d1) emp. entailer!.
  Qed.

  Lemma DFA_PairProduct' (H: forall d1 d2 c, CHrep d1 c -> CHrep d2 c -> d1=d2) x:
    DFA Sigma CHrep PairDFA DPair M state dfaops ops i f s cs x |--
    DFA (Sigma * Sigma) (fun c v => CHrep (fst c) v /\ CHrep (snd c) v)
        (ProductDFA ParentL ParentR) DProd M state dfaops ops i f s cs x.
  Proof.
    rewrite 2 DFA_fold_unfold. Intros fptrs cso m. Exists fptrs cso m.
    destruct x as [[st gv] v]; entailer!.
    Transparent ObjectBase.funcptrs. simpl. Opaque ObjectBase.funcptrs.
    (*apply sepcon_derives; [ apply derives_refl | simpl]. *)
    clear - H. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear - H. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer!; apply derives_refl |].
    clear - H. destruct fptrs; trivial. Intros sh u. Exists sh u.
    apply sepcon_derives; [ entailer! | trivial ].
    apply func_ptr'_mono. apply (step_funspec_sub_inv' H).
  Qed.

  Lemma makePair_makeProduct' (H: forall d1 d2 c, CHrep d1 c -> CHrep d2 c -> d1=d2):
        funspec_sub (@makePair_spec _ ParentL ParentR CHrep DParentL DParentR M state dfaops ops i f s cs DPair)
                    (@makeProduct_spec _ _ ParentL ParentR CHrep CHrep DParentL DParentR M state dfaops ops i f s cs DProd).
  Proof. do_funspec_sub. Exists w emp.
    destruct w as [[[[l r] stL] stR] gv].
    (*entailer! takes 10secs...*) Intros; apply andp_right; [apply prop_right; intuition; entailer! | cancel].
    intros; Exists (eval_id ret_temp rho'); entailer!.
    apply (DFA_PairProduct' H).
  Qed.
End Pair2Prod.

(*The following alternative construction of Pair from Product proposed by Dave Naumann does
  not require a condition on CHrep(s), as it yields the identical DFA predicate, for any CHrep*)
Section Prod2XPair.
  Variable SigmaL SigmaR: Type.
  Variable ParentL: @Dfa SigmaL.
  Variable ParentR: @Dfa SigmaR.

  Definition LeftDFA: @Dfa (SigmaL * SigmaR) :=
  {| Dfa_state := Dfa_state ParentL;
     Dfa_s := Dfa_s ParentL;
     Dfa_fin := Dfa_fin ParentL;
     Dfa_trans := fun s d => Dfa_trans ParentL s (fst d) |}.

  Definition RightDFA: @Dfa (SigmaL * SigmaR) :=
  {| Dfa_state := Dfa_state ParentR;
     Dfa_s := Dfa_s ParentR;
     Dfa_fin := Dfa_fin ParentR;
     Dfa_trans := fun s d => Dfa_trans ParentR s (snd d) |}.

  Notation ProdDFA := (ProductDFA ParentL ParentR).
  Notation XPairDFA:= (PairDFA LeftDFA RightDFA).

  Variable DParentL: DFA_APD SigmaL ParentL.
  Variable DParentR: DFA_APD SigmaR ParentR.

  Variable DProd: DFA_APD _ ProdDFA.
  Definition XDFA_APD: DFA_APD (SigmaL * SigmaR) XPairDFA := DProd.

  Lemma DFA_ProductX (CHrep: SigmaL * SigmaR->val->Prop):
  @DFA (SigmaL * SigmaR) CHrep _ DProd =
  @DFA (SigmaL * SigmaR) CHrep _ XDFA_APD.
  Proof. reflexivity. Qed.
End Prod2XPair.
