Require Import VST.floyd.proofauto.
Require Import ObjectBase.
Require Export DFA.
Require Export DFAbase.

Section DFA_specs.
Variable ch:Type.
Variable CHrep: ch -> val -> Prop.

Variable A: @Dfa ch.

Definition st_init_spec (P: GPred (Dfa_state A)):funspec := 
    WITH s:Dfa_state A, v:val, gv: globals
    PRE [ ]
     PROP()
     PARAMS () GLOBALS (gv)
     SEP(P (s, gv ,v))
    POST[ tvoid ]
      PROP()
      RETURN ()
      SEP(P (Dfa_s A, gv,v)).

Definition st_final_spec (P: GPred (Dfa_state A)):funspec := 
    WITH s:Dfa_state A, v:val, gv: globals
    PRE [ ]
     PROP()
     PARAMS () GLOBALS (gv)
     SEP(P (s, gv, v))
    POST[ tbool ]
      PROP()
      RETURN (if Dfa_fin A s then Vtrue else Vfalse)
      SEP(P (s, gv, v)).

Definition st_step_spec (P: GPred (Dfa_state A)):funspec := 
    WITH s:Dfa_state A, v:val,gv: globals, c:val, d:ch
    PRE [ tuchar ]
     PROP(CHrep d c)
     PARAMS (c) GLOBALS (gv)
     SEP(P (s, gv, v))
    POST[ tvoid ]
      PROP()
      RETURN ()
      SEP(P (Dfa_trans A s d, gv, v)).

Definition SDFA_ops_specs:list (GPred (Dfa_state A) -> funspec):= [st_init_spec; st_final_spec; st_step_spec].

End DFA_specs.

Section MkSDFAspec.

Variable char: Type.
Variable CHrep: char -> val -> Prop.

Variable A: @Dfa char.
Variable D: DFA_APD _ A.

Variable staticops:type.
Variable ops i f s:ident.
Variable cs: compspecs.

Definition SDFA: GPred (Dfa_state A) :=
           staticobj_rep cs (Dfa_state A) (DFA_Inv char A D) staticops [i; f; s] (SDFA_ops_specs char CHrep A).

Lemma SDFA_fold_unfold:
  SDFA = fun x => EX fptrs, DFA_Inv char A D x *
                            MTable A staticops [i; f; s] (SDFA_ops_specs char CHrep A) fptrs (snd x) ((*S*)DFA_Inv char A D).
Proof. reflexivity. Qed.
Lemma SDFA_fold_unfold':
  SDFA = fun x => (DFA_Inv char A D x *
                   EX fptrs, MTable A staticops [i; f; s] (SDFA_ops_specs char CHrep A) fptrs (snd x) (DFA_Inv char A D))%logic.
Proof. rewrite SDFA_fold_unfold. extensionality x. rewrite <- exp_sepcon2. reflexivity. Qed.

(*Variable DFAClass: globals -> mpred. (*location predicate for globals*)*)

(*Variable D:DFA_APD _ A.*)

Definition makeStaticDFA_spec : funspec :=
 WITH gv:globals
 PRE [ ]
    PROP () PARAMS() GLOBALS (gv)
    SEP (DFA_GV _ _ D gv)
 POST [ tptr staticops ]
    EX p: val, PROP () RETURN (p)
     SEP ((*SDFAClass _ _ D gv;*) (*Static objects are only instantiated once, so the "class" disappears*)
          SDFA (Dfa_s A, gv, p)).

End MkSDFAspec.

Lemma StaticDFA_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) char CHrep A D staticops i f s
      (Hstaticops: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) staticops = true)
      (Hidents: forall i0 : ident, @In ident i0 [i; f; s] ->
                @reptype cs1 (@nested_field_type cs1 staticops (DOT i0)) =
                @reptype cs2 (@nested_field_type cs2 staticops (DOT i0))):
      SDFA char CHrep A D staticops i f s cs1 = SDFA char CHrep A D staticops i f s cs2.
Proof. (*apply exp_congr. intros fptrs.*) apply (staticobj_rep_change_compspecs _ _ CCE12); trivial. Qed.


