Require Import VST.floyd.proofauto.
(*Require Import spec_stdlib. Includes ASI_malloc, and hence Malloc_APD 
Require Import surelyMalloc.*)
Global Open Scope funspec_scope.

Local Open Scope assert.

Parameter WORD: Z. (* := 4.  (* sizeof(size_t) is 4 for 32bit Clight *)*)
Parameter ALIGN:Z. (*:=2*)
Definition WA: Z := (WORD*ALIGN) - WORD. (* WASTE at start of block *)

(*The final two APDs are already defined in my file ASI_malloc, which we'll eventually import*)
Record MallocTokenAPD := {
  malloc_token': share -> Z -> val -> mpred;
  malloc_token'_valid_pointer: forall sh sz p, 
      malloc_token' sh sz p |-- valid_pointer p;
  malloc_token'_local_facts:  forall sh sz p, 
      malloc_token' sh sz p |-- !! malloc_compatible sz p;
}.

Record MallocFreeAPD := {
  MF_Tok :> MallocTokenAPD;
  mem_mgr: globals -> mpred;
}.
#[export]Hint Resolve malloc_token'_local_facts : saturate_local.
#[export]Hint Resolve malloc_token'_valid_pointer : saturate_local.

Definition malloc_token {cs:compspecs} M (sh: share) (t: type) (p: val): mpred := 
   !! field_compatible t [] p && 
   malloc_token' M sh (sizeof t) p.

Lemma malloc_token_valid_pointer: forall {cs: compspecs} M sh t p, 
      malloc_token M sh t p |-- valid_pointer p.
Proof. intros. unfold malloc_token.
 apply andp_left2. apply malloc_token'_valid_pointer.
Qed.
#[export]Hint Resolve malloc_token'_valid_pointer : valid_pointer.
#[export]Hint Resolve malloc_token_valid_pointer : valid_pointer.

Lemma malloc_token_local_facts:  forall {cs: compspecs} M sh t p,
      malloc_token M sh t p |-- !! (field_compatible t [] p /\ malloc_compatible (sizeof t) p).
Proof. intros.
 unfold malloc_token.
 normalize. rewrite prop_and.
 apply andp_right. apply prop_right; auto.
 apply malloc_token'_local_facts.
Qed.

#[export]Hint Resolve malloc_token'_local_facts : saturate_local.
#[export]Hint Resolve malloc_token_local_facts : saturate_local.

Section SM_ASI.
Variable M:MallocFreeAPD.
Definition free_spec' :=
(* DECLARE (*_free*)freeID*)
   WITH n:Z, p:val, gv: globals
   PRE [ tptr tvoid ]
       PROP ()
       PARAMS (p) GLOBALS (gv)
       SEP (mem_mgr M gv;
            if eq_dec p nullval then emp
            else (malloc_token' M Ews n p * memory_block Ews n p))
    POST [ Tvoid ]
       PROP ()
       LOCAL ()
       SEP (mem_mgr M gv).


Definition surely_malloc_numericspec :=
(* DECLARE _surely_malloc*)
   WITH n:Z, gv: globals
   PRE [ size_t ]
       PROP (0 <= n <= Ptrofs.max_unsigned - (WA+WORD))
       PARAMS (Vptrofs (Ptrofs.repr n))
       GLOBALS (gv)
       SEP (mem_mgr M gv)
    POST [ tptr tvoid ] EX p:_,
       PROP ()
       RETURN (p)
       SEP (mem_mgr M gv; malloc_token' M Ews n p; memory_block Ews n p).

Definition surely_malloc_spec {cs:compspecs} :=
(*  DECLARE _surely_malloc*)
   WITH t:type, gv: globals (*, cs:compspecs -- universe inconsistency*)
   PRE [ (*tuint*)size_t ]
       PROP (0 <= @sizeof cs t <= Int.max_unsigned - (WA+WORD);
                @complete_legal_cosu_type (@cenv_cs cs) t = true;
                @natural_aligned cs natural_alignment t = true)
       PARAMS (if Archi.ptr64
               then Vlong (Ptrofs.to_int64 (Ptrofs.repr(@sizeof cs t)))
               else Vint (Int.repr (@sizeof cs t)))
       GLOBALS (gv)
       SEP (mem_mgr M gv)
    POST [ tptr tvoid ] EX p:_,
       PROP ()
       RETURN (p)
       SEP (mem_mgr M gv; malloc_token M Ews t p; data_at_ Ews t p).


Definition SM_ASI free surely : funspecs := [(free,free_spec');(surely, @surely_malloc_numericspec)].

Lemma surelymalloc_spec_sub:
 forall {cs: compspecs}, 
   funspec_sub surely_malloc_numericspec (@surely_malloc_spec cs).
Proof.
do_funspec_sub. destruct w as [t gv]. clear H. simpl.
Exists (sizeof t, gv) emp. simpl; entailer!.
intros tau ? ?. Exists (eval_id ret_temp tau).
entailer!.
unfold malloc_token.
assert_PROP (field_compatible t [] (eval_id ret_temp tau)).
{ entailer!.
  apply malloc_compatible_field_compatible; auto. }
entailer!.
rewrite memory_block_data_at_; auto.
Qed.

Lemma natural_aligned_TarrayD {cs} t z a:
  @natural_aligned cs natural_alignment (Tarray t z a) = true ->
  @natural_aligned cs natural_alignment t = true.
Proof.
  intros. apply andb_true_iff. apply andb_true_iff in H; destruct H.
  simpl in *. split; trivial.
  unfold is_aligned, is_aligned_aux in *. simpl in *.
  apply andb_true_iff in H0. destruct H0. apply andb_true_iff in H0. destruct H0.
  apply andb_true_iff; split; trivial.
Qed.

Lemma ha_sub {cs cs'} (HSUB: ha_env_cs_sub (@ha_env_cs cs) (@ha_env_cs cs')): forall t
    (L: @complete_legal_cosu_type (@cenv_cs cs) t = true),
    hardware_alignof (@ha_env_cs cs) t = hardware_alignof (@ha_env_cs cs') t.
Proof.
  induction t; simpl; intros; auto.
  + remember (@Maps.PTree.get composite i (@cenv_cs cs)) as q. destruct q; [symmetry in Heqq | discriminate].
    specialize (HSUB i); red in HSUB.
    assert (exists ha : Z, @Maps.PTree.get Z i (@ha_env_cs cs) = @Some Z ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct H as [ha HA]. rewrite HA in *.
    rewrite HSUB; trivial.
  + remember (@Maps.PTree.get composite i (@cenv_cs cs)) as q. destruct q; [symmetry in Heqq | discriminate].
    specialize (HSUB i); red in HSUB.
    assert (exists ha : Z, @Maps.PTree.get Z i (@ha_env_cs cs) = @Some Z ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct H as [ha HA]. rewrite HA in *.
    rewrite HSUB; trivial.
Qed.
  
Lemma natural_aligned_cspecs_sub {cs cs'} (C: cspecs_sub cs cs'): forall t,
    @complete_legal_cosu_type (@cenv_cs cs) t = true ->
    @natural_aligned cs natural_alignment t = true ->
    @natural_aligned cs' natural_alignment t = true.
Proof.
  induction t; simpl; intros; auto.
  + specialize (IHt H (natural_aligned_TarrayD _ _ _ H0)). 
    apply andb_true_iff in H0; destruct H0. apply andb_true_iff in IHt; destruct IHt.
    apply andb_true_iff. simpl in *. split; trivial. 
    unfold is_aligned, is_aligned_aux in *. simpl in *.
    rewrite andb_true_r in *.
    apply andb_true_iff in H1; destruct H1.
    apply andb_true_iff. split; trivial.
    specialize (complete_legal_cosu_type_complete_type _ _ H); intros.
    destruct C as [C1 [C2 C3]].
    rewrite (@expr_lemmas4.cenv_sub_sizeof _ _ C1); trivial.
    rewrite <- (ha_sub C2); trivial.
  + unfold natural_aligned in *; simpl in *.
    remember (@Maps.PTree.get composite i (@cenv_cs cs)) as q; destruct q; [ symmetry in Heqq | discriminate].
    destruct C as [CEsub [HAsub LAsub]].
    specialize (HAsub i); red in HAsub.
    assert (HA: exists ha : Z, @Maps.PTree.get Z i (@ha_env_cs cs) = Some ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct HA as [ha HA]. rewrite HA in *.
    rewrite HAsub.
    apply andb_true_iff in H0; destruct H0.
    rewrite H0; simpl. 
    unfold is_aligned, is_aligned_aux in *.
    rewrite andb_true_r in *.
    specialize (LAsub i); red in LAsub.
    assert (LA: exists la, @Maps.PTree.get legal_alignas_obs i (@la_env_cs cs)  = Some la).
    { apply (@la_env_cs_complete cs i). eexists; eassumption. }
    unfold legal_alignas_type, legal_alignas_obs in *.
    destruct LA as [la LA]. rewrite LA in *. rewrite LAsub in *; trivial.
  + (*same proof as in previous case*)
    unfold natural_aligned in *; simpl in *.
    remember (@Maps.PTree.get composite i (@cenv_cs cs)) as q; destruct q; [ symmetry in Heqq | discriminate].
    destruct C as [CEsub [HAsub LAsub]].
    specialize (HAsub i); red in HAsub.
    assert (HA: exists ha : Z, @Maps.PTree.get Z i (@ha_env_cs cs)  = Some ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct HA as [ha HA]. rewrite HA in *.
    rewrite HAsub.
    apply andb_true_iff in H0; destruct H0.
    rewrite H0; simpl. 
    unfold is_aligned, is_aligned_aux in *.
    rewrite andb_true_r in *.
    specialize (LAsub i); red in LAsub.
    assert (LA: exists la, @Maps.PTree.get legal_alignas_obs i (@la_env_cs cs)  = Some la).
    { apply (@la_env_cs_complete cs i). eexists; eassumption. }
    unfold legal_alignas_type, legal_alignas_obs in *.
    destruct LA as [la LA]. rewrite LA in *. rewrite LAsub in *; trivial.
Qed.

Lemma cspecs_sub_cs_preserve_type {cs1 cs2} (CSUB:cspecs_sub cs1 cs2) t x i
      (T: t = field_type i (co_members (get_co x))) coeq
  (C: coeq = cs_preserve_env cs1 cs2):
  cs_preserve_type cs1 cs2 coeq t = true.
Proof.
Abort.
End SM_ASI.