#include "Client.h"

#include "SingleDFA.h"
#include "StaticWeekdays.h"
#include "StaticTrueFalse.h"
#include "TrueFalse.h"
#include "Weekdays.h"
#include "T_Weekdays.h"
#include "Twice_Weekdays.h"
#include "ComplementDFA.h"
#include "ProductDFA.h"
#include "IntersectionDFA.h"
#include <stdio.h>

struct storage {
  StaticOps ops_ref;
  Object obj_ref;
};

struct storage strg; /* = { NULL, NULL};*/

void invokeInit (Object s) {
  s -> ops -> init (s);
}
void invokeStep (Object s, uchar c) {
  s -> ops -> step (s, c);
}
bool invokeFinal (Object s) {
  return (s -> ops -> final (s));
}
void invokeTwosteps (Object s, uchar c1, uchar c2) {
  ((struct t_dfaOps *)(s -> ops)) -> twosteps (s, c1, c2);
}

bool runStaticDFAs(boolbinop f) {
  DFA_init();

  StaticOps w = make_StaticWeekdays();
  StaticOps tf = make_StaticTrueFalse();
  strg.ops_ref = tf;
  w -> st_init ();
  tf -> st_init ();
  strg.ops_ref  -> st_init ();
  w -> st_step ('x');
  tf -> st_step ('1');
  DFA_step ('1');
  w -> st_step ('n');
  DFA_step ('0');
  /*tf*/strg.ops_ref -> st_step ('0');
  return (f (w -> st_final (), f (tf -> st_final (), DFA_final())));
}

Object createWeekdayDFA(bool b) {
  if (b) {
    return (make_TWeekdays ()); 
  } else {
    return (make_Weekdays()); 
  }
}

bool runClient(void) {
  Object tf = make_TrueFalse();
  Object W = createWeekdayDFA(true);
  strg.obj_ref = W;

  tf -> ops -> init (tf);
  W -> ops -> init (W);
  tf -> ops -> step (tf, '0');
  W -> ops -> step (W, 'x');
  strg.obj_ref -> ops -> step (W, 'n');
  /*return (tf -> ops -> final (tf) && W -> ops -> final (W));*/
  return (invokeFinal (tf) && invokeFinal (W));
}

bool runTClient(void) {
  Object W = make_TWeekdays ();

  W -> ops -> init (W);
  ((struct t_dfaOps *) (W -> ops)) -> twosteps (W, 'x', 'n');
  invokeTwosteps (W, 'x', 'n');
  return (W -> ops -> final (W));
}

bool runTwiceClient(void) {
  Object W = make_TwiceWeekdays ();

  W -> ops -> init (W);
  ((struct t_dfaOps *) (W -> ops)) -> twosteps (W, 'x', 'n'); /*method is still called twostep!*/
  return (W -> ops -> final (W));
}

bool runClientComplement(void) {
  Object W = make_Weekdays ();

  W -> ops -> init (W);
  W -> ops -> step (W, 'x');
  Object U = mkCompl(W);
  U -> ops -> step (U, 'n');
  return (U -> ops -> final (U));
}

bool runClientBoth(boolbinop f) {
  Object P = mkProduct(make_Weekdays (),
                                           mkCompl(make_Weekdays ()));
  bool b = P-> ops -> final (P);

  strg.obj_ref = make_IntersectionDFA(make_Weekdays (), P);
 
  return (f (b, strg.obj_ref -> ops -> final (strg.obj_ref)));
}
