Record Dfa {Sigma : Type} := {
    Dfa_state :> Type;
    Dfa_s : Dfa_state;
    Dfa_fin : Dfa_state -> bool;
    Dfa_trans : Dfa_state -> Sigma -> Dfa_state
  }.

Definition ComplementDFA {Sigma} (Parent:@Dfa Sigma): @Dfa Sigma :=
   Build_Dfa Sigma (Dfa_state Parent)
                   (Dfa_s Parent)
                   (fun s => negb (Dfa_fin Parent s))
                   (Dfa_trans Parent).

Definition ProductDFA {SigmaL SigmaR} (ParentL: @Dfa SigmaL) (ParentR: @Dfa SigmaR): 
           @Dfa (SigmaL * SigmaR)%type :=
    Build_Dfa (SigmaL * SigmaR)%type 
              ((Dfa_state ParentL * Dfa_state ParentR)%type)
              (Dfa_s ParentL, Dfa_s ParentR)
              (fun s => orb (Dfa_fin ParentL (fst s)) (Dfa_fin ParentR (snd s)))
              (fun s c => (Dfa_trans ParentL (fst s) (fst c), Dfa_trans ParentR (snd s) (snd c))).

Definition IntersectionDFA {SigmaL SigmaR} (ParentL: @Dfa SigmaL) (ParentR: @Dfa SigmaR):
           @Dfa (SigmaL * SigmaR)%type :=
    Build_Dfa (SigmaL * SigmaR)%type 
              ((Dfa_state ParentL * Dfa_state ParentR)%type)
              (Dfa_s ParentL, Dfa_s ParentR)
              (fun s => andb (Dfa_fin ParentL (fst s)) (Dfa_fin ParentR (snd s)))
              (fun s c => (Dfa_trans ParentL (fst s) (fst c), Dfa_trans ParentR (snd s) (snd c))).

Definition PairDFA {Sigma} (ParentL: @Dfa Sigma) (ParentR: @Dfa Sigma):
           @Dfa (Sigma) :=
    Build_Dfa Sigma 
              ((Dfa_state ParentL * Dfa_state ParentR)%type)
              (Dfa_s ParentL, Dfa_s ParentR)
              (fun s => orb (Dfa_fin ParentL (fst s)) (Dfa_fin ParentR (snd s)))
              (fun s c => (Dfa_trans ParentL (fst s) c, Dfa_trans ParentR (snd s) c)).


Record DfaMorphism {Sigma Sigma'} (A:@Dfa Sigma) (A':@Dfa Sigma') := {
  DfaMorph_Sigma:  Sigma -> Sigma';
  DfaMorph_state:  Dfa_state A -> Dfa_state A';
  DfaMorph_s: DfaMorph_state (Dfa_s A) = Dfa_s A';
  DfaMorph_fin: forall s , Dfa_fin A s = Dfa_fin A' (DfaMorph_state s);
  DfaMorph_trans: forall s c, DfaMorph_state (Dfa_trans A s c) =
                              Dfa_trans A' (DfaMorph_state s) (DfaMorph_Sigma c);
}.