From mathcomp Require Import all_ssreflect.
Require Import RegLang.dfa.

Require VST.msl.Axioms.

Lemma exist_congr A P x Hx y Hy (XY:x=y): @exist A P x Hx = @exist A P y Hy.
Proof. subst. f_equal. apply Axioms.proof_irr. Qed.

Require Import DFA.

Record Equiv (T:Type) (t:finType) := {
  Equiv_F: T -> t;
  Equiv_f: t -> T;
  Equiv_Tt: forall x, Equiv_F (Equiv_f x) = x;
  Equiv_tT: forall y, Equiv_f (Equiv_F y) = y
}.

Record EquivDFA {C c} (Cc: Equiv C c) (A: @Dfa C) (a:@dfa c) := {
  Equiv_states: Equiv (Dfa_state A) (dfa_state a);
  Equiv_s: Equiv_F _ _ Equiv_states (Dfa_s A) = (dfa_s a);
  Equiv_fin: forall s, Dfa_fin A s = true <-> (Equiv_F _ _ Equiv_states s) \in (dfa_fin a);
  Equiv_trans: forall s x, Equiv_F _ _ Equiv_states (Dfa_trans A s x) = 
                           @dfa_trans _ a (Equiv_F _ _ Equiv_states s) (Equiv_F _ _ Cc x) }.

Lemma Equiv_s' {C c Cc A a} (E: @EquivDFA C c Cc A a):
  Equiv_f _ _ (Equiv_states _ _ _ E) (dfa_s a) = (Dfa_s A).
Proof. destruct E as [E1 E2 E3 E4]; simpl in *.
  rewrite <- E2; clear E2 E3 E4.
  destruct E1 as [F1 F2 F3 F4]. simpl. rewrite F4; trivial. 
Qed.
Lemma Equiv_fin' {C c Cc A a} (E: @EquivDFA C c Cc A a):
  forall s, Dfa_fin A (Equiv_f _ _ (Equiv_states _ _ _ E) s) = true <-> s \in (dfa_fin a). 
Proof. destruct E as [E1 E2 E3 E4]; simpl in *. intros.
  destruct (E3 (Equiv_f A a E1 s)) as [X Y].
  clear - X Y E1.
  destruct E1 as [F1 F2 F3 F4]. simpl in *. split; intros.
  + apply X in H. rewrite F3 in H; trivial.
  + apply Y. rewrite F3; trivial.
Qed.

Lemma Equiv_trans' {C c Cc A} {a:@dfa c} (E: @EquivDFA C c Cc A a):
  forall s x,
  Equiv_f _ _ (Equiv_states _ _ _ E) (@dfa_trans _ a s x) = 
                           @Dfa_trans _ A (Equiv_f _ _ (Equiv_states _ _ _ E) s) (Equiv_f _ _ Cc x).
Proof. destruct E as [E1 E2 E3 E4]; simpl in *. intros.
  specialize (E4 (Equiv_f A a E1 s) (Equiv_f _ _ Cc x)).
  clear - E4 E1.
  destruct E1 as [F1 F2 F3 F4]. simpl in *.
  rewrite F3 in E4.
  destruct Cc as [C1 C2 C3 C4]; simpl in *. rewrite C3 in E4.
  rewrite <- E4, F4. trivial.
Qed.

Lemma EquivCompl {C c Cc A a} (E: @EquivDFA C c Cc A a): @EquivDFA C c Cc (ComplementDFA A) (dfa_compl a).
destruct E as [E_states E_s E_fin E_trans].
eapply (Build_EquivDFA _ _ Cc (ComplementDFA A) (dfa_compl a) E_states E_s). 2: apply E_trans.
clear- E_fin. simpl; intros. destruct (E_fin s) as [E1 E2]; clear E_fin. split; intros.
+ rewrite in_set. destruct (in_mem (Equiv_F A a E_states s) (mem (dfa_fin a))); simpl in *; trivial.
  rewrite E2 in H; trivial.
+ rewrite in_set in H. destruct ((Dfa_fin A s)); simpl in *; trivial.
  rewrite E1 in H; trivial.
Qed.

Definition EquivPair_states {A a B b} (EA: Equiv A a) (EB:Equiv B b): Equiv (A*B)%type (prod_finType a b).
destruct EA as [F f Ff fF]. destruct EB as [G g Gg gG].
apply (Build_Equiv (A * B) (prod_finType a b) (fun x => (F (fst x), G (snd x))) (fun y => (f (fst y), g (snd y)))).
simpl; intros. rewrite Ff; rewrite Gg; destruct x; trivial.
simpl; intros. rewrite fF; rewrite gG; destruct y; trivial.
Defined.

Lemma EquivPair {C c Cc A a B b} (E: @EquivDFA C c Cc A a) (F: @EquivDFA C c Cc B b):
      @EquivDFA C c Cc (PairDFA A B) (dfa_op orb a b).
destruct E as [E_states E_s E_fin E_trans].
destruct F as [F_states F_s F_fin F_trans].
apply (Build_EquivDFA _ _ Cc (PairDFA A B) (dfa_op orb a b) (EquivPair_states E_states F_states)).
+ destruct E_states as [E1 E2 E3 E4]; simpl in *.
  destruct F_states as [F1 F2 F3 F4]; simpl in *. clear - E_s F_s.
  f_equal; trivial.
+ destruct E_states as [E1 E2 E3 E4]; simpl in *.
  destruct F_states as [F1 F2 F3 F4]; simpl in *.
  simpl; intros. destruct s as [sA sB]. clear - E_fin F_fin.
  destruct (E_fin sA) as [Efin1 Efin2]. destruct (F_fin sB) as [Ffin1 Ffin2].
  clear - Efin1 Efin2 Ffin1 Ffin2.
  split; intros.
  { rewrite in_set. simpl in *. destruct (Dfa_fin A sA); simpl in *. 
    { rewrite Efin1; trivial. } { rewrite Ffin1; trivial. rewrite orbT; trivial. } }
  { rewrite in_set in H; simpl in *. destruct (Dfa_fin A sA); simpl in *; trivial.
    apply Ffin2. destruct ((E1 sA \in dfa_fin a)); simpl in *; trivial. intuition. }
+ destruct E_states as [E1 E2 E3 E4]; simpl in *.
  destruct F_states as [F1 F2 F3 F4]; simpl in *.
  intros [sA sB] x. clear - E_trans F_trans. simpl.
  f_equal; trivial.
Qed.
(*
(*
Record Diag (T:Type):Type := { Diag_x : T * T; Diag_X: fst Diag_x = snd Diag_x}.
Lemma DiagEq (T:Type)(D1 D2: Diag T) (H: Diag_x _ D1 = Diag_x _ D2): D1=D2.
destruct D1; destruct D2; simpl in *. apply Axioms.proof_irr. rewrite H.  apply proof_irrelevance. f_equal.  = { Diag_x : T * T; Diag_X: fst Diag_x = snd Diag_x}.

Definition DiagF {T U} (F:T->U) (d:Diag T): U := F (fst (Diag_x _ d)).

Program Definition Diagf {T U} (f:U->T) (u:U): Diag T := Build_Diag _ (f u, f u) _.

Definition Equiv_ProdSigma {C c} (Cc : Equiv C c): Equiv (Diag C) c.
destruct Cc as [F f Ff fF].
apply (Build_Equiv (Diag C) c (DiagF F) (Diagf f) Ff).
intros [[y1 y2] Y]; simpl in *; subst. unfold Diagf, DiagF. simpl.
  rewrite (fF y2). rewrite fF.   f_equal.  unfold Diagf.
assert (HH: (f (F y2), f (F y2)).1 = (f (F y2), f (F y2)).2).
{ simpl. trivial. }
assert(GG: (f (F y2), f (F y2)) = (y2, y2)). rewrite fF; trivial.
apply eq_existT_uncurried. exists GG. apply proof_irr. *)

Definition Diag (T:Type):Type := { x : T * T | fst x = snd x}.

Definition DiagF {T U} (F:T->U):Diag T -> U := fun d => match d with exist x Hx => F (fst x) end.

Program Definition Diagf {T U} (f:U->T):U -> Diag T:= fun x => exist _ (f x, f x) _.

Definition Equiv_ProdSigma {C c} (Cc : Equiv C c): Equiv (Diag C) c.
destruct Cc as [F f Ff fF].
apply (Build_Equiv (Diag C) c (DiagF F) (Diagf f) Ff).
intros [[y1 y2] Y]; simpl in *; subst. unfold Diagf.
apply exist_congr. rewrite fF; trivial.
Qed.

(*Program Definition Equiv_ProdSigma {C c} (Cc : Equiv C c): Equiv (Diag C) c :=
 Build_Equiv (Diag C) c (fun d => match d with exist x Hx => (Equiv_F _ _ Cc) (fst x) end)
             (fun x => exist _ ((Equiv_f _ _ Cc)  x, (Equiv_f _ _ Cc)  x) _) _ _.
 
Definition Equiv_ProdSigma {C c} (Cc : Equiv C c): Equiv (Diag C)%type c.
destruct Cc as [F f Ff fF].
specialize (Build_Equiv (Diag C) c (fun d => match d with exist x Hx => (Equiv_F _ _ Cc) (fst x) end)).
 (fun y =>fun x => exist _ ((Equiv_f _ _ Cc)  x, (Equiv_f _ _ Cc)  x) _). 
assert (G: forall _ : Diag C, c).
{ intros [[x1 x2] Hx]. simpl in Hx. apply (F x1).
specialize (Build_Equiv (Diag C) c). *)

Lemma ProductCompl {C c Cc A a B b} (E: @EquivDFA C c Cc A a) (F: @EquivDFA C c Cc B b):
      @EquivDFA _ _ (Equiv_ProdSigma Cc) (ProductDFA A B) (dfa_op orb a b).
destruct E as [E_states E_s E_fin E_trans].
eapply (Build_EquivDFA _ _ Cc (ComplementDFA A) (dfa_compl a) E_states E_s). 2: apply E_trans.
clear- E_fin. simpl; intros. destruct (E_fin s) as [E1 E2]; clear E_fin. split; intros.
+ rewrite in_set. destruct (in_mem (Equiv_F A a E_states s) (mem (dfa_fin a))); simpl in *; trivial.
  rewrite E2 in H; trivial.
+ rewrite in_set in H. destruct ((Dfa_fin A s)); simpl in *; trivial.
  rewrite E1 in H; trivial.
Qed.
*)

Require Import TrueFalseDFA.

Definition TF_step (s a: bool_finType) : bool_finType := a.
Definition TrueFalse_dfa: @dfa bool_finType :=
 @Build_dfa bool_finType bool_finType false (set1 true) TF_step.

Program Definition Bool_equiv: Equiv bool bool_finType :=
  Build_Equiv _ _ (fun b => b) (fun b => b) _ _.

Program Definition TF_equiv:= Build_EquivDFA _ _ Bool_equiv TrueFalseDFA TrueFalse_dfa
  Bool_equiv _ _ _ .
Next Obligation.
rewrite in_set1; unfold eq_op; simpl. destruct s; intuition.
Qed.

Require Import ZArith.
Require Import Lia.
Require Import WeekdayDFA.

Program Definition WDState_Equiv_F (W:Weekday): ordinal_finType 7 :=
@Ordinal 7 (Z.to_nat (proj1_sig W)) _.
Next Obligation.
destruct W; simpl.
specialize (@leP (S (Z.to_nat x)) 7). intros Y.
destruct Y; trivial. lia.
Qed.

Program Definition WDState_Equiv_f (D:ordinal_finType 7):Weekday :=
  match D with @Ordinal _ m i => @exist Z (fun z : Z => and (Z.le 0 z) (Z.lt z 7)) (Z.of_nat m) _
  end.
Next Obligation.
specialize (@leP (S m) 7). intros Y.
destruct Y. lia. split. lia. exfalso; intuition.
Qed.

Lemma Ordinal_congr n m m' i i' (Hm: m=m'):
  @Ordinal n m i = @Ordinal n m' i'.
subst. f_equal. apply Axioms.proof_irr.
Qed.

Program Definition WDState_Equiv: Equiv Weekday (ordinal_finType 7) :=
      Build_Equiv Weekday (ordinal_finType 7) WDState_Equiv_F WDState_Equiv_f _ _ .
Next Obligation.
destruct x. simpl. apply Ordinal_congr. simpl. rewrite Nat2Z.id; trivial. Qed.
Next Obligation.
destruct y. simpl. apply exist_congr. rewrite Z2Nat.id; trivial. apply a. Qed.

  Definition day := ordinal_finType 7.
  Program Definition monday:day := @Ordinal 7 0 _.
  Program Definition tuesday:day := @Ordinal 7 1 _.
  Program Definition wednesday:day := @Ordinal 7 2 _.
  Program Definition thursday:day := @Ordinal 7 3 _.
  Program Definition friday:day := @Ordinal 7 4 _.
  Program Definition saturday:day := @Ordinal 7 5 _.
  Program Definition sunday:day := @Ordinal 7 6 _.

  Definition weekday_dfa:@dfa (ordinal_finType 256%nat).
 apply (@Build_dfa _ (ordinal_finType 7%nat) monday).
   simpl. apply set1. apply sunday.
   simpl. intros D [x X]. 
   remember (beq_nat x (Byte.to_nat "n")); destruct b.
   2: apply D. clear Heqb X x.
   remember (eq_op D monday) as b0. destruct b0. apply tuesday.
   remember (eq_op D tuesday) as b1. destruct b1. apply wednesday.
   remember (eq_op D wednesday) as b2. destruct b2. apply thursday.
   remember (eq_op D thursday) as b3. destruct b3. apply friday.
   remember (eq_op D friday) as b4. destruct b4. apply saturday.
   remember (eq_op D saturday) as b5. destruct b5. apply sunday.
   apply monday.
  Defined.

Program Definition ByteEquiv_F (c:ch):ordinal_finType 256 :=
   @Ordinal 256 (Z.to_nat (proj1_sig c)) _.
Next Obligation.
destruct (@leP (S (Z.to_nat c)) 256); trivial; lia.
Qed.

Program Definition ByteEquiv_f (o:ordinal_finType 256):ch :=
  match o with @Ordinal _ m i => exist _ (Z.of_nat m) _ end.
Next Obligation. destruct (@leP (S m) 256). lia. exfalso. intuition. Qed.

Program Definition Byte_equiv: Equiv ch (ordinal_finType 256%nat) :=
  Build_Equiv _ _ ByteEquiv_F ByteEquiv_f _ _.
Next Obligation.
destruct x; simpl. apply Ordinal_congr. simpl. rewrite Nat2Z.id; trivial. Qed.
Next Obligation.
apply exist_congr. rewrite Z2Nat.id; trivial. Qed.

Program Definition WD_Equiv:= Build_EquivDFA _ _ Byte_equiv WeekdayDFA weekday_dfa
  WDState_Equiv _ _ _ .
Next Obligation. apply Ordinal_congr; simpl; trivial. Qed.
Next Obligation. destruct s as [s Hs]; simpl.
  specialize (@eqnP (Z.to_nat s) 6%nat); intros X.
  rewrite in_set1; simpl. unfold eq_op. simpl. unfold eq_op. simpl.
  split; intros. apply Z.eqb_eq in H. subst s; simpl; trivial.
  destruct X. apply Z.eqb_eq. lia. exfalso. intuition.
Qed.
Next Obligation. destruct x; simpl; trivial. destruct s as [s Hs]; simpl.
  remember(Nat.eqb (Pos.to_nat p) 110) as b; destruct b.
{ assert (P: Pos.eqb 110 p = true).
  { apply beq_nat_eq in Heqb. specialize (Pos2Nat.id p). rewrite Heqb. simpl; intros.
    apply Pos.eqb_eq; auto. }
  rewrite P. clear P Heqb. 
  remember (Z.eqb s 0) as b0; destruct b0; simpl.
  { symmetry in Heqb0; apply Z.eqb_eq in Heqb0; subst s; apply Ordinal_congr; trivial. }
  remember (Z.eqb s 1) as b1; destruct b1; simpl.
  { symmetry in Heqb1; apply Z.eqb_eq in Heqb1; subst s; apply Ordinal_congr; trivial. }
  remember (Z.eqb s 2) as b2; destruct b2; simpl.
  { symmetry in Heqb2; apply Z.eqb_eq in Heqb2; subst s; apply Ordinal_congr; trivial. }
  remember (Z.eqb s 3) as b3; destruct b3; simpl.
  { symmetry in Heqb3; apply Z.eqb_eq in Heqb3; subst s; apply Ordinal_congr; trivial. }
  remember (Z.eqb s 4) as b4; destruct b4; simpl.
  { symmetry in Heqb4; apply Z.eqb_eq in Heqb4; subst s; apply Ordinal_congr; trivial. }
  remember (Z.eqb s 5) as b5; destruct b5; simpl.
  { symmetry in Heqb5; apply Z.eqb_eq in Heqb5; subst s; apply Ordinal_congr; trivial. }
  remember (Z.eqb s 6) as b6; destruct b6; simpl.
  { symmetry in Heqb6; apply Z.eqb_eq in Heqb6; subst s; apply Ordinal_congr; trivial. }
  lia. }
{ assert (P: Pos.eqb 110 p = false).
  { remember (Pos.eqb 110 p) as t; destruct t; trivial.
    symmetry in Heqt; apply Peqb_true_eq in Heqt. subst p; inversion Heqb. }
  rewrite P; trivial. }
Qed.
