Require Import VST.floyd.proofauto.
Require Export ASI_surelyMalloc.

Definition GPred (T:Type):= (T * globals * val) -> mpred.

Fixpoint funcptrs {cs} T (t:type) (meths:list ident) (specs: list (GPred T -> funspec))(fptrs:list val)(m:val) (P:GPred T): mpred :=
  match specs, meths, fptrs with 
    nil, nil, nil => emp
  | phi::specs', meth::meths', u::fptrs' =>
          ((EX sh v, !!(JMeq v u /\ readable_share sh) &&
            func_ptr' (phi P) u * @field_at cs sh t [StructField meth] v m)
          * funcptrs T t meths' specs' fptrs' m P)%logic
  | _,_,_ => FF
  end.

Notation MTable := funcptrs.

Definition defaultMethspec {T} (P:GPred T): funspec := ImpossibleFunspec.

Section MTable.
Variable cs:compspecs.
Variable T:Type.
Variable t:type.
Variable idents:list ident.
Variable specs: list (GPred T -> funspec).
Variable fptrs:list val.
Variable m:val.
Variable P:GPred T.

Lemma funcptrs_duplicable:
      @funcptrs cs T t idents specs fptrs m P |--
      @funcptrs cs T t idents specs fptrs m P * @funcptrs cs T t idents specs fptrs m P.
Proof. generalize dependent fptrs. generalize dependent specs. clear. induction idents; simpl; intros.
+ destruct specs; destruct fptrs; entailer!.
+ destruct specs. entailer!.
  destruct fptrs. entailer!.
  Intros sh w.
  Exists (fst (slice.cleave sh)).
  Exists (snd (slice.cleave sh)).
  Exists w w.
  sep_apply IHl; clear IHl.
  rewrite split_func_ptr' at 1. entailer!.
  - split. apply slice.cleave_readable1; auto.
    apply slice.cleave_readable2; auto.
  - rewrite (field_at_share_join (fst (slice.cleave sh)) (snd (slice.cleave sh)) sh). trivial.
    apply slice.cleave_join.
Qed.

Lemma funcptrs_length:
      @funcptrs cs T t idents specs fptrs m P |-- !!(length idents = length specs /\ length idents = length fptrs).
Proof. generalize dependent specs. generalize dependent fptrs. clear.
  induction idents; intros; destruct fptrs; destruct specs; simpl; entailer!.
Qed.
Lemma funcptrs_Zlength:
      @funcptrs cs T t idents specs fptrs m P |-- !!(Zlength idents = Zlength specs /\ Zlength idents = Zlength fptrs).
Proof. sep_apply funcptrs_length. rewrite ! Zlength_correct. entailer!. Qed.

Lemma funcptrs_consD {ide spec fptr}:
      @funcptrs cs T t (ide::idents) (spec::specs) (fptr::fptrs) m P =
      (@funcptrs cs T t [ide] [spec] [fptr] m P * @funcptrs cs T t idents specs fptrs m P)%logic.
Proof. simpl; apply pred_ext; Intros sh v; Exists sh v; entailer!. Qed.
 
Lemma funcptrs_app_inv: forall {idents2 specs2 fptrs2 },
      length idents = length specs -> length idents = length fptrs ->
      @funcptrs cs T t (idents++idents2) (specs++specs2) (fptrs++fptrs2) m P =
      (@funcptrs cs T t idents specs fptrs m P * @funcptrs cs T t idents2 specs2 fptrs2 m P)%logic.
Proof. 
  generalize dependent specs. generalize dependent fptrs. clear.
  induction idents; intros.
+ destruct fptrs; destruct specs; simpl in *; try lia. rewrite emp_sepcon; trivial.
+ destruct fptrs; destruct specs; simpl in *; try lia.
  assert (Hs1: Datatypes.length l = Datatypes.length l1) by lia. clear H.
  assert (Hf1: Datatypes.length l = Datatypes.length l0) by lia. clear H0.
  rewrite IHl by trivial; clear IHl. rewrite sepcon_assoc; trivial. 
Qed.
Lemma funcptrs_app_invZ: forall {idents2 specs2 fptrs2 },
      Zlength idents = Zlength specs -> Zlength idents = Zlength fptrs ->
      @funcptrs cs T t (idents++idents2) (specs++specs2) (fptrs++fptrs2) m P =
      (@funcptrs cs T t idents specs fptrs m P * @funcptrs cs T t idents2 specs2 fptrs2 m P)%logic.
Proof. intros. apply funcptrs_app_inv; rewrite ! Zlength_correct in *; lia. Qed.
End MTable.


Lemma extractEntryFromMT_Z {cs} T DEF i idents (Hi: 0 <= i < Zlength idents) (P:GPred T) t specs fptrs m:
  MTable T t idents specs fptrs m P |-- 
      !!(Zlength idents = Zlength specs /\ Zlength idents = Zlength fptrs) &&
      (MTable T t (firstn (Z.to_nat i) idents)
                      (firstn (Z.to_nat i) specs) (firstn (Z.to_nat i) fptrs) m P *
      MTable T t (skipn (Z.to_nat (i + 1)) idents)
                      (skipn (Z.to_nat (i + 1)) specs) (skipn (Z.to_nat (i + 1)) fptrs) m P) *
      EX sh v, !!(JMeq v (*(@Znth _ Inhabitant_val i fptrs)*) (Znth i fptrs) /\ readable_share sh) &&
                 func_ptr' (*((@Znth _ defaultMethspec i specs)*)((@Znth _ DEF i specs) P) (Znth i fptrs) *
                 @field_at cs sh t [StructField (Znth i idents)] v m.
Proof.
  assert_PROP (Zlength idents = Zlength specs /\ Zlength idents = Zlength fptrs)
    by apply funcptrs_Zlength.
  destruct H as [Ls Lf].
  rewrite (split3_full_length_list 0 i (Zlength idents) idents) at 1 by lia.
  rewrite (@split3_full_length_list _ (*defaultMethspec*)DEF 0 i (Zlength specs) specs) at 1 by lia.
  rewrite (split3_full_length_list 0 i (Zlength fptrs) fptrs) at 1 by lia.
  rewrite ! Zminus_0_r.
  rewrite funcptrs_app_invZ by ( rewrite 2 Zlength_firstn; lia). simpl app.
  rewrite funcptrs_consD. cancel.
  simpl. Intros sh u. Exists sh u(* (Znth i fptrs)*). entailer!.
Qed.

Lemma insertEntryInMT_Z {cs:compspecs} T DEF i idents (Hi: 0 <= i < Zlength idents) (P:GPred T) t specs (fptrs:list val) m
                      (sh:share) v (VU:JMeq v (Znth i fptrs)) (Rsh: readable_share sh)
      (LS: Zlength idents = Zlength specs) (LF: Zlength idents = Zlength fptrs) phi
      (Phi: phi = @Znth _ (*defaultMethspec*)DEF i specs P):
      (MTable T t (firstn (Z.to_nat i) idents)
                      (firstn (Z.to_nat i) specs) (firstn (Z.to_nat i) fptrs) m P *
      MTable T t (skipn (Z.to_nat (i + 1)) idents)
                      (skipn (Z.to_nat (i + 1)) specs) (skipn (Z.to_nat (i + 1)) fptrs) m P) *
      func_ptr' phi (Znth i fptrs) *
      @field_at cs sh t [StructField (Znth i idents)] v m
 |-- MTable T t idents specs fptrs m P.
Proof. subst phi.
  rewrite (split3_full_length_list 0 i (Zlength idents) idents) at 4 by lia.
  rewrite (@split3_full_length_list _ (*defaultMethspec*)DEF 0 i (Zlength specs) specs) at 4 by lia.
  rewrite (split3_full_length_list 0 i (Zlength fptrs) fptrs) at 4 by lia.
  rewrite ! Zminus_0_r.
  rewrite funcptrs_app_invZ by ( rewrite 2 Zlength_firstn; lia). simpl app.
  rewrite funcptrs_consD. cancel.
  simpl. Exists sh v. entailer!.
Qed.

Lemma extractEntryFromMT_sl {cs} T DEF i idents (Hi: 0 <= i < Zlength idents) (P:GPred T) t specs fptrs m:
  MTable T t idents specs fptrs m P |-- 
      !!(Zlength idents = Zlength specs /\ Zlength idents = Zlength fptrs) &&
      (MTable T t (sublist 0 i idents) (sublist 0 i specs) (sublist 0 i fptrs) m P *
      MTable T t (sublist (i + 1) (Zlength idents) idents)
                      (sublist (i + 1) (Zlength specs) specs) (sublist (i + 1) (Zlength fptrs) fptrs) m P) *
      EX sh v, !!(JMeq v (*(@Znth _ Inhabitant_val i fptrs)*) (Znth i fptrs) /\ readable_share sh) &&
                 func_ptr' (*((@Znth _ defaultMethspec i specs)*)((@Znth _ DEF i specs) P) (Znth i fptrs) *
                 @field_at cs sh t [StructField (Znth i idents)] v m.
Proof.
  sep_apply (extractEntryFromMT_Z T DEF i idents). entailer!.
  rewrite ! sublist_skip by lia; trivial.
Qed.

Lemma insertEntryInMT_sl {cs idents} T DEF i (Hi: 0 <= i < Zlength idents)
                         t (fptrs:list val)(sh:share) v (VU:JMeq v (Znth i fptrs)) m phi
                         specs (Rsh: readable_share sh)(P:GPred T)
      (LS: Zlength idents = Zlength specs) (LF: Zlength idents = Zlength fptrs)
      (Phi: phi = @Znth _ (*defaultMethspec*)DEF i specs P):
      (MTable T t (sublist 0 i idents) (sublist 0 i specs) (sublist 0 i fptrs) m P *
      MTable T t (sublist (i + 1) (Zlength idents) idents)
                      (sublist (i + 1) (Zlength specs) specs) (sublist (i + 1) (Zlength fptrs) fptrs) m P) *
      func_ptr' phi (Znth i fptrs) *
      @field_at cs sh t [StructField (Znth i idents)] v m
 |-- MTable T t idents specs fptrs m P.
Proof. rewrite ! sublist_skip by lia. rewrite ! sublist_firstn. eapply insertEntryInMT_Z; trivial. Qed.

Lemma insertEntryInMT_sl_N {cs idents} T DEF i (Hi: 0 <= i < Zlength idents)
                         t (fptrs:list val)(sh:share) v (VU:JMeq v (nth (Z.to_nat i) fptrs Inhabitant_val)) m phi
                         specs (Rsh: readable_share sh)(P:GPred T)
      (LS: Zlength idents = Zlength specs) (LF: Zlength idents = Zlength fptrs) ii (Hii: ii= Z.to_nat i)
      (Phi: phi = @nth _ (*(Z.to_nat i)*)ii specs DEF P):
      (MTable T t (sublist 0 i idents) (sublist 0 i specs) (sublist 0 i fptrs) m P *
      MTable T t (sublist (i + 1) (Zlength idents) idents)
                      (sublist (i + 1) (Zlength specs) specs) (sublist (i + 1) (Zlength fptrs) fptrs) m P) *
      func_ptr' phi (nth ii fptrs Inhabitant_val) *
      @field_at cs sh t [StructField (Znth i idents)] v m
 |-- MTable T t idents specs fptrs m P.
Proof. subst ii. rewrite ! sublist_skip by lia. rewrite ! sublist_firstn.
 assert (Q: JMeq v (Znth i fptrs)).
  { eapply JMeq_trans. apply VU. clear - Hi. unfold Znth. rewrite if_false. apply JMeq_refl. lia. }
 eapply derives_trans.
 2:{ apply (insertEntryInMT_sl T DEF i Hi t fptrs sh v Q m phi); trivial.
     unfold Znth. rewrite if_false; trivial; lia. }
 rewrite ! sublist_skip by lia. rewrite ! sublist_firstn. cancel.
 unfold Znth. rewrite if_false; trivial; lia.
Qed.

Lemma extractEntryFromMT {cs} T DEF (i:nat) idents (Hi: (0 <= i < length idents)%nat) (P:GPred T) t specs fptrs m:
  MTable T t idents specs fptrs m P |-- 
      !!(length idents = length specs /\ length idents = length fptrs) &&
      (MTable T t (firstn i idents)
                      (firstn i specs) (firstn i fptrs) m P *
      MTable T t (skipn (i + 1) idents)
                      (skipn (i + 1) specs) (skipn (i + 1) fptrs) m P) *
      EX sh v, !!(JMeq v (nth i fptrs Inhabitant_val)  /\ readable_share sh) &&
                 func_ptr' ((@nth _ i specs DEF ) P) (nth i fptrs Inhabitant_val) *
                 @field_at cs sh t [StructField (Znth (Z.of_nat i) idents (*Inhabitant_positive*))] v m.
Proof.
  sep_apply (extractEntryFromMT_Z T DEF (Z.of_nat i)).  rewrite Zlength_correct;lia.
  Intros sh v. Exists sh. entailer!. { clear - Hi H H0. rewrite ! Zlength_correct in *; lia. }
  clear H5 H4 H3. Exists v. rewrite ! Nat2Z.id. rewrite <- ! nth_Znth, Nat2Z.id in *.
  2: rewrite <- H0, Zlength_correct; rep_lia.
  2: rewrite <- H0, Zlength_correct; rep_lia.
  2: rewrite <- H, Zlength_correct; rep_lia.
  entailer!.
  apply derives_refl'; f_equal; f_equal; rep_lia.
Qed.

Lemma insertEntryInMT {cs idents} (*T DEF*) (i:nat) (Hi: (0 <= i < length idents)%nat) t (fptrs:list val) (sh:share) v (VU:JMeq v (nth i fptrs Inhabitant_val) )T DEF
                      (*(Hi: (0 <= i < length idents)%nat) *)(P:GPred T) specs (*(fptrs:list val)*) m
                      (*(sh:share)*) (*v (VU:JMeq v (nth i fptrs Inhabitant_val))*) (Rsh: readable_share sh)
      (LS: length idents = length specs) (LF: length idents = length fptrs) phi
      (Phi: phi = @nth _ i specs DEF P):
      (MTable T t (firstn i idents)
                      (firstn i specs) (firstn i fptrs) m P *
      MTable T t (skipn (i + 1) idents)
                      (skipn (i + 1) specs) (skipn (i + 1) fptrs) m P) *
      func_ptr' phi (nth i fptrs Inhabitant_val) *
      @field_at cs sh t [StructField (Znth (Z.of_nat i)  idents)] v m
 |-- MTable T t idents specs fptrs m P.
Proof.
  assert (Hii: 0 <= Z.of_nat i < Zlength idents ). { rewrite Zlength_correct; rep_lia. }
  assert (HV: JMeq v (Znth (Z.of_nat i) fptrs)).
  { rewrite <- ! nth_Znth, Nat2Z.id. apply VU.
    rewrite Zlength_correct; rep_lia. } (*
  assert (exists u : reptype (nested_field_type t (DOT Znth (Z.of_nat i) idents)),
          JMeq u (@Znth _ Inhabitant_val (Z.of_nat i) fptrs)).
  { exists v. rewrite <- ! nth_Znth, Nat2Z.id in *. trivial.
    rewrite Zlength_correct; rep_lia. }
  destruct H as [u U].*)
  assert (X: Zlength idents = Zlength specs). rewrite !Zlength_correct; rep_lia.
  assert (Y: Zlength idents = Zlength fptrs). rewrite !Zlength_correct; rep_lia.
  eapply derives_trans.
  2:{ apply (insertEntryInMT_Z T DEF (Z.of_nat i) idents Hii P t specs fptrs m sh v HV Rsh X Y phi).
      rewrite <- ! nth_Znth, Nat2Z.id. trivial. 
      rewrite Zlength_correct; rep_lia. }
  rewrite Nat2Z.id. cancel.
  apply sepcon_derives.
  + apply derives_refl'; f_equal; f_equal; rep_lia.
  + apply derives_refl'; f_equal. rewrite <- ! nth_Znth, Nat2Z.id; trivial.
    rewrite Zlength_correct; rep_lia.
Qed. 

Lemma funcptrs_change_compspecs cs1 cs2 (CCE : change_composite_env cs1 cs2) t
      (CST: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE) t = true) u T P: forall specs idents fptrs
      (Hidents: forall i, In i idents -> @reptype cs1 (@nested_field_type cs1 t [StructField i]) = @reptype cs2 (@nested_field_type cs2 t [StructField i])),
      @funcptrs cs1 T t idents specs fptrs u P = @funcptrs cs2 T t idents specs fptrs u P.
Proof. 
  induction specs; intros; simpl funcptrs. destruct idents; simpl; trivial. 
  destruct idents. simpl; trivial. destruct fptrs; simpl; trivial.
  rewrite IHspecs. 2:{ intros. apply Hidents. right; trivial. } f_equal.
  apply exp_congr. intros sh.
  assert (R: @reptype cs1 (@nested_field_type cs1 t [StructField i]) =
             @reptype cs2 (@nested_field_type cs2 t [StructField i])). apply Hidents; left; trivial.
  clear - R CST.
  apply pred_ext.
  + Intros v1. destruct (@JMeq_sigT _ _ v1 R) as [v2 V2]. 
    Exists v2. entailer!.
    - eapply JMeq_trans. apply JMeq_sym. apply V2. apply H. 
    - apply change_compspecs_field_at_cancel; trivial.
  + Intros v2. symmetry in R. destruct (@JMeq_sigT _ _ v2 R) as [v1 V1]. 
    Exists v1. apply JMeq_sym in V1. entailer!; clear H1.
    - eapply JMeq_trans. apply V1. apply H. 
    - erewrite (@field_at_change_composite _ _ CCE). apply derives_refl. apply V1. apply CST.
Qed.

Definition FldAt {cs:compspecs} T (o:type) (mt:ident) (m:val):GPred T := fun x=>
           EX m', !! (JMeq m m') && field_at Ews o [StructField mt] m' (snd x).

Lemma FldAt_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) T mt base
      (CST_Base: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) base = true)
      (Hbase : @reptype cs1 (@nested_field_type cs1 base (DOT mt)) =
               @reptype cs2 (@nested_field_type cs2 base (DOT mt))):
      @FldAt cs1 T base mt = @FldAt cs2 T base mt.
Proof. unfold FldAt. extensionality m. extensionality x. apply pred_ext.
  + Intros m1 (*sh*). specialize (@JMeq_sigT _ _ m1 Hbase) as [m2 M2]. 
    Exists m2 (*sh*). entailer!; clear H1.
    - eapply JMeq_trans. apply H. apply M2.
    - apply change_compspecs_field_at_cancel; trivial.
  + Intros m2 (*sh*). symmetry in Hbase. destruct (@JMeq_sigT _ _ m2 Hbase) as [m1 M1]. 
    Exists m1 (*sh*). entailer!; clear H1.
    - eapply JMeq_trans. apply H. apply M1.
    - erewrite (@field_at_change_composite _ _ CCE12). apply derives_refl. apply JMeq_sym; apply M1. apply CST_Base.
Qed.

Lemma FldAt_change_compspecs' cs1 cs2 (CCE12 : change_composite_env cs1 cs2) T mt base x P
      (CST_Base: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) base = true)
      (Hbase : @reptype cs1 (@nested_field_type cs1 base (DOT mt)) =
               @reptype cs2 (@nested_field_type cs2 base (DOT mt))):
      @FldAt cs1 T base mt x P = @FldAt cs2 T base mt x P.
Proof. rewrite (@FldAt_change_compspecs _ _ CCE12 T mt base); trivial. Qed.

Definition RdblFldAt {cs:compspecs} T (o:type) (mt:ident) (m:val): GPred T := fun x=>
           EX m' sh, !! (JMeq m m' /\ readable_share sh /\ sepalg.join_sub sh (fst (slice.cleave Ews))) && field_at sh o [StructField mt] m' (snd x).

Lemma RdblFldAt_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) T mt base
      (CST_Base: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) base = true)
      (Hbase : @reptype cs1 (@nested_field_type cs1 base (DOT mt)) =
               @reptype cs2 (@nested_field_type cs2 base (DOT mt))):
      @RdblFldAt cs1 T base mt = @RdblFldAt cs2 T base mt.
Proof. unfold RdblFldAt. extensionality m. extensionality x. apply pred_ext.
  + Intros m1 sh. specialize (@JMeq_sigT _ _ m1 Hbase) as [m2 M2]. 
    Exists m2 sh. entailer!; clear H1.
    - eapply JMeq_trans. apply H. apply M2.
    - apply change_compspecs_field_at_cancel; trivial.
  + Intros m2 sh. symmetry in Hbase. destruct (@JMeq_sigT _ _ m2 Hbase) as [m1 M1]. 
    Exists m1 sh. entailer!; clear H1.
    - eapply JMeq_trans. apply H. apply M1.
    - erewrite (@field_at_change_composite _ _ CCE12). apply derives_refl. apply JMeq_sym; apply M1. apply CST_Base.
Qed.

Lemma RdblFldAt_change_compspecs' cs1 cs2 (CCE12 : change_composite_env cs1 cs2) T mt base x P
      (CST_Base: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) base = true)
      (Hbase : @reptype cs1 (@nested_field_type cs1 base (DOT mt)) =
               @reptype cs2 (@nested_field_type cs2 base (DOT mt))):
      @RdblFldAt cs1 T base mt x P = @RdblFldAt cs2 T base mt x P.
Proof. rewrite (@RdblFldAt_change_compspecs _ _ CCE12 T mt base); trivial. Qed.

Section ObjRep.
Variable cs:compspecs.
Variable T:Type.
Variable P:GPred T.
Variable M :MallocFreeAPD.
Variable base:type.
Variable mt:ident.
Variable dfsops:type.
Variable methidents:list ident.
Variable specs : list (GPred T -> funspec).
(*Variable fptrs : list val.*)

Definition dynobj_rep:GPred T:= fun x => 
  EX fptrs cso, (*EX P:GPred T,*) (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
  EX m, !!(isptr m) && FldAt T base mt m x *
        @MTable cs T (*M*) dfsops methidents specs fptrs m P)%logic.

Definition Rdbldynobj_rep:GPred T:= fun x => 
  EX fptrs cso, (*EX P:GPred T,*) (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
  EX m, !!(isptr m) && RdblFldAt T base mt m x *
        @MTable cs T (*M*) dfsops methidents specs fptrs m P)%logic.
End ObjRep.

Section GObjRep.
Variable cs:compspecs.
Variable T:Type.
Variable Q:GPred T.
Variable dfsops:type.
Variable methidents:list ident.
Variable specs : list (GPred T -> funspec).

Definition staticobj_rep:GPred T:= fun x => EX fptrs, (Q x * @MTable cs _ (*M*) dfsops methidents specs fptrs (snd x) Q)%logic.

End GObjRep.

Lemma obj_rep_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) T P M base dfsops
      (CST_MT: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) dfsops = true)
      (CST_Base: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) base = true) specs idents
      (HMT: forall i, In i idents ->
            @reptype cs1 (@nested_field_type cs1 dfsops [StructField i]) = @reptype cs2 (@nested_field_type cs2 dfsops [StructField i]))
      mt (Hbase: @reptype cs1 (@nested_field_type cs1 base [StructField mt]) = @reptype cs2 (@nested_field_type cs2 base [StructField mt])):
      @dynobj_rep cs1 T P M base mt dfsops idents specs = @dynobj_rep cs2 T P M base mt dfsops idents specs.
Proof.
  unfold dynobj_rep. extensionality x.
  apply exp_congr; intros fptrs.
  apply exp_congr; intros cso.
  f_equal.
  apply exp_congr; intros m.
  rewrite (@funcptrs_change_compspecs _ _ CCE12) by trivial.
  rewrite (FldAt_change_compspecs _ _ CCE12); trivial.
Qed.

Lemma staticobj_rep_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) T Q dfsops
      (CST_MT: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) dfsops = true) specs idents
      (HMT: forall i, In i idents ->
            @reptype cs1 (@nested_field_type cs1 dfsops [StructField i]) = @reptype cs2 (@nested_field_type cs2 dfsops [StructField i])):
       @staticobj_rep cs1 T Q dfsops idents specs = @staticobj_rep cs2 T Q dfsops idents specs.
Proof.
  unfold staticobj_rep. extensionality x. apply exp_congr; intros fptrs.
  rewrite (@funcptrs_change_compspecs _ _ CCE12); trivial.
Qed.

Lemma funcptrs_approx cs T u m P n: forall idents specs fptrs
      (SpecsApprox: 
        Forall (fun phi => forall n v,
                 compcert_rmaps.R.approx n (func_ptr (phi P) v) =
                 compcert_rmaps.R.approx n (func_ptr (phi (fun x : T * globals * val => compcert_rmaps.R.approx n (P x))) v)) specs),
      compcert_rmaps.R.approx n (@funcptrs cs T u idents specs fptrs m P) =
      compcert_rmaps.R.approx n (@funcptrs cs T u idents specs fptrs m (fun x => compcert_rmaps.R.approx n (P x))).
Proof. induction idents; intros.
+ destruct specs; simpl; trivial.
+ destruct specs. simpl; trivial.
  destruct fptrs. simpl; trivial.
  inv SpecsApprox. exploit (IHidents specs fptrs); clear IHidents; [ apply H2 | intros IHidents].
  rewrite 2 (funcptrs_consD _ T u idents). rewrite !approx_sepcon. rewrite <- IHidents; clear IHidents. 
  f_equal. specialize (H1 n v). simpl. rewrite 2 sepcon_emp.
  rewrite ! approx_exp. apply exp_congr. intros vv.
  rewrite ! approx_exp. apply exp_congr. intros w.
  rewrite ! approx_sepcon. rewrite !approx_andp. f_equal. f_equal.
  unfold func_ptr'. rewrite !approx_andp. f_equal. apply H1.
Qed.