Require Import VST.veric.rmaps.
Require Import VST.veric.compcert_rmaps. (*For TypeTree etc in invokeInit spec etc*)

Lemma approx_Sn_eq_weaken:
  forall n a b, approx (S n) a = approx (S n) b -> approx n a = approx n b.
Proof.
intros.
apply predicates_hered.pred_ext.
-
intros ? ?.
destruct H0.
split; auto.
assert (approx (S n) b a0).
rewrite <- H.
split; auto.
apply H2.
-
intros ? ?.
destruct H0.
split; auto.
assert (approx (S n) a a0).
rewrite H.
split; auto.
apply H2.
Qed.

Lemma approx_idem : forall n P, compcert_rmaps.R.approx n (compcert_rmaps.R.approx n P) =
  compcert_rmaps.R.approx n P.
Proof.
  intros.
  transitivity (base.compose (compcert_rmaps.R.approx n) (compcert_rmaps.R.approx n) P); auto.
  rewrite compcert_rmaps.RML.approx_oo_approx; auto.
Qed.

Lemma approx_idem' : forall n P, approx n (approx n P) =
  approx n P.
Proof. intros. apply approx_idem. Qed.

Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Export ASI_surelyMalloc.
Require Import ObjectBase.
Require Import ASI_DFA.
Require Import ASI_SingleDFA.
Require Import ASI_StaticDFA.
Require Import ASI_Client.
Require Import WeekdayDFA.
Require Import TrueFalseDFA.
Require Import ASI_TrueFalse.
Require Import ASI_Weekdays.
Require Import ASI_ComplementDFA.
Require Import ASI_ProductDFA.
Require Import ASI_IntersectionDFA.
Require Import ASI_StaticWeekdays.
Require Import ASI_StaticTrueFalse.
Require Import ASI_EDFA. 
Require Import ASI_T_Weekdays. 
Require Import ASI_Twice_Weekdays. 

Lemma sepcon_congr (A A' B B':mpred): A=A' -> B=B' -> (A * B = A' * B')%logic.
Proof. intros; subst. trivial. Qed.

Lemma andp_congr P Q P' Q': P=P' -> Q=Q' -> P && Q = P' && Q'.
Proof. intros; subst; auto. Qed.

Lemma funcptr_f_equal fs fs' v v': fs=fs' -> v=v' -> func_ptr fs v = func_ptr fs' v'.
Proof. intros; subst; trivial. Qed.

Lemma init_spec_nonexpansive object {Sigma} {A:@Dfa Sigma} P n v:
approx n (func_ptr (init_spec object Sigma A P) v) =
approx n (func_ptr (init_spec object Sigma A (fun x => approx n (P x))) v).
Proof. unfold func_ptr, init_spec.
      apply approx_Sn_eq_weaken. 
      rewrite (@approx_func_ptr ((A * globals * val)%type)) at 1.
      setoid_rewrite approx_func_ptr at 2. apply f_equal. apply funcptr_f_equal; trivial.
      (* apply approx_func_ptr_si_general'.*)
      apply semax_prog.funspec_eq.
      * extensionality tss w rho'. destruct w as [[a gv] p].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
      * extensionality tss w rho'. destruct w as [[a gv] p].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
Qed.
Lemma step_spec_nonexpansive object {Sigma CHrep} {A:@Dfa Sigma} P n v:
approx n (func_ptr (step_spec object Sigma CHrep A P) v) =
approx n (func_ptr (step_spec object Sigma CHrep A (fun x => approx n (P x))) v).
Proof. unfold func_ptr, step_spec.
      apply approx_Sn_eq_weaken. 
      rewrite (@approx_func_ptr ((A * globals * val * val * Sigma)%type)) at 1.
      setoid_rewrite approx_func_ptr at 2. apply f_equal. apply funcptr_f_equal; trivial.
      (* apply approx_func_ptr_si_general'.*)
      apply semax_prog.funspec_eq.
      * extensionality tss w rho'. destruct w as [[[[a gv] p] c] d].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
      * extensionality tss w rho'. destruct w as [[[[a gv] p] c] d].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
Qed.
Lemma final_spec_nonexpansive object {Sigma } {A:@Dfa Sigma} P n v:
approx n (func_ptr (final_spec object Sigma A P) v) =
approx n (func_ptr (final_spec object Sigma A (fun x => approx n (P x))) v).
Proof. unfold func_ptr, final_spec.
      apply approx_Sn_eq_weaken. 
      rewrite (@approx_func_ptr ((A * globals * val)%type)) at 1.
      setoid_rewrite approx_func_ptr at 2. apply f_equal. apply funcptr_f_equal; trivial.
      (* apply approx_func_ptr_si_general'.*)
      apply semax_prog.funspec_eq.
      * extensionality tss w rho'. destruct w as [[a gv] p].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
      * extensionality tss w rho'. destruct w as [[a gv] p].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
Qed.
Lemma twosteps_spec_nonexpansive object {Sigma CHrep } {A:@EDfa Sigma} P n v:
approx n (func_ptr (twosteps_spec object Sigma CHrep A P) v) =
approx n (func_ptr (twosteps_spec object Sigma CHrep A (fun x => approx n (P x))) v).
Proof. unfold func_ptr, twosteps_spec.
      apply approx_Sn_eq_weaken. 
      rewrite (@approx_func_ptr ((A * globals * val * val * val *
        Sigma * Sigma)%type)) at 1.
      setoid_rewrite approx_func_ptr at 2. apply f_equal. apply funcptr_f_equal; trivial.
      (* apply approx_func_ptr_si_general'.*)
      apply semax_prog.funspec_eq.
      * extensionality tss w rho'. destruct w as [[[[[[a gv] p] c1] c2] d1] d2].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
      * extensionality tss w rho'. destruct w as [[[[[[a gv] p] c1] c2] d1] d2].
        unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
Qed.

Opaque funcptrs.
Lemma DFA_approx {cs:compspecs} M object dfaops ops i f s {Sigma} (CHrep: Sigma -> val -> Prop) (A:@Dfa Sigma) w GV INV n:
approx n (DFA Sigma CHrep A {| DFA_GV := GV; DFA_Inv := INV |} M object dfaops ops i f s cs w) = 
approx n
  (DFA Sigma CHrep A
     {| DFA_GV := fun a0 : globals => approx n (GV a0); DFA_Inv := fun a0 : A * globals * val => approx n (INV a0) |} M object
     dfaops ops i f s cs w).
Proof.
  rewrite 2 DFA_fold_unfold. simpl. rewrite ! approx_exp. apply exp_congr. intros fptrs.
  rewrite ! approx_exp. apply exp_congr. intros sco.
  rewrite !approx_sepcon. apply sepcon_congr. apply sepcon_congr; [ rewrite approx_idem |]; trivial.
  rewrite ! approx_exp. apply exp_congr. intros m.
  rewrite !approx_sepcon. apply sepcon_congr. trivial.
  apply funcptrs_approx.
  constructor. apply init_spec_nonexpansive.
  constructor. apply final_spec_nonexpansive.
  constructor. apply step_spec_nonexpansive.
  constructor.
Qed.

Lemma EDFA_approx {cs:compspecs} M object dfaops ops i f s t {Sigma} (CHrep: Sigma -> val -> Prop) (A:@EDfa Sigma) w GV INV n:
approx n (EDFA Sigma CHrep A {| EDFA_GV := GV; EDFA_Inv := INV |} M object dfaops ops i f s t cs w) = 
approx n
  (EDFA Sigma CHrep A
     {| EDFA_GV := fun a0 : globals => approx n (GV a0); EDFA_Inv := fun a0 : A * globals * val => approx n (INV a0) |} M object
     dfaops ops i f s t cs w).
Proof.
  rewrite 2 EDFA_fold_unfold. simpl. rewrite ! approx_exp. apply exp_congr. intros fptrs.
  rewrite ! approx_exp. apply exp_congr. intros sco.
  rewrite !approx_sepcon. apply sepcon_congr. apply sepcon_congr; [ rewrite approx_idem |]; trivial.
  rewrite ! approx_exp. apply exp_congr. intros m.
  rewrite !approx_sepcon. apply sepcon_congr. trivial.
  apply funcptrs_approx.
  constructor. apply init_spec_nonexpansive.
  constructor. apply final_spec_nonexpansive.
  constructor. apply step_spec_nonexpansive.
  constructor. apply twosteps_spec_nonexpansive.
  constructor.
Qed.

Require Import Client.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition clientQP:= ltac:(QPprog prog).
Definition clientVprog := QPvarspecs clientQP.

Notation object := (Tstruct _object noattr).
Notation dfaops := (Tstruct _dfaOps noattr).
Notation staticops := (Tstruct _staticOps noattr).
Notation tdfaops := (Tstruct _t_dfaOps noattr).

(*Need to repeat the following proofs from VSU_TwiceWeekdays for our local compspecs, or import that file and adjust compspecs,
  or investigate whether we can isolate them in some new file that has suitably general compspecs*)
  Lemma align_compatible_tdfaops_dfaops {b i} (AL: align_compatible tdfaops (Vptr b i)):
        align_compatible dfaops (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

  Lemma field_compatible_tdfaops_dfaops_init p: field_compatible tdfaops [StructField _init] p ->
        field_compatible dfaops [StructField _init] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tdfaops_dfaops; trivial.
    - split; simpl; auto. left; trivial.
  Qed.
  Lemma field_compatible_tdfaops_dfaops_final p: field_compatible tdfaops [StructField _final] p ->
        field_compatible dfaops [StructField _final] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tdfaops_dfaops; trivial.
    - split; simpl; auto. right; left; trivial.
  Qed.
  Lemma field_compatible_tdfaops_dfaops_step p: field_compatible tdfaops [StructField _step] p ->
        field_compatible dfaops [StructField _step] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tdfaops_dfaops; trivial.
    - split; simpl; auto. right; right; left; trivial.
  Qed.

Lemma DFA_tdfaops_dfaops M ch CHrep A APD x:
      DFA ch CHrep A APD M object tdfaops _ops _init _final _step CompSpecs x
  |-- DFA ch CHrep A APD M object dfaops _ops _init _final _step CompSpecs x.
Proof.
  rewrite 2 DFA_fold_unfold. Intros fptrs cso m. Exists fptrs cso m. cancel.
  entailer!.
Transparent funcptrs. simpl. Opaque funcptrs.
destruct fptrs; [trivial | Intros ish iv]. Exists ish iv. entailer!.
destruct fptrs; [ eapply derives_trans; [ apply sepcon_FF_derives'; trivial | apply FF_left] | Intros fsh fv; Exists fsh fv; entailer!].
destruct fptrs; [ eapply derives_trans; [ apply sepcon_FF_derives'; trivial | apply FF_left] | Intros ssh sv; Exists ssh sv; entailer!].
unfold field_at; simpl; entailer!. split3.
apply field_compatible_tdfaops_dfaops_init; trivial.
apply field_compatible_tdfaops_dfaops_step; trivial.
apply field_compatible_tdfaops_dfaops_final; trivial.
Qed.

  Definition Client_GV (gv:globals): mpred := data_at_ Ews (Tstruct _storage noattr) (gv _strg).

Section Client_VSU.
  Variable M:MallocFreeAPD.

  Variable TF_APD: DFA_APD bool TrueFalseDFA.
  Notation GV_TF:= (DFA_GV _ _  TF_APD).
  Notation mkTFspec := (makeTrueFalse_spec TF_APD M object dfaops _ops _init _final _step CompSpecs).

  Variable WD_APD: DFA_APD ch WeekdayDFA.
  Notation GV_WD:= (DFA_GV _ _  WD_APD).
  Notation mkWDspec := (makeWeekdays_spec WD_APD M object dfaops _ops _init _final _step CompSpecs).

  Variable Compl_APD: DFA_APD ch (ComplementDFA (*ch*) WeekdayDFA).
  Notation GV_Complem:= (DFA_GV _ _ Compl_APD).
  Notation mkComplemspec := (makeComplement_spec _ WeekdayDFA CHrepW WD_APD M object dfaops _ops _init _final _step CompSpecs Compl_APD).

  Notation Prod := (ProductDFA (*ch ch*) WeekdayDFA (ComplementDFA (*ch*) WeekdayDFA)).
  Variable Prod_APD: DFA_APD _ Prod.
  Notation Prod_GV:= (DFA_GV _ _ Prod_APD).

  Definition Inter := (IntersectionDFA (*ch ch*) WeekdayDFA Prod).
  Variable Inter_APD: DFA_APD _ Inter.
  Notation Inter_GV:= (DFA_GV _ _ Inter_APD).
(*
  Variable InterInherit_APD: DFA_APD _ Inter.
  Notation InterInherit_GV:= (DFA_GV _ _ InterInherit_APD).
*)
  Notation ProdCHrep:= (fun (c: ch * ch) (v:val) => CHrepW (fst c) v /\ CHrepW (snd c) v).

  Definition mkUnionspec := (makeProduct_spec _ _ _ _ CHrepW CHrepW WD_APD Compl_APD M object dfaops _ops _init _final _step CompSpecs Prod_APD).

  Notation InterCHrep:= (fun (c: ch * (ch * ch)) (v:val) => CHrepW (fst c) v /\ ProdCHrep (snd c) v).

  Definition mkInterspec := (makeIntersection_spec _ _ _ _ CHrepW ProdCHrep WD_APD Prod_APD M object dfaops _ops _init _final _step CompSpecs Inter_APD).
(*
  Definition mkInterInheritspec := (makeIntersection_spec _ _ _ _ CHrepW ProdCHrep WD_APD Prod_APD M object dfaops _ops _init _final _step CompSpecs InterInherit_APD).
*)
  Notation runClientComplementspec := (_runClientComplement,
                                       runClientComplement_spec M WD_APD object dfaops _ops _init _final _step CompSpecs Compl_APD).
(*
  Notation runClientEitherspec := (_runClientEither,
                                    runClientEither_spec M WD_APD object dfaops _ops _init _final _step CompSpecs Compl_APD Prod_APD Inter_APD  InterInherit_APD).
*)

  Notation runClientBothspec := (_runClientBoth,
                                 runClientBoth_spec M WD_APD object dfaops _ops _init _final _step CompSpecs Compl_APD Prod_APD Inter_APD (*InterInherit_APD*) Client_GV).

  Variable SW_APD : DFA_APD ch WeekdayDFA. 
  Notation makestaticWeekdaysspec := (makeStaticWeekdays_spec SW_APD staticops _st_init _st_final _st_step CompSpecs).

  Variable STF_APD : DFA_APD _ TrueFalseDFA.
  Notation makestaticTrueFalsespec := (makeStaticTrueFalse_spec STF_APD staticops _st_init _st_final _st_step CompSpecs).

  Variable SGL_APD: SingleDFA_APD _ TrueFalseDFA.

  Notation runStaticDFAsspec := (_runStaticDFAs, runStaticDFAs_spec SW_APD STF_APD SGL_APD Client_GV CompSpecs staticops _st_init _st_final _st_step).

  Notation SingleDFA_specs := (combine [_DFA_init; _DFA_final; _DFA_step]
                                 (map (fun phi => phi (SingleGV bool TrueFalseDFA SGL_APD)) (SingleDFA_ops_specs bool CHrep TrueFalseDFA))).

  Variable TW_APD : EDFA_APD TWeekdayDFA.
  Notation mkTWspec := (make_TWeekdays_spec TW_APD M object tdfaops _ops _init _final _step _twosteps CompSpecs).

  Variable TWC_APD : EDFA_APD TwiceWeekdayDFA.
  Notation mkTwiceWspec := (make_TwiceWeekdays_spec TWC_APD M object tdfaops _ops _init _final _step _twosteps CompSpecs).

  Notation runTClientspec := (_runTClient,
                               runTClient_spec M object _ops _init _final _step CompSpecs TW_APD tdfaops _twosteps).

  Notation runTwiceClientspec := (_runTwiceClient,
                               runTwiceClient_spec M object _ops _init _final _step CompSpecs TWC_APD tdfaops _twosteps).

  Notation createWDspec := (_createWeekdayDFA, createWeekdayDFA_spec M WD_APD object dfaops _ops _init _final _step CompSpecs TW_APD).

  Notation runClientspec := (_runClient,
                             runClient_spec M WD_APD TF_APD object dfaops _ops _init _final _step CompSpecs (*TWC_APD*)TW_APD Client_GV).

(******************* Definition of specs for the wrapper functions *********************)

(********************* invokeInit *********************)
Definition InvokeInitDFAWITH:Type := 
           sigT (fun Sigma => (Sigma -> val -> Prop) * @sigT (@Dfa Sigma) (fun A => (globals -> mpred) * GPred A * (Dfa_state A * globals * val)))%type.

Definition InvokeInitDFAWitness Sigma (CHrep: Sigma -> val -> Prop) (Parent:@Dfa Sigma)
           (w:(globals->mpred) * GPred Parent * (Parent * globals * val)): InvokeInitDFAWITH.
Proof. exists Sigma. split; trivial. exists Parent. apply w. Defined.

Definition InvokeInitDFATypetree:TypeTree.
apply (@SigType Type). intros Sigma.
apply (ProdType (ConstType (Sigma -> val -> Prop))).
apply (@SigType (@Dfa Sigma)). intros A.
eapply ProdType.
+ apply ProdType.
  - apply (ArrowType (ConstType globals) Mpred).
  - apply (ArrowType (ConstType (Dfa_state A * globals * val)) Mpred).
+ apply (ConstType (Dfa_state A * globals * val)).
Defined.

Definition InvokeInitDFAPre (ts:list Type) (x:InvokeInitDFAWITH): argsEnviron -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP () PARAMS(s) GLOBALS (gv)
       SEP (DFA Sigma CHrep A (Build_DFA_APD _ A GV INV) M object dfaops _ops _init _final _step CompSpecs (a, gv, s))).
Defined.

Definition InvokeInitDFAPost (ts:list Type) (x:InvokeInitDFAWITH): environ -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP ( ) LOCAL ()
       SEP (DFA Sigma CHrep A (Build_DFA_APD _ A GV INV) M object dfaops _ops _init _final _step CompSpecs (Dfa_s A, gv, s))).
Defined.

Lemma InvokeInitDFAPre_NExp: @args_super_non_expansive InvokeInitDFATypetree InvokeInitDFAPre.
Proof. clear. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeInitDFAPre; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LAMBDAx. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp. apply DFA_approx.
Qed.

Lemma InvokeInitDFAPost_NExp: @super_non_expansive InvokeInitDFATypetree InvokeInitDFAPost.
Proof. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeInitDFAPost; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold SEPx; simpl. rewrite !sepcon_emp. apply DFA_approx.
Qed.

Definition invokeInitDFA_spec :funspec :=
   mk_funspec (pair (cons (tptr object) nil) tvoid) cc_default _ _ _ InvokeInitDFAPre_NExp InvokeInitDFAPost_NExp.

Definition InvokeInitEDFAWITH:Type := 
           sigT (fun Sigma => (Sigma -> val -> Prop) * @sigT (@EDfa Sigma) (fun A => (globals -> mpred) * GPred A * (Dfa_state A * globals * val)))%type.

Definition InvokeInitEDFAWitness Sigma (CHrep: Sigma -> val -> Prop) (Parent:@EDfa Sigma)
           (w:(globals->mpred) * GPred Parent * (Parent * globals * val)): InvokeInitEDFAWITH.
Proof. exists Sigma. split; trivial. exists Parent. apply w. Defined.

Definition InvokeInitEDFATypetree:TypeTree.
apply (@SigType Type). intros Sigma.
apply (ProdType (ConstType (Sigma -> val -> Prop))).
apply (@SigType (@EDfa Sigma)). intros A.
eapply ProdType.
+ apply ProdType.
  - apply (ArrowType (ConstType globals) Mpred).
  - apply (ArrowType (ConstType (Dfa_state A * globals * val)) Mpred).
+ apply (ConstType (Dfa_state A * globals * val)).
Defined.

Definition InvokeInitEDFAPre (ts:list Type) (x:InvokeInitEDFAWITH): argsEnviron -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP () PARAMS(s) GLOBALS (gv)
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (a, gv, s))).
Defined.

Definition InvokeInitEDFAPost (ts:list Type) (x:InvokeInitEDFAWITH): environ -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP ( ) LOCAL ()
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (Dfa_s A, gv, s))).
Defined.

Lemma InvokeInitEDFAPre_NExp: @args_super_non_expansive InvokeInitEDFATypetree InvokeInitEDFAPre.
Proof. clear. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeInitEDFAPre; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LAMBDAx. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Lemma InvokeInitEDFAPost_NExp: @super_non_expansive InvokeInitEDFATypetree InvokeInitEDFAPost.
Proof. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeInitEDFAPost; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Definition invokeInitEDFA_spec :funspec :=
   mk_funspec (pair (cons (tptr object) nil) tvoid) cc_default _ _ _ InvokeInitEDFAPre_NExp InvokeInitEDFAPost_NExp.

Definition invokeInit_spec := binary_intersection' invokeInitDFA_spec invokeInitEDFA_spec (eq_refl _) (eq_refl _).

Lemma invokeInitDFA: funspec_sub invokeInit_spec invokeInitDFA_spec.
Proof. apply (binary_intersection'_sub' invokeInitDFA_spec invokeInitEDFA_spec (eq_refl _) (eq_refl _) invokeInit_spec (eq_refl _) ).
Qed. 
Lemma invokeInitEDFA: funspec_sub invokeInit_spec invokeInitEDFA_spec.
Proof. apply (binary_intersection'_sub' invokeInitDFA_spec invokeInitEDFA_spec (eq_refl _) (eq_refl _) invokeInit_spec (eq_refl _) ).
Qed.

(******************* invokeFinal_spec***************)

Definition InvokeFinalDFAWITH:Type := 
           sigT (fun Sigma => (Sigma -> val -> Prop) * @sigT (@Dfa Sigma) (fun A => (globals -> mpred) * GPred A * (Dfa_state A * globals * val)))%type.

Definition InvokeFinalDFAWitness Sigma (CHrep: Sigma -> val -> Prop) (Parent:@Dfa Sigma)
           (w:(globals->mpred) * GPred Parent * (Parent * globals * val)): InvokeFinalDFAWITH.
Proof. exists Sigma. split; trivial. exists Parent. apply w. Defined.

Definition InvokeFinalDFATypetree:TypeTree.
apply (@SigType Type). intros Sigma.
apply (ProdType (ConstType (Sigma -> val -> Prop))).
apply (@SigType (@Dfa Sigma)). intros A.
eapply ProdType.
+ apply ProdType.
  - apply (ArrowType (ConstType globals) Mpred).
  - apply (ArrowType (ConstType (Dfa_state A * globals * val)) Mpred).
+ apply (ConstType (Dfa_state A * globals * val)).
Defined.

Definition InvokeFinalDFAPre (ts:list Type) (x:InvokeFinalDFAWITH): argsEnviron -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP () PARAMS(s) GLOBALS (gv)
       SEP (DFA Sigma CHrep A (Build_DFA_APD _ A GV INV) M object dfaops _ops _init _final _step CompSpecs (a, gv, s))).
Defined.

Definition InvokeFinalDFAPost (ts:list Type) (x:InvokeFinalDFAWITH): environ -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP ( )  RETURN (if Dfa_fin A a then Vtrue else Vfalse)
       SEP (DFA Sigma CHrep A (Build_DFA_APD _ A GV INV) M object dfaops _ops _init _final _step CompSpecs (a, gv, s))).
Defined.

Lemma InvokeFinalDFAPre_NExp: @args_super_non_expansive InvokeFinalDFATypetree InvokeFinalDFAPre.
Proof. clear. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeFinalDFAPre; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LAMBDAx. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp. apply DFA_approx.
Qed.

Lemma InvokeFinalDFAPost_NExp: @super_non_expansive InvokeFinalDFATypetree InvokeFinalDFAPost.
Proof. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeFinalDFAPost; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold SEPx; simpl. rewrite !sepcon_emp. apply DFA_approx.
Qed.

Definition invokeFinalDFA_spec :funspec :=
   mk_funspec (pair (cons (tptr object) nil) tbool) cc_default _ _ _ InvokeFinalDFAPre_NExp InvokeFinalDFAPost_NExp.

Definition InvokeFinalEDFAWITH:Type := 
           sigT (fun Sigma => (Sigma -> val -> Prop) * @sigT (@EDfa Sigma) (fun A => (globals -> mpred) * GPred A * (Dfa_state A * globals * val)))%type.

Definition InvokeFinalEDFAWitness Sigma (CHrep: Sigma -> val -> Prop) (Parent:@EDfa Sigma)
           (w:(globals->mpred) * GPred Parent * (Parent * globals * val)): InvokeFinalEDFAWITH.
Proof. exists Sigma. split; trivial. exists Parent. apply w. Defined.

Definition InvokeFinalEDFATypetree:TypeTree.
apply (@SigType Type). intros Sigma.
apply (ProdType (ConstType (Sigma -> val -> Prop))).
apply (@SigType (@EDfa Sigma)). intros A.
eapply ProdType.
+ apply ProdType.
  - apply (ArrowType (ConstType globals) Mpred).
  - apply (ArrowType (ConstType (Dfa_state A * globals * val)) Mpred).
+ apply (ConstType (Dfa_state A * globals * val)).
Defined.

Definition InvokeFinalEDFAPre (ts:list Type) (x:InvokeFinalEDFAWITH): argsEnviron -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP () PARAMS(s) GLOBALS (gv)
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (a, gv, s))).
Defined.

Definition InvokeFinalEDFAPost (ts:list Type) (x:InvokeFinalEDFAWITH): environ -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
apply (PROP ( )  RETURN (if Dfa_fin A a then Vtrue else Vfalse)
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (a, gv, s))).
Defined.

Lemma InvokeFinalEDFAPre_NExp: @args_super_non_expansive InvokeFinalEDFATypetree InvokeFinalEDFAPre.
Proof. clear. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeFinalEDFAPre; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LAMBDAx. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Lemma InvokeFinalEDFAPost_NExp: @super_non_expansive InvokeFinalEDFATypetree InvokeFinalEDFAPost.
Proof. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]]. unfold InvokeFinalEDFAPost; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Definition invokeFinalEDFA_spec :funspec :=
   mk_funspec (pair (cons (tptr object) nil) tbool) cc_default _ _ _ InvokeFinalEDFAPre_NExp InvokeFinalEDFAPost_NExp.

(*special cases, dead.
  Definition invokeFinalWD_spec : funspec :=
   WITH s:val, d:_, gv:globals
   PRE [ tptr object ]
     PROP () PARAMS(s) GLOBALS (gv)
     SEP (DFA _ CHrepW WeekdayDFA WD_APD M object dfaops _ops _init _final _step CompSpecs (d, gv, s))
   POST [ tbool ]
        PROP () RETURN (if Dfa_fin WeekdayDFA d then Vtrue else Vfalse)
        SEP (DFA _ CHrepW WeekdayDFA WD_APD M object dfaops _ops _init _final _step CompSpecs (d, gv, s)).

  Definition invokeFinalTF_spec : funspec :=
   WITH s:val, d:_, gv:globals
   PRE [ tptr object ]
     PROP () PARAMS(s) GLOBALS (gv)
     SEP (DFA _ CHrep TrueFalseDFA TF_APD M object dfaops _ops _init _final _step CompSpecs (d, gv, s))
   POST [ tbool ]
        PROP () RETURN (if Dfa_fin TrueFalseDFA d then Vtrue else Vfalse)
        SEP (DFA _ CHrep TrueFalseDFA TF_APD M object dfaops _ops _init _final _step CompSpecs (d, gv, s)).

  Definition invokeFinal_spec := binary_intersection' invokeFinalWD_spec invokeFinalTF_spec (eq_refl _) (eq_refl _).
*)
Definition invokeFinal_spec := binary_intersection' invokeFinalDFA_spec invokeFinalEDFA_spec (eq_refl _) (eq_refl _).

Lemma invokeFinalDFA: funspec_sub invokeFinal_spec invokeFinalDFA_spec.
Proof. apply (binary_intersection'_sub' invokeFinalDFA_spec invokeFinalEDFA_spec (eq_refl _) (eq_refl _) invokeFinal_spec (eq_refl _) ).
Qed. 
Lemma invokeFinalEDFA: funspec_sub invokeFinal_spec invokeFinalEDFA_spec.
Proof. apply (binary_intersection'_sub' invokeFinalDFA_spec invokeFinalEDFA_spec (eq_refl _) (eq_refl _) invokeFinal_spec (eq_refl _) ).
Qed.

(************************* Construction of invokeStep_spec *****************************************)

Definition InvokeStepDFAWITH:Type := 
           sigT (fun Sigma => (Sigma -> val -> Prop) * @sigT (@Dfa Sigma) (fun A => (globals -> mpred) * GPred A * 
                                (Dfa_state A * globals * val * val * Sigma)))%type.

Definition InvokeStepDFAWitness Sigma (CHrep: Sigma -> val -> Prop) (Parent:@Dfa Sigma)
           (w:(globals->mpred) * GPred Parent * (Parent * globals * val * val * Sigma)): InvokeStepDFAWITH.
Proof. exists Sigma. split; trivial. exists Parent. apply w. Defined.

Definition InvokeStepDFATypetree:TypeTree.
apply (@SigType Type). intros Sigma.
apply (ProdType (ConstType (Sigma -> val -> Prop))).
apply (@SigType (@Dfa Sigma)). intros A.
eapply ProdType.
+ apply ProdType.
  - apply (ArrowType (ConstType globals) Mpred).
  - apply (ArrowType (ConstType (Dfa_state A * globals * val)) Mpred).
+ apply (ConstType (Dfa_state A * globals * val * val * Sigma)).
Defined.

Definition InvokeStepDFAPre (ts:list Type) (x:InvokeStepDFAWITH): argsEnviron -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]].
apply (PROP (CHrep d c) PARAMS(v;c) GLOBALS (gv)
       SEP (DFA Sigma CHrep A (Build_DFA_APD _ A GV INV) M object dfaops _ops _init _final _step CompSpecs (a, gv, v))).
Defined.

Definition InvokeStepDFAPost (ts:list Type) (x:InvokeStepDFAWITH): environ -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]].
apply (PROP ( ) LOCAL ( )
       SEP (DFA Sigma CHrep A (Build_DFA_APD _ A GV INV) M object dfaops _ops _init _final _step CompSpecs (Dfa_trans A a d, gv, v))).
Defined.

Lemma InvokeStepDFAPre_NExp: @args_super_non_expansive InvokeStepDFATypetree InvokeStepDFAPre.
Proof. clear. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]]. unfold InvokeStepDFAPre; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LAMBDAx. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp. apply DFA_approx.
Qed.

Lemma InvokeStepDFAPost_NExp: @super_non_expansive InvokeStepDFATypetree InvokeStepDFAPost.
Proof. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]]. unfold InvokeStepDFAPost; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold SEPx; simpl. rewrite !sepcon_emp. apply DFA_approx.
Qed.

Definition invokeStepDFA_spec :funspec :=
   mk_funspec (pair (cons (tptr object) (cons tuchar nil)) tvoid) cc_default _ _ _ InvokeStepDFAPre_NExp InvokeStepDFAPost_NExp.

Definition InvokeStepEDFAWITH:Type := 
           sigT (fun Sigma => (Sigma -> val -> Prop) * @sigT (@EDfa Sigma) (fun A => (globals -> mpred) * GPred A * 
                                (Dfa_state A * globals * val * val * Sigma)))%type.

Definition InvokeStepEDFAWitness Sigma (CHrep: Sigma -> val -> Prop) (Parent:@EDfa Sigma)
           (w:(globals->mpred) * GPred Parent * (Parent * globals * val * val * Sigma)): InvokeStepEDFAWITH.
Proof. exists Sigma. split; trivial. exists Parent. apply w. Defined.

Definition InvokeStepEDFATypetree:TypeTree.
apply (@SigType Type). intros Sigma.
apply (ProdType (ConstType (Sigma -> val -> Prop))).
apply (@SigType (@EDfa Sigma)). intros A.
eapply ProdType.
+ apply ProdType.
  - apply (ArrowType (ConstType globals) Mpred).
  - apply (ArrowType (ConstType (Dfa_state A * globals * val)) Mpred).
+ apply (ConstType (Dfa_state A * globals * val * val * Sigma)).
Defined.

Definition InvokeStepEDFAPre (ts:list Type) (x:InvokeStepEDFAWITH): argsEnviron -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]].
apply (PROP (CHrep d c) PARAMS(v;c) GLOBALS (gv)
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (a, gv, v))).
Defined.

Definition InvokeStepEDFAPost (ts:list Type) (x:InvokeStepEDFAWITH): environ -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]].
apply (PROP ( ) LOCAL ( )
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (Dfa_trans A a d, gv, v))).
Defined.

Lemma InvokeStepEDFAPre_NExp: @args_super_non_expansive InvokeStepEDFATypetree InvokeStepEDFAPre.
Proof. clear. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]]. unfold InvokeStepEDFAPre; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LAMBDAx. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Lemma InvokeStepEDFAPost_NExp: @super_non_expansive InvokeStepEDFATypetree InvokeStepEDFAPost.
Proof. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] v] c] d]]]]]. unfold InvokeStepEDFAPost; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Definition invokeStepEDFA_spec :funspec :=
   mk_funspec (pair (cons (tptr object) (cons tuchar nil)) tvoid) cc_default _ _ _ InvokeStepEDFAPre_NExp InvokeStepEDFAPost_NExp.

Definition invokeStep_spec := binary_intersection' invokeStepDFA_spec invokeStepEDFA_spec (eq_refl _) (eq_refl _).

Lemma invokeStepDFA: funspec_sub invokeStep_spec invokeStepDFA_spec.
Proof. apply (binary_intersection'_sub' invokeStepDFA_spec invokeStepEDFA_spec (eq_refl _) (eq_refl _) invokeStep_spec (eq_refl _) ).
Qed. 
Lemma invokeStepEDFA: funspec_sub invokeStep_spec invokeStepEDFA_spec.
Proof. apply (binary_intersection'_sub' invokeStepDFA_spec invokeStepEDFA_spec (eq_refl _) (eq_refl _) invokeStep_spec (eq_refl _) ).
Qed.

(************************* Construction of invokeTwostepsEDFA_spec *****************************************)

Definition InvokeTwostepsEDFAWITH:Type := 
           sigT (fun Sigma => (Sigma -> val -> Prop) * @sigT (@EDfa Sigma) (fun A => (globals -> mpred) * GPred A * 
                                (Dfa_state A * globals * val * val * val * Sigma * Sigma)))%type.

Definition InvokeTwostepsEDFAWitness Sigma (CHrep: Sigma -> val -> Prop) (Parent:@EDfa Sigma)
           (w:(globals->mpred) * GPred Parent * (Parent * globals * val * val * val * Sigma * Sigma)): InvokeTwostepsEDFAWITH.
Proof. exists Sigma. split; trivial. exists Parent. apply w. Defined.

Definition InvokeTwostepsEDFATypetree:TypeTree.
apply (@SigType Type). intros Sigma.
apply (ProdType (ConstType (Sigma -> val -> Prop))).
apply (@SigType (@EDfa Sigma)). intros A.
eapply ProdType.
+ apply ProdType.
  - apply (ArrowType (ConstType globals) Mpred).
  - apply (ArrowType (ConstType (Dfa_state A * globals * val)) Mpred).
+ apply (ConstType (Dfa_state A * globals * val * val * val * Sigma * Sigma)).
Defined.

Definition InvokeTwostepsEDFAPre (ts:list Type) (x:InvokeTwostepsEDFAWITH): argsEnviron -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[[[[[a gv] v] c1] c2] d1] d2]]]]].
apply (PROP (CHrep d1 c1; CHrep d2 c2) PARAMS(v;c1;c2) GLOBALS (gv)
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (a, gv, v))).
Defined.

Definition InvokeTwostepsEDFAPost (ts:list Type) (x:InvokeTwostepsEDFAWITH): environ -> mpred. 
destruct x as [Sigma [CHrep [A [[GV INV] [[[[[[a gv] v] c1] c2] d1] d2]]]]].
apply (PROP ( ) LOCAL ( )
       SEP (EDFA Sigma CHrep A (Build_EDFA_APD _ A GV INV) M object tdfaops _ops _init _final _step _twosteps CompSpecs (Dfa_t A a d1 d2, gv, v))).
Defined.

Lemma InvokeTwostepsEDFAPre_NExp: @args_super_non_expansive InvokeTwostepsEDFATypetree InvokeTwostepsEDFAPre.
Proof. clear. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[[[[[a gv] v] c1] c2] d1] d2]]]]]. unfold InvokeTwostepsEDFAPre; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LAMBDAx. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Lemma InvokeTwostepsEDFAPost_NExp: @super_non_expansive InvokeTwostepsEDFATypetree InvokeTwostepsEDFAPost.
Proof. repeat intro. simpl in x. simpl.
  destruct x as [Sigma [CHrep [A [[GV INV] [[[[[[a gv] v] c1] c2] d1] d2]]]]]. unfold InvokeTwostepsEDFAPost; simpl.
  unfold PROPx. simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold LOCALx; simpl. rewrite !approx_andp. apply andp_congr. trivial. 
  unfold SEPx; simpl. rewrite !sepcon_emp. apply EDFA_approx.
Qed.

Definition invokeTwostepsEDFA_spec :funspec :=
   mk_funspec (pair (cons (tptr object) (cons tuchar (cons tuchar nil))) tvoid) cc_default _ _ _ InvokeTwostepsEDFAPre_NExp InvokeTwostepsEDFAPost_NExp.

  Definition Client_internalspecs := [createWDspec; runClientspec; runClientComplementspec; (*runClientEitherspec;*)
                                      runClientBothspec; runStaticDFAsspec; runTClientspec; runTwiceClientspec;
                                      (_invokeInit, invokeInit_spec);
                                      (_invokeFinal, invokeFinal_spec);
                                      (_invokeStep, invokeStep_spec);
                                      (_invokeTwosteps, invokeTwostepsEDFA_spec)].

  Definition Client_importedspecs := [(_make_TrueFalse, mkTFspec); (_make_Weekdays, mkWDspec);
                                      (_mkCompl, mkComplemspec);
                                      (_mkProduct, mkUnionspec);
                                      (_make_IntersectionDFA, mkInterspec);
                                      (*(_make_IntersectionInheritDFA, mkInterInheritspec);*)
                                      (_make_StaticWeekdays,makestaticWeekdaysspec);
                                      (_make_StaticTrueFalse,makestaticTrueFalsespec);
                                      (_make_TWeekdays, mkTWspec);
                                      (_make_TwiceWeekdays, mkTwiceWspec)] ++ SingleDFA_specs.

  Definition Client_Gprog : funspecs := Client_internalspecs ++ Client_importedspecs.

Opaque funcptrs.
(*for case-by case specs
Lemma verif_invokeFinalWD: semax_body clientVprog Client_Gprog f_invokeFinal (_invokeFinal, invokeFinalWD_spec ).
  Proof.
    start_function.
    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W1: 0 <= 1 < Zlength [_init; _final; _step]) by list_solve.
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 1).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward. forward. forward_call. forward. 
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 1 W1 dfaops fptrs sh _  (JMeq_refl _)).
    entailer!. 1: solve [ simple_if_tac; trivial].
    rewrite DFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.
Lemma verif_invokeFinalTF: semax_body clientVprog Client_Gprog f_invokeFinal (_invokeFinal, invokeFinalTF_spec ).
  Proof.
    start_function.
    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W1: 0 <= 1 < Zlength [_init; _final; _step]) by list_solve.
    sep_apply (extractEntryFromMT_sl TrueFalseDFA defaultMethspec 1).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward. forward. forward_call. forward. 
    sep_apply (insertEntryInMT_sl_N TrueFalseDFA defaultMethspec 1 W1 dfaops fptrs sh _  (JMeq_refl _)).
    entailer!. 1: solve [ simple_if_tac; trivial].
    rewrite DFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.

Lemma verif_invokeFinal: semax_body clientVprog Client_Gprog f_invokeFinal (_invokeFinal, invokeFinal_spec ).
  Proof. eapply semax_body_binaryintersection''. apply verif_invokeFinalWD. apply verif_invokeFinalTF. Qed. *)
Lemma verif_invokeFinalDFA: semax_body clientVprog Client_Gprog f_invokeFinal (_invokeFinal, invokeFinalDFA_spec).
  Proof. unfold invokeFinalDFA_spec, InvokeFinalDFAPre, InvokeFinalDFAPost.
    start_function1. simpl in x. destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
    start_function2. start_function3.
    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W1: 0 <= 1 < Zlength [_init; _final; _step]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 1).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward. forward. forward_call. forward.
    entailer!. destruct (Dfa_fin A a); trivial. 
    sep_apply (insertEntryInMT_sl_N A defaultMethspec 1 W1 dfaops fptrs sh _  (JMeq_refl _)).
    rewrite DFA_fold_unfold. entailer!. simpl. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.
Lemma verif_invokeFinalEDFA: semax_body clientVprog Client_Gprog f_invokeFinal (_invokeFinal, invokeFinalEDFA_spec).
  Proof. unfold invokeFinalEDFA_spec, InvokeFinalEDFAPre, InvokeFinalEDFAPost.
    start_function1. simpl in x. destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
    start_function2. start_function3.
    rewrite EDFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W1: 0 <= 1 < Zlength [_init; _final; _step; _twosteps]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 1).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward.
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _final) m)
         && field_at sh dfaops (DOT _final) (nth 1 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_final in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCinit.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _final) (nth 1 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call. forward.
    entailer!. destruct (Dfa_fin A a); trivial. 
    sep_apply (insertEntryInMT_sl_N A defaultMethspec 1 W1 tdfaops fptrs sh _  (JMeq_refl _)).
    rewrite EDFA_fold_unfold. entailer!. simpl. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.
Lemma verif_invokeFinal: semax_body clientVprog Client_Gprog f_invokeFinal (_invokeFinal, invokeFinal_spec ).
  Proof. eapply semax_body_binaryintersection''. apply verif_invokeFinalDFA. apply verif_invokeFinalEDFA. Qed.

(*
Lemma verif_invokeInit ch CH (A:@Dfa ch) APD: semax_body clientVprog Client_Gprog f_invokeInit (_invokeInit, invokeInit_spec  ch CH A APD).
  Proof.
    start_function. 
    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W0: 0 <= 0 < Zlength [_init; _final; _step]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 0).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward. forward. forward_call.
    sep_apply (insertEntryInMT_sl_N A defaultMethspec 0 W0 dfaops fptrs sh _  (JMeq_refl _)).
    rewrite DFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.*)
Lemma verif_invokeInitDFA: semax_body clientVprog Client_Gprog f_invokeInit (_invokeInit, invokeInitDFA_spec).
  Proof. unfold invokeInitDFA_spec, InvokeInitDFAPre, InvokeInitDFAPost.
    start_function1. simpl in x. destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
    start_function2. start_function3.
    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W0: 0 <= 0 < Zlength [_init; _final; _step]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 0).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward. forward. forward_call.
    sep_apply (insertEntryInMT_sl_N A defaultMethspec 0 W0 dfaops fptrs sh _  (JMeq_refl _)).
    rewrite DFA_fold_unfold. entailer!. simpl. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.
Lemma verif_invokeInitEDFA: semax_body clientVprog Client_Gprog f_invokeInit (_invokeInit, invokeInitEDFA_spec).
  Proof. unfold invokeInitEDFA_spec, InvokeInitEDFAPre, InvokeInitEDFAPost.
    start_function1. simpl in x. destruct x as [Sigma [CHrep [A [[GV INV] [[a gv] s]]]]].
    start_function2. start_function3.
    rewrite EDFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W0: 0 <= 0 < Zlength [_init; _final; _step; _twosteps]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 0).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward.
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _init) m)
         && field_at sh dfaops (DOT _init) (nth 0 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_init in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCfinal.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _init) (nth 0 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call.
    sep_apply (insertEntryInMT_sl_N A defaultMethspec 0 W0 tdfaops fptrs sh _  (JMeq_refl _)).
    rewrite EDFA_fold_unfold. entailer!. simpl. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.
Lemma verif_invokeInit: semax_body clientVprog Client_Gprog f_invokeInit (_invokeInit, invokeInit_spec ).
  Proof. eapply semax_body_binaryintersection''. apply verif_invokeInitDFA. apply verif_invokeInitEDFA. Qed.

Lemma verif_invokeStepDFA: semax_body clientVprog Client_Gprog f_invokeStep (_invokeStep, invokeStepDFA_spec).
  Proof. unfold invokeStepDFA_spec, InvokeStepDFAPre, InvokeStepDFAPost.
    start_function1. simpl in x. destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] u] c] d]]]]].
    start_function2. start_function3. rename H into Hcd.
    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W2: 0 <= 2 < Zlength [_init; _final; _step]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 2).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward. forward. forward_call (a, gv, u, c, d).

    sep_apply (insertEntryInMT_sl_N A defaultMethspec 2 W2 dfaops fptrs sh _  (JMeq_refl _)).
    rewrite DFA_fold_unfold. entailer!. simpl. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.
Lemma verif_invokeStepEDFA: semax_body clientVprog Client_Gprog f_invokeStep (_invokeStep, invokeStepEDFA_spec).
  Proof. unfold invokeStepEDFA_spec, InvokeStepEDFAPre, InvokeStepEDFAPost.
    start_function1. simpl in x. destruct x as [Sigma [CHrep [A [[GV INV] [[[[a gv] u] c] d]]]]].
    start_function2. start_function3. rename H into Hcd.
    rewrite EDFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W2: 0 <= 2 < Zlength [_init; _final; _step; _twosteps]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 2).
    Intros sh v; apply JMeq_eq in H1; subst v; unfold Znth; simpl.
    forward.
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _step) m)
         && field_at sh dfaops (DOT _step) (nth 2 fptrs Inhabitant_val) m).
    { entailer!. clear - H3. apply field_compatible_tdfaops_dfaops_step in H3.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCstep.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _step) (nth 2 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call (a, gv, u, c, d).

    sep_apply (insertEntryInMT_sl_N A defaultMethspec 2 W2 tdfaops fptrs sh _  (JMeq_refl _)).
    rewrite EDFA_fold_unfold. entailer!. simpl. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!.
  Qed.
Lemma verif_invokeStep: semax_body clientVprog Client_Gprog f_invokeStep (_invokeStep, invokeStep_spec ).
  Proof. eapply semax_body_binaryintersection''. apply verif_invokeStepDFA. apply verif_invokeStepEDFA. Qed.

Lemma verif_invokeTwostepsEDFA: semax_body clientVprog Client_Gprog  f_invokeTwosteps (_invokeTwosteps, invokeTwostepsEDFA_spec).
  Proof. unfold invokeTwostepsEDFA_spec, InvokeTwostepsEDFAPre, InvokeTwostepsEDFAPost.
    start_function1. simpl in x. destruct x as [Sigma [CHrep [A [[GV INV] [[[[[[a gv] u] c1] c2] d1] d2]]]]].
    start_function2. start_function3. rename H into Hcd1. rename H0 into Hcd2.
    rewrite EDFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W3: 0 <= 3 < Zlength [_init; _final; _step; _twosteps]) by list_solve.
    sep_apply (extractEntryFromMT_sl A defaultMethspec 3).
    Intros sh v; apply JMeq_eq in H1. subst v; unfold Znth; simpl.
    forward. forward. forward_call (a, gv, u, c1, c2, d1, d2). forward. 
    sep_apply (insertEntryInMT_sl_N A defaultMethspec 3 W3 tdfaops fptrs sh _  (JMeq_refl _)).
    entailer!.
    rewrite EDFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!. rewrite sepcon_comm; apply derives_refl.
  Qed.

  Definition CharX: {n:Z | 0<= n < 256}. exists 120. lia. Defined.
  Definition CharN: {n:Z | 0<= n < 256}. exists 110. lia. Defined.

Lemma verif_createWeekdayDFA: semax_body clientVprog Client_Gprog f_createWeekdayDFA createWDspec.
  Proof. unfold createWeekdayDFA_spec.
    start_function. forward_if.
  + forward_call. Intros p. forward. Exists p. entailer!. sep_apply EDFA2DFA. 
    sep_apply DFA_tdfaops_dfaops. simpl.
    Exists (EApd2DApd TW_APD); trivial.
  + forward_call. Intros p. forward. Exists p. entailer!.
    Exists WD_APD. apply derives_refl.
Qed.

  Lemma verif_runTClient: semax_body clientVprog Client_Gprog f_runTClient runTClientspec.
  Proof. unfold runTClient_spec.
    start_function. rename H into HWW.
    forward_call.
    Intros w.
    rewrite EDFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W0: 0 <= 0 < Zlength [_init; _final; _step; _twosteps]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 0).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward. 
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _init) m)
         && field_at sh dfaops (DOT _init) (nth 0 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_init in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCinit.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _init) (nth 0 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call.
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 0 W0 tdfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - FCinit W0  Pm (*Msh Hmsh*) HWW.
(*
    assert (W2: 0 <= 2 < Zlength [_init; _final; _step; _twosteps]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _step) m)
         && field_at sh dfaops (DOT _step) (nth 2 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_step in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCstep.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _step) (nth 2 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call (Monday, gv, w, Vint (Int.repr 120), CharX). (*step X*)
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 2 W2 tdfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - FCinit FCstep W2 W0 Pm Msh Hmsh HWW.*)

    assert (W3: 0 <= 3 < Zlength [_init; _final; _step; _twosteps]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 3).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call (Monday, gv, w, Vint (Int.repr 120), Vint (Int.repr 110), CharX, CharN). (*step X; step N*)
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 3 W3 tdfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear -FCinit Pm (*Msh Hmsh*) HWW.

(*Invocation of invokeTwosteps*)
    gather_SEP 0 1 4 5.
    replace_SEP 0 (EDFA ch ASI_StaticWeekdays.CHrep TWeekdayDFA TW_APD M object tdfaops _ops _init _final _step _twosteps CompSpecs (Dfa_t TWeekdayDFA Monday CharX CharN, gv, w)).
    { rewrite EDFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!. }

    forward_call (@InvokeTwostepsEDFAWitness ch ASI_StaticWeekdays.CHrep TWeekdayDFA
          ((@EDFA_GV _ TWeekdayDFA TW_APD, @EDFA_Inv _ TWeekdayDFA TW_APD), (Dfa_t TWeekdayDFA Monday CharX CharN, gv, w, Vint (Int.repr 120), Vint (Int.repr 110), CharX, CharN))).
    { rewrite ! sepcon_assoc. apply sepcon_derives; [ apply derives_refl | cancel]. }
    clear.
    rewrite EDFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).
    assert (W1: 0 <= 1 < Zlength [_init; _final; _step; _twosteps]) by list_solve.
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 1).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _final) m)
         && field_at sh dfaops (DOT _final) (nth 1 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_final in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCfinal.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _final) (nth 1 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call.
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 1 W1 tdfaops fptrs sh _  (JMeq_refl _)).
    forward. Exists w. entailer!.
    rewrite EDFA_fold_unfold. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!. apply derives_refl. 
  Qed.

  Lemma verif_runClient: semax_body clientVprog Client_Gprog f_runClient runClientspec.
  Proof. unfold runClient_spec.
    start_function. rename H into HWW.  

(*Object tf = make_TrueFalse ();*)
    forward_call.
    Intros tf.
(*
(*Object W = make_Weekdays ();*)
    forward_call.
    Intros w.
*)
(*Object W = createWeekdayDFA (1);*)
    forward_call (true, gv).
    Intros w APD.

(*  strg.object_ref = W;*)
    unfold Client_GV. forward.

    assert (W0: 0 <= 0 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
(* tf -> ops -> init (tf);*)
  rewrite (DFA_fold_unfold bool). Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    sep_apply (extractEntryFromMT_sl TrueFalseDFA defaultMethspec 0).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N TrueFalseDFA defaultMethspec 0 W0 dfaops fptrs sh _  (JMeq_refl _)).
    gather_SEP 0 1 7 8.
    replace_SEP 0 (DFA bool CHrep TrueFalseDFA TF_APD M object dfaops _ops _init _final _step CompSpecs (Dfa_s TrueFalseDFA, gv, tf)).
    { rewrite DFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!. }
    deadvars!. clear - (*Pm Msh Hmsh *) W0(* HWW*).

(* W -> ops -> init (W);*)
    rewrite (DFA_fold_unfold ch). Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 0); trivial.
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 0 W0 dfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pm (* Msh Hmsh HWW*).

    assert (W2: 0 <= 2 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }

(*  tf -> ops -> step (W, '0');*)
  rewrite (DFA_fold_unfold bool). Intros fptrsTF csoTF n.
        first [unfold FldAt; Intros n' | unfold RdblFldAt; Intros n' msh].
        apply JMeq_eq in H; subst n'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).
    sep_apply (extractEntryFromMT_sl TrueFalseDFA defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call (false, gv, tf, Vint (Int.repr 48), true). (*step 0*)
    { split; auto. }
    sep_apply (insertEntryInMT_sl_N TrueFalseDFA defaultMethspec 2 W2 dfaops fptrsTF sh _  (JMeq_refl _)).
    deadvars!. clear - Pn Pm (*Msh Hmsh*) W2 (*HWW*).

(*  W -> ops -> step (W, 'x');*)
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call (Monday, gv, w, Vint (Int.repr 120), CharX). (*step X*)
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 2 W2 dfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pn Pm (*Msh Hmsh*) W2 (*HWW*).

(*  /*W */ strg.object_ref -> ops -> step (W, 'n');*) forward.
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call (Monday, gv, w, Vint (Int.repr 110), CharN). (*step N*)
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 2 W2 dfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pn Pm (*Msh Hmsh HWW*). simpl.

(*Version if the NEW spec of invokeFinal is used*)
    gather_SEP 2 3 4 5.
    replace_SEP 0 (DFA bool CHrep TrueFalseDFA TF_APD M object dfaops _ops _init _final _step CompSpecs (true, gv, tf)).
    { rewrite DFA_fold_unfold. entailer!. 
      Exists fptrsTF csoTF n.
      first [ unfold FldAt; Exists n | unfold RdblFldAt; Exists n nsh].
      entailer!. }
    forward_call invokeFinalDFA (@InvokeFinalDFAWitness bool CHrep TrueFalseDFA
                      ((DFA_GV _ TrueFalseDFA TF_APD, DFA_Inv _ TrueFalseDFA TF_APD), (true, gv, tf))).
    { destruct TF_APD. rewrite ! sepcon_assoc. apply sepcon_derives; [ apply derives_refl | cancel]. }
    change (if Dfa_fin TrueFalseDFA true then Vtrue else Vfalse) with (Vtrue).
    forward_if (PROP ( ) LOCAL (temp _t'4 Vfalse; gvars gv)
      SEP (DFA bool CHrep TrueFalseDFA TF_APD M object dfaops _ops _init _final _step CompSpecs (true, gv, tf);
           DFA _ CHrepW WeekdayDFA APD M object dfaops _ops _init _final _step CompSpecs  (Tuesday, gv, w);
           mem_mgr M gv; GV_WD gv; EDFA_GV _ TW_APD gv;
           GV_TF gv;
           data_at Ews (Tstruct _storage noattr) (Vundef, w) (gv _strg))).
    { gather_SEP 1 2 6 7.
      replace_SEP 0 (DFA ch CHrepW WeekdayDFA (*WD_APD*)APD M object dfaops _ops _init _final _step CompSpecs  (Tuesday, gv, w)).
      { rewrite DFA_fold_unfold. entailer!. 
        Exists fptrs cso m.
        first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
        entailer!. }
      forward_call invokeFinalDFA (@InvokeFinalDFAWitness ch CHrepW WeekdayDFA
                      ((DFA_GV _ WeekdayDFA APD, DFA_Inv _ WeekdayDFA APD), (Tuesday, gv, w))).
      { destruct APD. rewrite ! sepcon_assoc. apply sepcon_derives; [ apply derives_refl | cancel]. }
      forward. entailer!. rewrite sepcon_comm; apply derives_refl. }
    { (*else*) inv H. }
    (*return instruction*)
    forward. Exists w tf APD. entailer!. unfold Client_GV. cancel.

(* (slightly outdated but repairable proof of final instructions if invokeFinal is used whhere the spec of invokefinal is
    the binaryintersection'' of apply verif_invokeFinalWD. apply verif_invokeFinalTF -- see above
    destruct (binary_intersection'_sub' invokeFinalWD_spec invokeFinalTF_spec (eq_refl _) (eq_refl _) invokeFinal_spec (eq_refl _) ) as [FinalWD FinalTF].

    gather_SEP 2 3 4 5.
    replace_SEP 0 (DFA bool CHrep TrueFalseDFA TF_APD M object dfaops _ops _init _final _step CompSpecs (true, gv, tf)).
    { rewrite DFA_fold_unfold. entailer!. 
      Exists fptrsTF csoTF n.
      first [ unfold FldAt; Exists n | unfold RdblFldAt; Exists n nsh].
      entailer!. }
    gather_SEP 1 2 6 7.
    replace_SEP 0 (DFA _ CHrepW WeekdayDFA (*WD_APD*)APD M object dfaops _ops _init _final _step CompSpecs  (Tuesday, gv, w)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!. }
    forward_call FinalTF (tf, true, gv).
    forward_if (PROP ( ) LOCAL (temp _t'4 Vfalse; gvars gv)
      SEP (DFA bool CHrep TrueFalseDFA TF_APD M object dfaops
               _ops _init _final _step CompSpecs (true, gv, tf);
           DFA _ CHrepW WeekdayDFA (*WD_APD*)APD M object dfaops _ops _init _final _step CompSpecs  (Tuesday, gv, w);

           mem_mgr M gv; GV_WD gv; EDFA_GV _ TW_APD gv;
           GV_TF gv;
           data_at Ews (Tstruct _storage noattr) (Vundef, w) (gv _strg))).
(*    forward_if (emp:environ ->mpred).*)
    { (*assert (APD=WD_APD) by admit.*) forward_call FinalWD (w, Tuesday, gv). (* rewrite H0.*) cancel. simpl.
       forward. rewrite Int.eq_true, H0. entailer!. }
    { forward. entailer!. }
    forward. Exists w tf APD. entailer!. unfold Client_GV. cancel.*)

(*Proof for the code that does not invoke invokeFinal twice but uses manual method dispatch:
    assert (W1: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
(*  tf -> ops -> final (tf);*)
    sep_apply (extractEntryFromMT_sl TrueFalseDFA defaultMethspec 1).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N TrueFalseDFA defaultMethspec 1 W1 dfaops fptrsTF sh _  (JMeq_refl _)).
    gather_SEP 0 1 4 5.
    replace_SEP 0 (DFA bool CHrep TrueFalseDFA TF_APD M object dfaops _ops _init _final _step CompSpecs (true, gv, tf)).
    { rewrite DFA_fold_unfold. entailer!. 
      Exists fptrsTF csoTF n.
      first [ unfold FldAt; Exists n | unfold RdblFldAt; Exists n nsh].
      entailer!. }
    simpl. deadvars!. clear - Pm (*Msh Hmsh*) W1 (*HWW*).

    forward_if (PROP ( ) LOCAL (temp _t'4 Vfalse; gvars gv)
      SEP (DFA bool CHrep TrueFalseDFA TF_APD M object dfaops
               _ops _init _final _step CompSpecs (true, gv, tf);
           DFA _ CHrepW WeekdayDFA (*WD_APD*)APD M object dfaops _ops _init _final _step CompSpecs  (Tuesday, gv, w);

           mem_mgr M gv; GV_WD gv; EDFA_GV _ TWC_APD gv;
           GV_TF gv;
           data_at Ews (Tstruct _storage noattr) (Vundef, w) (gv _strg))).
    + deadvars!. clear - Pm (*Msh Hmsh*) W1 (*HWW*).

  (*  W -> ops -> final (W);*)
      sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 1).
      Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
      forward.
      forward.
      forward_call.
      sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 1 W1 dfaops fptrs sh _  (JMeq_refl _)).
      forward. entailer!.
      { rewrite DFA_fold_unfold. entailer!. 
        Exists fptrs cso m.
        first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
        entailer!. }
    + inv H. 
    + clear. forward. Exists w tf APD. entailer!.
      unfold Client_GV; entailer!.*)
  Qed.

(* A future test case for when we have inheritance
Lemma verif_runClientEither: semax_body clientVprog Client_Gprog f_runClientEither runClientEitherspec.
  Proof. unfold runClientEither_spec.
    start_function. rename H into HWW.
(*_t'1 = _make_Weekdays([])*)
    forward_call. Intros w1.
(*_t'2 = _mkCompl(_t'*)
    forward_call. Intros c.
    forward.
(*_t'3 = _make_Weekdays([])*)
    forward_call. Intros w. try deadvars!; simpl.
(*_t'4 = _mkProduct([(_t'3)%expr; (_C)%expr])*)
    forward_call. Intros p.
    forward. try deadvars!; simpl.
    forward_if (EX i:val,
      PROP ( )
      LOCAL (temp _I i; gvars gv; temp _b (Val.of_bool b))
      SEP (mem_mgr M gv; GV_WD gv; GV_Complem gv; Prod_GV gv; Inter_GV gv; InterInherit_GV gv;
           (*DFA (ch * ch) ProdCHrep _ Prod_APD M object dfaops _ops _init _final _step CompSpecs (Monday, Monday,gv, p)*)
           if b 
           then DFA _ InterCHrep _ InterInherit_APD M object dfaops _ops _init _final _step CompSpecs (Monday,(Monday, Monday), gv, i)
           else DFA _ InterCHrep _ Inter_APD M object dfaops _ops _init _final _step CompSpecs (Monday, (Monday, Monday), gv, i))).
    + (*_t'5 = _make_Weekdays([])*)
        forward_call. Intros w2.
      (*_t'6 = _make_IntersectionInheritDFA ([(_t'5)%expr; (_P))*)
         forward_call. Intros i.
         forward.
      Exists i. entailer!. apply derives_refl.
    + (*_t'7 = _make_Weekdays([])*)
        forward_call. Intros w2.
      (*_t'8 = _make_IntersectionDFA ([(_t'7)%expr; (_P))*)
         forward_call. Intros i.
         forward.
      Exists i. entailer!. apply derives_refl.
    + Intros i.
      destruct b.
      - rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

        assert (W: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
        sep_apply (extractEntryFromMT_sl Inter defaultMethspec 1).
        Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
        forward.
        forward.
        forward_call.
        sep_apply (insertEntryInMT_sl_N Inter defaultMethspec 1 W dfaops fptrs sh _  (JMeq_refl _)).
        deadvars!. clear - (*Msh Hmsh*) Pm. Time simpl; forward; Exists i; apply andp_right; [ apply prop_right; trivial | cancel].
        rewrite DFA_fold_unfold. Exists fptrs cso m. simpl. cancel.
        first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh]; entailer!. apply derives_refl.
      - rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

        assert (W: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
        sep_apply (extractEntryFromMT_sl Inter defaultMethspec 1).
        Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
        forward.
        forward.
        forward_call.
        sep_apply (insertEntryInMT_sl_N Inter defaultMethspec 1 W dfaops fptrs sh _  (JMeq_refl _)).
        deadvars!. clear - (*Msh Hmsh*) Pm. Time simpl; forward; Exists i; apply andp_right; [ apply prop_right; trivial | cancel].
        rewrite DFA_fold_unfold. Exists fptrs cso m. simpl. cancel.
        first [unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh]; entailer!. apply derives_refl.
  Qed.*)

  Lemma verif_runClientBoth: semax_body clientVprog Client_Gprog f_runClientBoth runClientBothspec.
  Proof. unfold runClientBoth_spec.
    start_function. rename H into HWW.
    forward_call. Intros w.
    forward_call. Intros w1.
    forward_call. Intros c. 
    forward_call. Intros p. deadvars!. clear - HWW.

    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

        assert (W1: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
        sep_apply (extractEntryFromMT_sl Prod defaultMethspec 1).
        Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
        forward.
        forward.
        forward.
        forward_call.
        sep_apply (insertEntryInMT_sl_N Prod  defaultMethspec 1 W1 dfaops fptrs sh _  (JMeq_refl _)).
    simpl. deadvars!.
    gather_SEP 0 1 4 5.
    replace_SEP 0 (DFA (ch * ch)
     (fun (c : ch * ch) (v : val) =>
      v = Vint (Int.repr (proj1_sig (fst c))) /\
      v = Vint (Int.repr (proj1_sig (snd c)))) Prod Prod_APD M
     object dfaops _ops _init _final _step CompSpecs ((Monday, Monday), gv, p)).
    { rewrite DFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!; try apply derives_refl. }
    clear - W1 HWW.
    forward.
    forward_call. Intros w.
    forward_call. Intros i. simpl.
    unfold Client_GV. forward. deadvars!. clear - W1. simpl. (*
    forward. { entailer!.
    forward_call. Intros w.
    forward_call. Intros w1.
    forward_call. Intros c. 
    forward_call. Intros p. forward. deadvars!. clear - HWW.
    forward_call. Intros w.
    forward_call. Intros ii. forward. deadvars!. clear - HWW.

    freeze [0;1;3;4;5;6] ADMIN.
    rewrite (DFA_fold_unfold _ _ _ InterInherit_APD). Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

        assert (W: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
        sep_apply (extractEntryFromMT_sl Inter defaultMethspec 1).
        Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
        forward.
        forward.
        forward_call.
        sep_apply (insertEntryInMT_sl_N Inter defaultMethspec 1 W dfaops fptrs sh _  (JMeq_refl _)).
    simpl. deadvars!.
    gather_SEP 0 1 3 4.
    replace_SEP 0 (DFA _ InterCHrep _ InterInherit_APD M object dfaops _ops _init _final _step CompSpecs ((Monday, (Monday, Monday)), gv, ii)).
    { rewrite DFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!; try apply derives_refl. }
    clear - W.
    freeze [0;1] FR.*)

    try forward. (*_t'12 = _strg._object_ref;*)
    try solve [sep_apply DFA_isptr; entailer! ].
  
    rewrite (DFA_fold_unfold _ _ _ Inter_APD). Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

        sep_apply (extractEntryFromMT_sl Inter defaultMethspec 1).
        Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
        forward.
        forward.
        try forward. (*_t'15 = _strg._object_ref;*)
        forward_call.
        sep_apply (insertEntryInMT_sl_N Inter defaultMethspec 1 W1 dfaops fptrs sh _  (JMeq_refl _)).
    simpl. deadvars!.
    gather_SEP 0 1 4 5.
    replace_SEP 0 (DFA _ InterCHrep _ Inter_APD M object dfaops _ops _init _final _step CompSpecs ((Monday, (Monday, Monday)), gv, i)).
    { rewrite DFA_fold_unfold. entailer!. 
      Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!; apply derives_refl. }
    clear.

    forward_call (true, false). deadvars.
    (*Time*) (sep_apply func_ptr'_emp; freeze [0;1;3] FRX; forward; simpl;
         Exists i; apply andp_right; [ entailer! | thaw FRX; (*; thaw FR; thaw ADMIN;*) unfold Client_GV; cancel]). (*5 secs*)
  Qed.

  Lemma verif_runStaticDFAs: semax_body clientVprog Client_Gprog f_runStaticDFAs runStaticDFAsspec.
  Proof. unfold runStaticDFAs_spec.
    start_function.

(*DFA_init*)
    forward_call.

(*_make_StaticWeekdays*)
    forward_call. Intros w.

(*_make_StaticTrueFalse*)
    forward_call. Intros tf.


(*  strg.ops_ref = tf;*) 
    unfold Client_GV. forward. simpl.

    assert (W0: 0 <= 0 < Zlength [_st_init; _st_final; _st_step]). { unfold Zlength; simpl. rep_lia. }

  (*w -> st_init ();*)
    rewrite (SDFA_fold_unfold ch). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ Weekday defaultMethspec 0).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward. forward_call. simpl.
    sep_apply (insertEntryInMT_sl_N Weekday defaultMethspec 0 W0 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA ch ASI_StaticWeekdays.CHrep WeekdayDFA SW_APD staticops
                        _st_init _st_final _st_step CompSpecs (Monday, gv, w)).
    { rewrite (SDFA_fold_unfold ch). thaw FR; entailer!. Exists fptrs. cancel. }
    deadvars!. clear - W0.

  (*tf -> st_init ();*)
    rewrite (SDFA_fold_unfold bool). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ bool defaultMethspec 0).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward. forward_call. simpl.
    sep_apply (insertEntryInMT_sl_N bool defaultMethspec 0 W0 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA bool CHrepTF TrueFalseDFA STF_APD staticops
                        _st_init _st_final _st_step CompSpecs (false, gv, tf)).
    { rewrite (SDFA_fold_unfold bool). thaw FR; entailer!. Exists fptrs. cancel. }
    deadvars!. clear - W0.

  (*strg.ops_ref -> st_init ();*)
    forward.
    rewrite (SDFA_fold_unfold bool). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ bool defaultMethspec 0).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward. forward_call. simpl.
    sep_apply (insertEntryInMT_sl_N bool defaultMethspec 0 W0 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA bool CHrepTF TrueFalseDFA STF_APD staticops
                        _st_init _st_final _st_step CompSpecs (false, gv, tf)).
    { rewrite (SDFA_fold_unfold bool). thaw FR; entailer!. Exists fptrs. cancel. }
    deadvars!. clear.

    assert (W2: 0 <= 2 < Zlength [_st_init; _st_final; _st_step]). { unfold Zlength; simpl. rep_lia. }

  (*w -> st_step ('x');*)
    rewrite (SDFA_fold_unfold ch). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ Weekday defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward_call (Monday, w, gv, Vint (Int.repr 120), CharX). (*step X*) simpl.
    sep_apply (insertEntryInMT_sl_N Weekday defaultMethspec 2 W2 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA ch ASI_StaticWeekdays.CHrep WeekdayDFA SW_APD staticops
                        _st_init _st_final _st_step CompSpecs (Monday, gv, w)).
    { rewrite (SDFA_fold_unfold ch). thaw FR; entailer!. Exists fptrs. cancel. }
    deadvars!. clear - W2.

  (*tf -> st_step ('1');*)
    rewrite (SDFA_fold_unfold bool). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ bool defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward_call (false, tf, gv, Vint (Int.repr 49), false). (*step 1*) simpl.
    { clear; split; intros N; inv N. }
    sep_apply (insertEntryInMT_sl_N bool defaultMethspec 2 W2 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA bool CHrepTF TrueFalseDFA STF_APD staticops
                        _st_init _st_final _st_step CompSpecs (false, gv, tf)).
    { rewrite (SDFA_fold_unfold bool). thaw FR; entailer!. Exists fptrs. simpl; cancel. }
    deadvars!. clear - W2.

  (*_DFA_step('1')*)
   forward_call (false, gv, Vint (Int.repr 49), false).
    { clear; split; intros N; inv N. }

  (*w -> st_step ('n');*)
    rewrite (SDFA_fold_unfold ch). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ Weekday defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward_call (Monday, w, gv, Vint (Int.repr 110), CharN). (*step N*) simpl.
    sep_apply (insertEntryInMT_sl_N Weekday defaultMethspec 2 W2 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA ch ASI_StaticWeekdays.CHrep WeekdayDFA SW_APD staticops
                        _st_init _st_final _st_step CompSpecs (Tuesday, gv, w)).
    { rewrite (SDFA_fold_unfold ch). thaw FR; entailer!. Exists fptrs. cancel. }
    deadvars!. clear - W2.

  (*_DFA_step('0')*)
   forward_call (false, gv, Vint (Int.repr 48), true).
    { clear; split; trivial. }

  (*strg.ops_ref -> st_step ('0');*) forward.
    rewrite (SDFA_fold_unfold bool). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ bool defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward_call (false, tf, gv, Vint (Int.repr 48), true). (*step 0*) simpl.
    { clear; split; trivial. }
    sep_apply (insertEntryInMT_sl_N bool defaultMethspec 2 W2 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA bool CHrepTF TrueFalseDFA STF_APD staticops
                        _st_init _st_final _st_step CompSpecs (true, gv, tf)).
    { rewrite (SDFA_fold_unfold bool). thaw FR; entailer!. Exists fptrs. simpl; cancel. }
    deadvars!. clear.
  (*tf -> st_step ('0');
    rewrite (SDFA_fold_unfold bool). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ bool defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward_call (false, tf, gv, Vint (Int.repr 48), true). (*step 0*) simpl.
    { clear; split; trivial. }
    sep_apply (insertEntryInMT_sl_N bool defaultMethspec 2 W2 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA bool CHrepTF TrueFalseDFA STF_APD staticops
                        _st_init _st_final _st_step CompSpecs (true, gv, tf)).
    { rewrite (SDFA_fold_unfold bool). thaw FR; entailer!. Exists fptrs. simpl; cancel. }
    deadvars!. clear.*)

    assert (W1: 0 <= 1 < Zlength [_st_init; _st_final; _st_step]). { unfold Zlength; simpl. rep_lia. }

  (*w -> final();*)
    rewrite (SDFA_fold_unfold ch). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ Weekday defaultMethspec 1).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward_call. simpl.
    sep_apply (insertEntryInMT_sl_N Weekday defaultMethspec 1 W1 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA ch ASI_StaticWeekdays.CHrep WeekdayDFA SW_APD staticops
                        _st_init _st_final _st_step CompSpecs (Tuesday, gv, w)).
    { rewrite (SDFA_fold_unfold ch). thaw FR; entailer!. Exists fptrs. cancel. }
    deadvars!. clear - W1.

  (*tf -> final();*)
    rewrite (SDFA_fold_unfold bool). Intros fptrs. simpl.
    sep_apply (@extractEntryFromMT_sl _ bool defaultMethspec 1).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward_call. simpl.
    sep_apply (insertEntryInMT_sl_N bool defaultMethspec 1 W1 staticops fptrs sh _ (JMeq_refl _)).
    freeze [0;1] FR.
    replace_SEP 0 (SDFA bool CHrepTF TrueFalseDFA STF_APD staticops
                        _st_init _st_final _st_step CompSpecs (true, gv, tf)).
    { rewrite (SDFA_fold_unfold bool). thaw FR; entailer!. Exists fptrs. simpl; cancel. }
    deadvars!. clear.

  (*_DFA_final*)
   forward_call. simpl.

  (*invoke bop*)
    forward_call (true, true).

  (*invoke bop*)
    forward_call (false, bop true true). sep_apply func_ptr'_emp.

  (*return*)
    forward. Exists w tf. entailer!. unfold Client_GV; cancel.
  Qed.

  Lemma verif_runTwiceClient: semax_body clientVprog Client_Gprog f_runTwiceClient runTwiceClientspec.
  Proof. unfold runTwiceClient_spec.
    start_function. rename H into HWW.
    forward_call.
    Intros w.
    rewrite EDFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
        apply JMeq_eq in H; subst m'; simpl snd.
        try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W0: 0 <= 0 < Zlength [_init; _final; _step; _twosteps]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 0).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward. 
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _init) m)
         && field_at sh dfaops (DOT _init) (nth 0 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_init in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCinit.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _init) (nth 0 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call.
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 0 W0 tdfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - FCinit W0 Pm (*Msh Hmsh*) HWW.
(*
    assert (W2: 0 <= 2 < Zlength [_init; _final; _step; _twosteps]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _step) m)
         && field_at sh dfaops (DOT _step) (nth 2 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_step in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCstep.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _step) (nth 2 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call (Monday, gv, w, Vint (Int.repr 120), CharX). (*step X*)
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 2 W2 tdfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - FCinit FCstep W2 W0 Pm Msh Hmsh HWW.*)

    assert (W3: 0 <= 3 < Zlength [_init; _final; _step; _twosteps]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 3).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call (Monday, gv, w, Vint (Int.repr 120), Vint (Int.repr 110), CharX, CharN). (*step X; step N*)
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 3 W3 tdfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear -FCinit Pm (*Msh Hmsh*) HWW.

    assert (W1: 0 <= 1 < Zlength [_init; _final; _step; _twosteps]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 1).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    replace_SEP 3 (!!(field_compatible tdfaops (DOT _final) m)
         && field_at sh dfaops (DOT _final) (nth 1 fptrs Inhabitant_val) m).
    { entailer!. clear - H1. apply field_compatible_tdfaops_dfaops_final in H1.
      unfold field_at. entailer. }
    simpl; Intros. rename H1 into FCfinal.
    forward.
    replace_SEP 3 (field_at sh tdfaops (DOT _final) (nth 1 fptrs Inhabitant_val) m).
    { entailer!. unfold field_at; entailer. }
    forward_call.
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 1 W1 tdfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pm (*Msh Hmsh*) HWW.

    forward. Exists w.
    rewrite EDFA_fold_unfold; Exists fptrs cso m.
    first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
    entailer!. try apply derives_refl.
  Qed.
 
  Lemma verif_runClientComplement:
  semax_body clientVprog Client_Gprog f_runClientComplement runClientComplementspec.
  Proof. unfold runClientComplement_spec.
    start_function. rename H into HWW.
    forward_call.
    Intros w.
    rewrite DFA_fold_unfold. Intros fptrs cso m.
        first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
    apply JMeq_eq in H; subst m'; simpl snd.
    try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W: 0 <= 0 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 0).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 0 W dfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pm (*Msh Hmsh*) HWW.

    assert (W: 0 <= 2 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl WeekdayDFA defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call (Monday, gv, w, Vint (Int.repr 120), CharX). (*step X*)
    sep_apply (insertEntryInMT_sl_N WeekdayDFA defaultMethspec 2 W dfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pm (*Msh Hmsh*) HWW.

    forward_call (w, Monday, gv).
    { rewrite DFA_fold_unfold;  Exists fptrs cso m.
      first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
      entailer!. rewrite ! sepcon_assoc.  apply sepcon_derives.  apply derives_refl.
      rewrite sepcon_comm. apply sepcon_derives. apply derives_refl. cancel. }

    Intros u. deadvars!. clear - HWW.

    rewrite DFA_fold_unfold. Intros fptrs cso m.
    first [unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' msh].
    apply JMeq_eq in H; subst m'; simpl snd.
    try (rename H0 into Msh; rename H1 into Hmsh).

    assert (W: 0 <= 2 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl (ComplementDFA (*ch*) WeekdayDFA) defaultMethspec 2).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call (Monday, gv, u, Vint (Int.repr 110), CharN). (*step N*)
    sep_apply (insertEntryInMT_sl_N (ComplementDFA (*ch*) WeekdayDFA) defaultMethspec 2 W dfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pm (*Msh Hmsh*) HWW.

    assert (W: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl. rep_lia. }
    sep_apply (extractEntryFromMT_sl (ComplementDFA (*ch*) WeekdayDFA) defaultMethspec 1).
    Intros sh v. apply JMeq_eq in H1; subst v. unfold Znth; simpl.
    forward.
    forward.
    forward_call. (*final*)
    sep_apply (insertEntryInMT_sl_N (ComplementDFA (*ch*) WeekdayDFA) defaultMethspec 1 W dfaops fptrs sh _  (JMeq_refl _)).
    deadvars!. clear - Pm (*Msh Hmsh*) HWW.

    forward. Exists u.
    rewrite DFA_fold_unfold; Exists fptrs cso m.
    first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m msh].
    entailer!. try apply derives_refl.
  Qed.

  Lemma Client_GP gv (GV: globals_ok gv): InitGPred (Vardefs clientQP ) gv |-- Client_GV gv.
  Proof.
    unfold Vardefs. unfold InitGPred, Client_GV.
    simpl filter; unfold Maps.PTree.prev.
    simpl; Intros. rewrite sepcon_emp.
    assert (FC: field_compatible (Tstruct _storage noattr) []( gv _strg)) by auto with field_compatible.
    assert (MD:forall v (Hp: field_compatible (Tstruct _storage noattr) [] v),
            mapsto_zeros (AST.init_data_list_size (gvar_init v_strg)) Ews v
              |-- data_at_ Ews (Tstruct _storage noattr) v).
    { specialize (mapsto_zero_data_at_zero (Tstruct _storage noattr)  Ews).
      rewrite zero_val_eq. simpl. simpl; intros MD' pp Hpp.
      sep_apply MD'; clear MD'; auto. cancel. }
    unfold globvar2pred; simpl. entailer!.
  Qed.

  Definition Client_VSU: @VSU NullExtension.Espec nil Client_importedspecs clientQP 
             Client_internalspecs (*emp*)Client_GV.
  Proof.
     mkVSU prog Client_internalspecs.
     - solve_SF_internal verif_createWeekdayDFA.
     - solve_SF_internal verif_runClient.
     - solve_SF_internal verif_runClientComplement.
(*     - solve_SF_internal verif_runClientEither.*)
     - solve_SF_internal verif_runClientBoth.
     - solve_SF_internal verif_runStaticDFAs.
     - solve_SF_internal verif_runTClient.
     - solve_SF_internal verif_runTwiceClient.
     - solve_SF_internal verif_invokeInit.
     - solve_SF_internal verif_invokeFinal.
     - solve_SF_internal verif_invokeStep.
     - solve_SF_internal verif_invokeTwostepsEDFA.
     - apply Client_GP.
  Qed.

End Client_VSU.