Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Export ASI_surelyMalloc.
Require Import ObjectBase.
Require Import ASI_DFA.
Require Import ASI_ComplementDFA. 

Require Import ComplementDFA.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition complementQP:= ltac:(QPprog prog).
Definition complementVprog := QPvarspecs complementQP.
Notation object := (Tstruct _object noattr).
Notation complement := (Tstruct _complement noattr).
Notation dfaops := (Tstruct _dfaOps noattr).

Notation val := (Values.val).

Section Complement_VSU.
  Variable ch: Type.
  Variable Parent: @Dfa ch.
  Variable CHrep: ch -> val -> Prop.

  Variable Parent_APD : DFA_APD ch Parent.
  Variable M: MallocFreeAPD.
  Notation MTok := (MF_Tok M).
 (*we're a client of Parent but use OUR compspecs here*)

  Notation Prep :=
    (DFA ch CHrep Parent Parent_APD M object dfaops _ops _init _final _step CompSpecs). 
     (*we're a client of Parent but use OUR compspecs here. We also use the
    'client-exposed' structs (i.e. object dfaops) and identifiers (_ops _init _final _step)*)

  Notation ComplDFA := (ComplementDFA Parent).

  Definition Complement_Inv: GPred (Dfa_state ComplDFA):= fun x =>
  match x with ((d, gv), v) =>
   EX parent, Prep ((d, gv), parent) * field_at Ews complement [StructField _parent] parent v
  end.

  Definition complement_specs: list (GPred (Dfa_state ComplDFA) -> funspec) :=
     [init_spec object ch ComplDFA; final_spec object ch ComplDFA; step_spec object ch CHrep ComplDFA].

  Definition ComplMT (i f s:val): val -> GPred (Dfa_state ComplDFA) -> mpred :=
    MTable (Dfa_state ComplDFA) (*M*) dfaops [_init; _final; _step] complement_specs [i;f;s].

  Lemma ComplMT_duplicable i f s m P: ComplMT i f s m P |-- ComplMT i f s m P * ComplMT i f s m P.
  Proof. apply funcptrs_duplicable. Qed.

  Lemma make_ComplMT:
  forall sh1 sh2 sh3 i f s m, readable_share sh1 -> readable_share sh2 -> readable_share sh3 ->
  (func_ptr' (init_spec object ch ComplDFA Complement_Inv) i *
   func_ptr' (final_spec object ch ComplDFA Complement_Inv) f *
   func_ptr' (step_spec object ch CHrep ComplDFA Complement_Inv) s *
   field_at sh1 dfaops [StructField _init] i m *
   field_at sh2 dfaops [StructField _final] f m *
   field_at sh3 dfaops [StructField _step] s m
   (*data_at Ews methods (g,s) m*)
  |-- ComplMT i f s m Complement_Inv)%logic.
  Proof. intros. unfold ComplMT. simpl. Exists sh1 i sh2 f sh3 s. entailer!. Qed.

  Lemma Prep_isptr_or_null s p: Prep (s, p) |-- !! is_pointer_or_null p.
  Proof. sep_apply DFA_isptr. entailer!. Qed.
  #[local] Hint Resolve Prep_isptr_or_null: saturate_local.

Opaque funcptrs.


  Lemma verif_init_aux: semax_body complementVprog nil f_complement_init (_complement_init, init_spec object ch _ Complement_Inv).
  Proof.
    start_function. unfold Complement_Inv(*, withspacer*). Intros parent.
    forward.
    rewrite DFA_fold_unfold. Intros fptrs cso m. simpl.
    first [unfold FldAt;  Intros m' | unfold RdblFldAt; Intros m' rsh; rename H0 into Rsh; rename H1 into Hrsh].
    apply JMeq_eq in H. subst m'. simpl. forward.
Transparent funcptrs. simpl. Opaque funcptrs. 
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros ish iv. apply JMeq_eq in H. subst v0.
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros fsh fv. apply JMeq_eq in H. subst v0.
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros ssh sv. apply JMeq_eq in H. subst v0.
destruct fptrs. 2:{ apply (semax_pre FF). entailer!. apply semax_ff. } 
    forward.
    forward_call.
    forward.
    unfold Complement_Inv(*, withspacer*). Exists parent (*msh*). cancel.
    rewrite DFA_fold_unfold. Exists [iv;fv;sv] cso m. entailer!.
    rewrite ! sepcon_assoc. apply sepcon_derives. apply derives_refl.
    apply sepcon_derives. apply derives_refl.
    apply sepcon_derives. 
    first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m rsh]; entailer!.
    Transparent funcptrs. simpl. Opaque funcptrs.
    Exists ish iv fsh fv ssh sv. entailer!.
  Qed.

  Lemma verif_final_aux: semax_body complementVprog nil f_complement_final (_complement_final, final_spec object ch _ (*ComplementAPD*)Complement_Inv).
  Proof.
    start_function. unfold Complement_Inv. Intros parent (*msh. rename H into Msh. rename H0 into Hmsh*).
    forward.
    rewrite DFA_fold_unfold. Intros fptrs cso m. simpl.
    first [ unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' rsh; rename H0 into Rsh; rename H1 into Hrsh].
    apply JMeq_eq in H. subst m'. simpl. forward.
Transparent funcptrs. simpl. Opaque funcptrs. 
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros ish iv. apply JMeq_eq in H. subst v0.
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros fsh fv. apply JMeq_eq in H. subst v0.
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros ssh sv. apply JMeq_eq in H. subst v0.
destruct fptrs. 2:{ apply (semax_pre FF). entailer!. apply semax_ff. } 
    forward. forward_call. forward.
    apply andp_right.
    { entailer!. unfold ComplDFA; simpl. destruct (Dfa_fin Parent s); reflexivity. }
    unfold Complement_Inv(*, withspacer*). Exists parent (*msh*). cancel.
    rewrite DFA_fold_unfold. Exists [iv;fv;sv] cso m. entailer!.
    rewrite ! sepcon_assoc. apply sepcon_derives. apply derives_refl.
(*    apply sepcon_derives. apply derives_refl.*)
    apply sepcon_derives. 
    first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m rsh]; entailer!.
    Transparent funcptrs. simpl. Opaque funcptrs.
    Exists ish iv fsh fv ssh sv. entailer!.
  Qed.

  Lemma verif_step_aux: semax_body complementVprog nil f_complement_step (_complement_step, step_spec object ch CHrep _ (*ComplementAPD*)Complement_Inv).
  Proof. unfold step_spec.
    start_function. rename H into HCH. unfold Complement_Inv. simpl; Intros parent (*msh. rename H into Msh. rename H0 into Hmsh*).
    forward.
    rewrite DFA_fold_unfold. Intros fptrs cso m. simpl.
    first [ unfold FldAt; Intros m' | unfold RdblFldAt; Intros m' rsh; rename H0 into Rsh; rename H1 into Hrsh].
    apply JMeq_eq in H. subst m'. simpl. forward.
Transparent funcptrs. simpl. Opaque funcptrs. 
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros ish iv. apply JMeq_eq in H. subst v0.
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros fsh fv. apply JMeq_eq in H. subst v0.
destruct fptrs. apply (semax_pre FF). entailer!. apply semax_ff. Intros ssh sv. apply JMeq_eq in H. subst v0.
destruct fptrs. 2:{ apply (semax_pre FF). entailer!. apply semax_ff. }
     forward.
     forward_call (s, gv, parent, c, d).
    entailer!.
    unfold Complement_Inv(*, withspacer*). Exists parent (*msh*). cancel.
    rewrite DFA_fold_unfold. Exists [iv;fv;sv] cso m. entailer!.
    rewrite ! sepcon_assoc. apply sepcon_derives. apply derives_refl.
    apply sepcon_derives. apply derives_refl.
    apply sepcon_derives. 
    first [ unfold FldAt; Exists m | unfold RdblFldAt; Exists m rsh]; entailer!.
    Transparent funcptrs. simpl. Opaque funcptrs.
    Exists ish iv fsh fv ssh sv. entailer!.
  Qed.

  Definition ComplementGV (gv:globals): mpred :=
           EX sh1 sh2 sh3, !!(readable_share sh1 /\ readable_share sh2 /\ readable_share sh3) &&
                               (field_at sh1 dfaops (DOT _init)  (gv _complement_init)  (gv _complementops) *
                                field_at sh2 dfaops (DOT _final) (gv _complement_final) (gv _complementops) *
                                field_at sh3 dfaops (DOT _step)  (gv _complement_step)  (gv _complementops)).

  Lemma Complement_Inv_isptrnull a v: Complement_Inv (a, v) |-- !! is_pointer_or_null v.
  Proof. unfold Complement_Inv. destruct a; Intros parent (*msh*); entailer!. Qed.

  Definition ComplementAPD:= Build_DFA_APD ch ComplDFA  ComplementGV Complement_Inv (*Complement_Inv_isptrnull*).

  Notation mkComplspec := (makeComplement_spec _ Parent CHrep Parent_APD M object dfaops _ops _init _final _step CompSpecs   ComplementAPD).

  Definition Complement_internalspecs := [(_complement_init, init_spec object ch _ Complement_Inv);
                                       (_complement_final, final_spec object ch _ Complement_Inv);
                                       (_complement_step, step_spec object ch CHrep _ Complement_Inv);
                                       (_mkCompl, mkComplspec)]. 

  Definition Complement_Gprog : funspecs := Complement_internalspecs ++ [(_surely_malloc, surely_malloc_numericspec M)].

  Lemma align_compatible_complement_object {b i} (AL: align_compatible complement (Vptr b i)):
        align_compatible object (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

  Lemma field_compatible_complements_object_ops p: field_compatible complement [StructField _ops] p ->
        field_compatible object [StructField _ops] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_complement_object; trivial.
    - split; simpl; auto. left; trivial.
  Qed.

  Lemma ComplMT_GV' gv:
        ComplMT (gv _complement_init) (gv _complement_final) (gv _complement_step)
                 (gv _complementops) Complement_Inv
        |-- ComplementGV gv.
  Proof. unfold ComplMT. simpl.
    Transparent funcptrs. unfold MTable. Opaque funcptrs. simpl.
    Intros Ish gi Fsh gf Ssh gs. Exists Ish Fsh Ssh. entailer!.
    do 3 sep_apply func_ptr'_emp. cancel.
    apply JMeq_eq in H. subst gi. 
    apply JMeq_eq in H1. subst gf. 
    apply JMeq_eq in H3. subst gs. cancel.
  Qed.

  Lemma body_makeComplements: semax_body complementVprog Complement_Gprog f_mkCompl (_mkCompl, mkComplspec).
  Proof.
  unfold mkComplspec, makeComplement_spec, ComplementAPD, ComplementGV.
    (*redundant function_pointers.*)
  start_function. rename H into HWW.
  forward_call surelymalloc_spec_sub (complement, gv). { simpl. rep_lia. }
  Intros c. simpl.
  make_func_ptr _complement_init.
  make_func_ptr _complement_final.
  make_func_ptr _complement_step.
  simpl. Intros ish fsh ssh. rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh.
  sep_apply(make_ComplMT ish fsh ssh).
  forward. forward. forward.
  sep_apply ComplMT_duplicable. sep_apply ComplMT_GV'. unfold ComplementGV.
  Exists c. entailer!.
  rewrite (DFA_fold_unfold ch CHrep ComplDFA). (* ComplementAPD M object dfaops _ops _init _final _step CompSpecs).*)
  (*simpl. unfold Complement_Inv.*)
  Exists [gv _complement_init; gv _complement_final; gv _complement_step] (CompSpecs, complement).
  simpl. Exists q (*(gv _complementops)*) (gv _complementops). cancel.
  rewrite sepcon_comm. apply sepcon_derives.
  2: apply derives_refl. entailer!.
  unfold_data_at (data_at Ews complement _ c). 
  unfold Complement_Inv, FldAt, RdblFldAt.

  (*Exists (snd (slice.cleave Ews)) (gv _complementops) (fst (slice.cleave Ews)).
  rewrite <- (field_at_share_join (fst (slice.cleave Ews)) (snd (slice.cleave Ews))
                                  Ews complement [ StructField _ops])
  by apply slice.cleave_join.*)
  Exists (gv _complementops).
   
  entailer!.
  (*{ split; split. apply slice.cleave_readable1; auto. apply sepalg.join_sub_refl.
    apply slice.cleave_readable2; auto. apply sepalg.join_sub_refl. }*)
  clear - H2. apply field_compatible_complements_object_ops in H2.
  unfold field_at. simpl; entailer!.
  Qed.

  Lemma verif_init: semax_body complementVprog Complement_Gprog f_complement_init (_complement_init, init_spec object ch _ Complement_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body complementVprog Complement_Gprog f_complement_final (_complement_final, final_spec object ch _ Complement_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body complementVprog Complement_Gprog f_complement_step (_complement_step, step_spec object ch CHrep _ Complement_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.

  Lemma Complement_GP gv (GV: globals_ok gv):
        InitGPred (Vardefs complementQP) gv |-- ComplementGV gv.
  Proof. unfold Vardefs. unfold InitGPred, ComplementGV. Exists Ews Ews Ews.
    assert (X: filter isGvar (Maps.PTree.elements (QP.prog_defs complementQP)) = [(_complementops, Gvar v_complementops)]) by reflexivity.
    rewrite X; clear X.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer.
    (*unfold_data_at (data_at Ews dfaops _ (gv _complementops)).*)
    erewrite <- 3 mapsto_field_at; try reflexivity; auto with field_compatible.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _complement_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _complement_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _complement_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  Definition Complement_VSU: @VSU NullExtension.Espec nil [(_surely_malloc, surely_malloc_numericspec M)] complementQP 
             [(_mkCompl, mkComplspec)]
             ComplementGV.
 Proof.
     mkVSU prog Complement_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal body_makeComplements.
     - apply Complement_GP. 
  Qed.

End Complement_VSU.