Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Export ASI_surelyMalloc.
Require Import ObjectBase.
Require Import ASI_DFA.
Require Import ASI_IntersectionDFA.

Require Import IntersectionDFA.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition intersectionQP:= ltac:(QPprog prog).
Definition intersectionVprog := QPvarspecs intersectionQP.
Notation object := (Tstruct _object noattr).
Notation intersection := (Tstruct _intersection noattr).
Notation dfaops := (Tstruct _dfaOps noattr).

Notation val := (Values.val).

Section Intersection_VSU.
  Variable charL charR: Type.
  Variable ParentL: @Dfa charL.
  Variable ParentR: @Dfa charR.

  Variable CHrepL: charL -> val -> Prop.
  Variable CHrepR: charR -> val -> Prop.

  Variable L_APD : DFA_APD _ ParentL.
  Variable R_APD : DFA_APD _ ParentR.

  Variable M: MallocFreeAPD.
  Notation MTok := (MF_Tok M).
 (*we're a client of Parent but use OUR compspecs here*)

  Notation Lrep :=
    (DFA charL CHrepL ParentL L_APD M object dfaops _ops _init _final _step CompSpecs).

  Notation Rrep :=
    (DFA charR CHrepR ParentR R_APD M object dfaops _ops _init _final _step CompSpecs).
 
     (*we're a client of Parent but use OUR compspecs here. We also use the
    'client-exposed' structs (i.e. object dfaops) and identifiers (_ops _init _final _step)*)
  Definition IntersectionDFA := (IntersectionDFA ParentL ParentR).

  Definition Intersection_Inv: GPred (Dfa_state IntersectionDFA):= fun dv =>
   match dv with ((l,r), gv, v) =>
     EX left right, (Lrep (l, gv, left) * field_at Ews intersection [StructField _left] left v) *
                    (Rrep (r, gv, right) * field_at Ews intersection [StructField _right] right v)
   end.

  Notation ch:= (charL * charR)%type.
  Notation CHrep := (fun c v => CHrepL (fst c) v /\ CHrepR (snd c) v).

  Definition intersection_specs: list (GPred (Dfa_state IntersectionDFA) -> funspec) :=
     [init_spec object ch IntersectionDFA; final_spec object ch IntersectionDFA; step_spec object ch CHrep IntersectionDFA].

  Definition IntersectionMT (i f s:val): val -> GPred (Dfa_state IntersectionDFA) -> mpred :=
    MTable (Dfa_state IntersectionDFA) (*M*) dfaops [_init; _final; _step] intersection_specs [i;f;s].

  Lemma IntersectionMT_duplicable i f s m P: IntersectionMT i f s m P |-- IntersectionMT i f s m P * IntersectionMT i f s m P.
  Proof. apply funcptrs_duplicable. Qed.

  Lemma make_IntersectionMT:
  forall sh1 sh2 sh3 i f s m, readable_share sh1 -> readable_share sh2 -> readable_share sh3 ->
  (func_ptr' (init_spec object ch IntersectionDFA Intersection_Inv) i *
   func_ptr' (final_spec object ch IntersectionDFA Intersection_Inv) f *
   func_ptr' (step_spec object ch CHrep IntersectionDFA Intersection_Inv) s *
   field_at sh1 dfaops [StructField _init] i m *
   field_at sh2 dfaops [StructField _final] f m *
   field_at sh3 dfaops [StructField _step] s m
  |-- IntersectionMT i f s m Intersection_Inv)%logic.
  Proof. intros. unfold IntersectionMT. simpl. Exists sh1 i sh2 f sh3 s. entailer!. Qed.

  Lemma Lrep_isptr_or_null s p: Lrep (s, p) |-- !! is_pointer_or_null p.
  Proof. sep_apply DFA_isptr. entailer!. Qed.
  #[local] Hint Resolve Lrep_isptr_or_null: saturate_local.
  Lemma Rrep_isptr_or_null s p: Rrep (s, p) |-- !! is_pointer_or_null p.
  Proof. sep_apply DFA_isptr. entailer!. Qed.
  #[local] Hint Resolve Rrep_isptr_or_null: saturate_local.

Opaque funcptrs.

  Lemma verif_init_aux: semax_body intersectionVprog nil f_intersection_init (_intersection_init, init_spec object ch _ Intersection_Inv).
  Proof.
    start_function. unfold Intersection_Inv. 
    destruct s as [l r]. Intros left right (*msh. rename H into Msh. rename H0 into Hmsh*).
    forward.
    rewrite (DFA_fold_unfold charL). Intros fptrsL csoL mL.
    first [unfold FldAt; Intros mL' | unfold RdblFldAt; Intros mL' l_rsh; rename H0 into L_rsh; rename H1 into Hl_rsh].
    apply JMeq_eq in H. subst mL'. simpl. forward.
    
    assert (W: 0 <= 0 < Zlength [_init; _final; _step]). { unfold Zlength; simpl; rep_lia. }
    sep_apply (extractEntryFromMT_sl ParentL defaultMethspec 0).
    Intros sh ll. apply JMeq_eq in H1; subst ll. unfold Znth; simpl.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentL defaultMethspec 0 W dfaops fptrsL sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentL dfaops [_init; _final; _step] (DFA_ops_specs object charL CHrepL ParentL) fptrsL mL (DFA_Inv charL ParentL L_APD))
               (DFA_Inv charL ParentL L_APD (Dfa_s ParentL, gv, left))
               (malloc_token MTok Ews (snd csoL) left)
               (field_at Ews(*l_rsh*) object (DOT _ops) _ left).
(*    gather_SEP 0 1 4 5.*)
    replace_SEP 0 (Lrep (Dfa_s ParentL, gv, left)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsL csoL mL. entailer!. unfold FldAt, RdblFldAt.
      Exists mL (*l_rsh*); entailer!. }
    deadvars!. clear - (*Msh Hmsh*) W.

    forward.
    rewrite (DFA_fold_unfold charR). Intros fptrsR csoR mR.
    first [unfold FldAt; Intros mR' | unfold RdblFldAt; Intros mR' r_rsh; rename H0 into R_rsh; rename H1 into Hr_rsh].
    apply JMeq_eq in H. subst mR'. simpl. forward.
    sep_apply (extractEntryFromMT_sl ParentR defaultMethspec 0).
    Intros sh rr. apply JMeq_eq in H1; subst rr. unfold Znth; simpl.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentR defaultMethspec 0 W dfaops fptrsR sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentR dfaops [_init; _final; _step] (DFA_ops_specs object charR CHrepR ParentR) fptrsR mR (DFA_Inv charR ParentR R_APD))
               (DFA_Inv charR ParentR R_APD (Dfa_s ParentR, gv, right))
               (malloc_token MTok Ews (snd csoR) right)
               (field_at Ews(*r_rsh*) object (DOT _ops) _ right).
(*    gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Rrep (Dfa_s ParentR, gv, right)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsR csoR mR. entailer!. unfold FldAt, RdblFldAt.
      Exists mR (*r_rsh*); entailer!. }
    deadvars!. clear (*- Msh Hmsh*).

    forward. simpl. Exists left right (*msh*). entailer!.
  Qed.

  Lemma verif_final_aux: semax_body intersectionVprog nil f_intersection_final (_intersection_final, final_spec object ch _ (*IntersectionAPD*)Intersection_Inv).
  Proof.
    start_function. unfold Intersection_Inv.
    destruct s as [l r]. Intros left right (*msh; rename H into Msh. rename H0 into Hmsh*).

    rewrite (DFA_fold_unfold charL). Intros fptrsL csoL mL.
    first [ unfold FldAt; Intros mL' | unfold RdblFldAt; Intros mL' l_rsh; rename H0 into L_rsh; rename H1 into Hl_rsh].
    apply JMeq_eq in H. subst mL'. simpl. forward.
    
    assert (W: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl; rep_lia. }
    sep_apply (extractEntryFromMT_sl ParentL defaultMethspec 1).
    Intros sh ll. apply JMeq_eq in H1; subst ll. unfold Znth; simpl.
    forward. forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentL defaultMethspec 1 W dfaops fptrsL sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentL dfaops [_init; _final; _step] (DFA_ops_specs object charL CHrepL ParentL) fptrsL mL (DFA_Inv charL ParentL L_APD))
               (DFA_Inv charL ParentL L_APD (l, gv, left))
               (malloc_token MTok Ews (snd csoL) left)
               (field_at Ews(*l_rsh*) object (DOT _ops) _ left).
(*    gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Lrep (l, gv, left)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsL csoL mL. entailer!. unfold FldAt, RdblFldAt.
      Exists mL (*l_rsh*); entailer!. }
    forward. deadvars!. clear - (*Msh Hmsh*) W.

    rewrite (DFA_fold_unfold charR). Intros fptrsR csoR mR.
    first [ unfold FldAt; Intros mL' |  unfold RdblFldAt; Intros mR' r_rsh; rename H0 into R_rsh; rename H1 into Hr_rsh].
    apply JMeq_eq in H. subst (*mR'*)mL'. simpl. forward.

    sep_apply (extractEntryFromMT_sl ParentR defaultMethspec 1).
    Intros sh rr. apply JMeq_eq in H1; subst rr. unfold Znth; simpl.
    forward. forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentR defaultMethspec 1 W dfaops fptrsR sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentR dfaops [_init; _final; _step] (DFA_ops_specs object charR CHrepR ParentR) fptrsR mR (DFA_Inv charR ParentR R_APD))
               (DFA_Inv charR ParentR R_APD (r, gv, right))
               (malloc_token MTok Ews (snd csoR) right)
               (field_at Ews(*r_rsh*) object (DOT _ops) _ right).
(*  gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Rrep (r, gv, right)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsR csoR mR. entailer!. unfold FldAt, RdblFldAt.
      Exists mR (*r_rsh*); entailer!. }
    forward. deadvars!. clear (*- Msh Hmsh*).

    replace (force_val (sem_cast tbool tbool (if Dfa_fin ParentR r then Vtrue else Vfalse)))
       with (if (Dfa_fin ParentR r) then Vtrue else Vfalse ).
    2: destruct (Dfa_fin ParentR r); reflexivity.
    replace (force_val (sem_cast_i2bool (if Dfa_fin ParentL l then Vtrue else Vfalse)))
       with (if (Dfa_fin ParentL l) then Vtrue else Vfalse ).
    2: destruct (Dfa_fin ParentL l); reflexivity.

    forward_if
     (PROP ( )
      LOCAL (temp _br (if Dfa_fin ParentR r then Vtrue else Vfalse);
             temp _bl (if Dfa_fin ParentL l then Vtrue else Vfalse);
             temp _t'3 (if Dfa_fin ParentL l then (if Dfa_fin ParentR r then Vtrue else Vfalse) else Vfalse))
      SEP (Rrep (r, gv, right); Lrep (l, gv, left);
           field_at Ews intersection (DOT _left) left v;
           field_at Ews intersection (DOT _right) right v(*;
           field_at msh intersection (DOT _ops) (gv _intersectionops) v*))).
    + forward. entailer!. rewrite H. destruct (Dfa_fin ParentR r); reflexivity. 
    + forward. entailer!. rewrite H. reflexivity.
    + forward. clear (*- Msh Hmsh*). entailer!.
      - simpl. destruct (Dfa_fin ParentL l);
        destruct (Dfa_fin ParentR r); simpl; trivial.
      - simpl. Exists left right (*msh*). entailer!.
  Qed. 

  Lemma verif_step_aux: semax_body intersectionVprog nil f_intersection_step (_intersection_step, step_spec object ch CHrep _ (*IntersectionAPD*)Intersection_Inv).
  Proof. unfold step_spec.
    start_function. destruct H as [CHL CHR].
    unfold Intersection_Inv. destruct s as [l r]; Intros left right (*msh. rename H into Msh. rename H0 into Hmsh*).

    rewrite (DFA_fold_unfold charL). Intros fptrsL csoL mL.
    first [ unfold FldAt; Intros mL' | unfold RdblFldAt; Intros mL' l_rsh; rename H0 into L_rsh; rename H1 into Hl_rsh].
    apply JMeq_eq in H. subst mL'. simpl. forward.
    
    assert (W: 0 <= 2 < Zlength [_init; _final; _step]). { unfold Zlength; simpl; rep_lia. }
    sep_apply (extractEntryFromMT_sl ParentL defaultMethspec 2).
    Intros sh ll. apply JMeq_eq in H1; subst ll. unfold Znth; simpl.
    forward. forward.
    forward_call (l, gv, left, c, fst d).
    sep_apply (insertEntryInMT_sl_N ParentL defaultMethspec 2 W dfaops fptrsL sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentL dfaops [_init; _final; _step] (DFA_ops_specs object charL CHrepL ParentL) fptrsL mL (DFA_Inv charL ParentL L_APD))
               (DFA_Inv charL ParentL L_APD (Dfa_trans ParentL l (fst d), gv, left))
               (malloc_token MTok Ews (snd csoL) left)
               (field_at Ews(*l_rsh*) object (DOT _ops) _ left).
(*    gather_SEP 0 1 2 Msh 3.*)
    replace_SEP 0 (Lrep (Dfa_trans ParentL l (fst d), gv, left)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsL csoL mL. entailer!. unfold FldAt, RdblFldAt.
      Exists mL (*l_rsh*); entailer!. }
    deadvars!. clear - (*Msh Hmsh*) W CHR.

    rewrite (DFA_fold_unfold charR). Intros fptrsR csoR mR. simpl.
    first [ unfold FldAt; Intros mR' | unfold RdblFldAt; Intros mR' r_rsh; rename H0 into R_rsh; rename H1 into Hr_rsh].
    apply JMeq_eq in H. subst mR'. simpl. forward.

    sep_apply (extractEntryFromMT_sl ParentR defaultMethspec 2).
    Intros sh rr. apply JMeq_eq in H1; subst rr. unfold Znth; simpl.
    forward. forward.
    forward_call (r, gv, right, c, snd d).
    sep_apply (insertEntryInMT_sl_N ParentR defaultMethspec 2 W dfaops fptrsR sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentR dfaops [_init; _final; _step] (DFA_ops_specs object charR CHrepR ParentR) fptrsR mR (DFA_Inv charR ParentR R_APD))
               (DFA_Inv charR ParentR R_APD (Dfa_trans ParentR r (snd d), gv, right))
               (malloc_token MTok Ews (snd csoR) right)
               (field_at Ews(*r_rsh*) object (DOT _ops) _ right).
(*    gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Rrep (Dfa_trans ParentR r (snd d), gv, right)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsR csoR mR. entailer!. unfold FldAt, RdblFldAt.
      Exists mR (*r_rsh*); entailer!. }
    clear (*- Msh Hmsh*). entailer!. simpl. Exists left right (*msh*). entailer!.
  Qed.

  Definition IntersectionGV (gv:globals): mpred :=
           EX sh1 sh2 sh3, !!(readable_share sh1 /\ readable_share sh2 /\ readable_share sh3) &&
                               (field_at sh1 dfaops (DOT _init)  (gv _intersection_init)  (gv _intersectionops) *
                                field_at sh2 dfaops (DOT _final) (gv _intersection_final) (gv _intersectionops) *
                                field_at sh3 dfaops (DOT _step)  (gv _intersection_step)  (gv _intersectionops)).

  Lemma Intersection_Inv_isptrnull agv v: Intersection_Inv (agv, v) |-- !! is_pointer_or_null v.
  Proof. simpl. destruct agv as [[l r] gv]; Intros left right. entailer!. Qed.

  Definition IntersectionAPD:= Build_DFA_APD ch IntersectionDFA IntersectionGV  Intersection_Inv (*Intersection_Inv_isptrnull*).

  Definition mkIntersectionspec := (@makeIntersection_spec charL charR ParentL ParentR CHrepL CHrepR L_APD R_APD M object dfaops _ops _init _final _step CompSpecs IntersectionAPD).

  Definition Intersection_internalspecs := [(_intersection_init, init_spec object ch _ Intersection_Inv);
                                       (_intersection_final, final_spec object ch _ Intersection_Inv);
                                       (_intersection_step, step_spec object ch CHrep _ Intersection_Inv);
                                       (_make_IntersectionDFA, mkIntersectionspec)]. 

  Definition Intersection_Gprog : funspecs := Intersection_internalspecs ++ [(_surely_malloc, surely_malloc_numericspec M)].
  (* SM_ASI free surelyMalloc.*)

  Lemma align_compatible_intersection_object {b i} (AL: align_compatible intersection (Vptr b i)):
        align_compatible object (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { subst i0. inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

  Lemma field_compatible_intersection_object_ops p: field_compatible intersection [StructField _ops] p ->
        field_compatible object [StructField _ops] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_intersection_object; trivial.
    - split; simpl; auto. left; trivial.
  Qed.

  Lemma IntersectionMT_GV' gv:
        IntersectionMT (gv _intersection_init) (gv _intersection_final) (gv _intersection_step)
                 (gv _intersectionops) Intersection_Inv
        |-- IntersectionGV gv.
  Proof. unfold IntersectionMT. simpl.
    Transparent funcptrs. unfold MTable. Opaque funcptrs. simpl.
    Intros Ish gi Fsh gf Ssh gs. Exists Ish Fsh Ssh. entailer!.
    do 3 sep_apply func_ptr'_emp. cancel.
    apply JMeq_eq in H. subst gi. 
    apply JMeq_eq in H1. subst gf. 
    apply JMeq_eq in H3. subst gs. cancel.
  Qed.

  Lemma body_makeIntersections: semax_body intersectionVprog Intersection_Gprog f_make_IntersectionDFA (_make_IntersectionDFA, mkIntersectionspec).
  Proof.
  unfold mkIntersectionspec, makeIntersection_spec, IntersectionAPD, IntersectionGV.
    (*redundant function_pointers.*)
  start_function. rename H into HWW.
  forward_call surelymalloc_spec_sub (intersection, gv). { simpl. rep_lia. }
  Intros s.
  make_func_ptr _intersection_init.
  make_func_ptr _intersection_final.
  make_func_ptr _intersection_step.
  simpl. Intros ish fsh ssh. rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh.
  sep_apply(make_IntersectionMT ish fsh ssh).
  forward. forward. forward. forward.
  sep_apply IntersectionMT_duplicable. sep_apply IntersectionMT_GV'. unfold IntersectionGV.
  Exists s. entailer!.
  rewrite (DFA_fold_unfold ch CHrep IntersectionDFA). (* IntersectionAPD M object dfaops _ops _init _final _step CompSpecs).*)
  (*simpl. unfold Intersection_Inv.*)
  Exists [gv _intersection_init; gv _intersection_final; gv _intersection_step] (CompSpecs, intersection).
  simpl. Exists l r. cancel. Exists (gv _intersectionops).

 cancel.
  rewrite sepcon_comm. apply sepcon_derives.
  2: apply derives_refl. entailer!.
  unfold_data_at (data_at Ews intersection _ s). 
  unfold FldAt, RdblFldAt.
  (*Exists (snd (slice.cleave Ews)) (gv _intersectionops) (fst (slice.cleave Ews)).
  rewrite <- (field_at_share_join (fst (slice.cleave Ews)) (snd (slice.cleave Ews))
                                  Ews intersection [ StructField _ops])
  by apply slice.cleave_join.*) Exists (gv _intersectionops).
  entailer!.
  (*{ split; split. apply slice.cleave_readable1; auto. apply sepalg.join_sub_refl.
    apply slice.cleave_readable2; auto. apply sepalg.join_sub_refl. }*)
  clear - H2. apply field_compatible_intersection_object_ops in H2.
  unfold field_at. simpl; entailer!.
  Qed.

  Lemma verif_init: semax_body intersectionVprog Intersection_Gprog f_intersection_init (_intersection_init, init_spec object ch _ Intersection_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body intersectionVprog Intersection_Gprog f_intersection_final (_intersection_final, final_spec object ch _ Intersection_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body intersectionVprog Intersection_Gprog f_intersection_step (_intersection_step, step_spec object ch CHrep _ Intersection_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.

  Lemma Intersection_GP gv (GV: globals_ok gv):
        InitGPred (Vardefs intersectionQP) gv |-- (*data_at Ews dfaops (gv _intersection_init, (gv _intersection_final, gv _intersection_step)) (gv _intersectionops).*)
        IntersectionGV gv.
  Proof. unfold Vardefs. unfold InitGPred, IntersectionGV. Exists Ews Ews Ews.
    assert (X: filter isGvar (Maps.PTree.elements (QP.prog_defs intersectionQP)) = [(_intersectionops, Gvar v_intersectionops)]) by reflexivity.
    rewrite X; clear X.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer.
    (*unfold_data_at (data_at Ews dfaops _ (gv _intersectionops)).*)
    erewrite <- 3 mapsto_field_at; try reflexivity; auto with field_compatible.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _intersection_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _intersection_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _intersection_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  (*For now, we only export the constructor. If later "subclasses" inherit init, final, or step, well need to export these here, too*)
  Definition Intersection_VSU: @VSU NullExtension.Espec nil [(_surely_malloc, surely_malloc_numericspec M)] intersectionQP 
             [(*(_cell_get, get_spec object _ m1 CellGPred); 
              (_cell_set, set_spec object _ m1 CellGPred);*)
              (_make_IntersectionDFA,mkIntersectionspec)
             ]
             IntersectionGV.
  Proof.
     mkVSU prog Intersection_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal body_makeIntersections.
     - apply Intersection_GP. 
  Qed.

End Intersection_VSU.