Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Export ASI_surelyMalloc.
Require Import ObjectBase.
Require Import ASI_DFA.
Require Import ASI_ProductDFA.

Require Import ProductDFA.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition productQP:= ltac:(QPprog prog).
Definition productVprog := QPvarspecs productQP.
Notation object := (Tstruct _object noattr).
Notation product := (Tstruct _product noattr).
Notation dfaops := (Tstruct _dfaOps noattr).

Notation val := (Values.val).

Section Product_VSU.
  Variable SigmaL SigmaR: Type.
  Variable ParentL: @Dfa SigmaL.
  Variable ParentR: @Dfa SigmaR.

  Variable CHrepL: SigmaL -> val -> Prop.
  Variable CHrepR: SigmaR -> val -> Prop.

  Variable L_APD : DFA_APD _ ParentL.
  Variable R_APD : DFA_APD _ ParentR.

  Variable M: MallocFreeAPD.
  Notation MTok := (MF_Tok M).
 (*we're a client of Parent but use OUR compspecs here*)

  Notation Lrep :=
    (DFA SigmaL CHrepL ParentL L_APD M object dfaops _ops _init _final _step CompSpecs).

  Notation Rrep :=
    (DFA SigmaR CHrepR ParentR R_APD M object dfaops _ops _init _final _step CompSpecs).
 
     (*we're a client of Parent but use OUR compspecs here. We also use the
    'client-exposed' structs (i.e. object dfaops) and identifiers (_ops _init _final _step)*)
  Definition ProductDFA := (ProductDFA ParentL ParentR).

  Definition Product_Inv: GPred (Dfa_state ProductDFA):= fun dv =>
   match dv with ((l,r), gv, v) =>
     EX left right, (Lrep (l, gv, left) * field_at Ews product [StructField _left] left v) *
                    (Rrep (r, gv, right) * field_at Ews product [StructField _right] right v)
   end.

  Notation ch:= (SigmaL * SigmaR)%type.
  Notation CHrep := (fun c v => CHrepL (fst c) v /\ CHrepR (snd c) v).

  Definition product_specs: list (GPred (Dfa_state ProductDFA) -> funspec) :=
     [init_spec object ch ProductDFA; final_spec object ch ProductDFA; step_spec object ch CHrep ProductDFA].

  Definition ProductMT (i f s:val): val -> GPred (Dfa_state ProductDFA) -> mpred :=
    MTable (Dfa_state ProductDFA) (*M*) dfaops [_init; _final; _step] product_specs [i;f;s].

  Lemma ProductMT_duplicable i f s m P: ProductMT i f s m P |-- ProductMT i f s m P * ProductMT i f s m P.
  Proof. apply funcptrs_duplicable. Qed.

  Lemma make_ProductMT:
  forall sh1 sh2 sh3 i f s m, readable_share sh1 -> readable_share sh2 -> readable_share sh3 ->
  (func_ptr' (init_spec object ch ProductDFA Product_Inv) i *
   func_ptr' (final_spec object ch ProductDFA Product_Inv) f *
   func_ptr' (step_spec object ch CHrep ProductDFA Product_Inv) s *
   field_at sh1 dfaops [StructField _init] i m *
   field_at sh2 dfaops [StructField _final] f m *
   field_at sh3 dfaops [StructField _step] s m
   (*data_at Ews methods (g,s) m*)
  |-- ProductMT i f s m Product_Inv)%logic.
  Proof. intros. unfold ProductMT. simpl. Exists sh1 i sh2 f sh3 s. entailer!. Qed.

  Lemma Lrep_isptr_or_null s p: Lrep (s, p) |-- !! is_pointer_or_null p.
  Proof. sep_apply DFA_isptr. entailer!. Qed.
  #[local] Hint Resolve Lrep_isptr_or_null: saturate_local.
  Lemma Rrep_isptr_or_null s p: Rrep (s, p) |-- !! is_pointer_or_null p.
  Proof. sep_apply DFA_isptr. entailer!. Qed.
  #[local] Hint Resolve Rrep_isptr_or_null: saturate_local.

Opaque funcptrs.

  Lemma verif_init_aux: semax_body productVprog nil f_product_init (_product_init, init_spec object ch _ Product_Inv).
  Proof.
    start_function. unfold Product_Inv. 
    destruct s as [l r]. Intros left right (*msh. rename H into Msh. rename H0 into Hmsh*).
    forward.
    rewrite (DFA_fold_unfold SigmaL). Intros fptrsL csoL mL.
    first [ unfold FldAt; Intros mL' | unfold RdblFldAt; Intros mL' l_rsh; rename H0 into L_rsh; rename H1 into Hl_rsh].
    apply JMeq_eq in H. subst mL'. simpl. forward.
    
    assert (W: 0 <= 0 < Zlength [_init; _final; _step]). { unfold Zlength; simpl; rep_lia. }
    sep_apply (extractEntryFromMT_sl ParentL defaultMethspec 0).
    Intros sh ll. apply JMeq_eq in H1; subst ll. unfold Znth; simpl.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentL defaultMethspec 0 W dfaops fptrsL sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentL dfaops [_init; _final; _step] (DFA_ops_specs object SigmaL CHrepL ParentL) fptrsL mL (DFA_Inv SigmaL ParentL L_APD))
               (DFA_Inv SigmaL ParentL L_APD (Dfa_s ParentL, gv, left))
               (malloc_token MTok Ews (snd csoL) left)
               (field_at Ews(*l_rsh*) object (DOT _ops) _ left).
(*    gather_SEP 0 1 4 5.*)
    replace_SEP 0 (Lrep (Dfa_s ParentL, gv, left)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsL csoL mL. entailer!. unfold FldAt, RdblFldAt.
      Exists mL (*l_rsh*); entailer!. }
    deadvars!. clear - (*Msh Hmsh*) W.

    forward.
    rewrite (DFA_fold_unfold SigmaR). Intros fptrsR csoR mR.
    first [ unfold FldAt; Intros mR' | unfold RdblFldAt; Intros mR' r_rsh; rename H0 into R_rsh; rename H1 into Hr_rsh].
    apply JMeq_eq in H. subst mR'. simpl. forward.
    sep_apply (extractEntryFromMT_sl ParentR defaultMethspec 0).
    Intros sh rr. apply JMeq_eq in H1; subst rr. unfold Znth; simpl.
    forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentR defaultMethspec 0 W dfaops fptrsR sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentR dfaops [_init; _final; _step] (DFA_ops_specs object SigmaR CHrepR ParentR) fptrsR mR (DFA_Inv SigmaR ParentR R_APD))
               (DFA_Inv SigmaR ParentR R_APD (Dfa_s ParentR, gv, right))
               (malloc_token MTok Ews (snd csoR) right)
               (field_at Ews(*r_rsh*) object (DOT _ops) _ right).
(*    gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Rrep (Dfa_s ParentR, gv, right)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsR csoR mR. entailer!. unfold FldAt, RdblFldAt.
      Exists mR (*r_rsh*); entailer!. }
    deadvars!. (* clear - Msh Hmsh.*)

    forward. simpl. Exists left right (*msh*). entailer!.
  Qed.

  Lemma verif_final_aux: semax_body productVprog nil f_product_final (_product_final, final_spec object ch _ (*ProductAPD*)Product_Inv).
  Proof.
    start_function. unfold Product_Inv.
    destruct s as [l r]. Intros left right (*msh. rename H into Msh. rename H0 into Hmsh*).

    rewrite (DFA_fold_unfold SigmaL). Intros fptrsL csoL mL.
    first [unfold FldAt; Intros mL' | unfold RdblFldAt; Intros mL' l_rsh; rename H0 into L_rsh; rename H1 into Hl_rsh].
    apply JMeq_eq in H. subst mL'. simpl. forward.
    
    assert (W: 0 <= 1 < Zlength [_init; _final; _step]). { unfold Zlength; simpl; rep_lia. }
    sep_apply (extractEntryFromMT_sl ParentL defaultMethspec 1).
    Intros sh ll. apply JMeq_eq in H1; subst ll. unfold Znth; simpl.
    forward. forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentL defaultMethspec 1 W dfaops fptrsL sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentL dfaops [_init; _final; _step] (DFA_ops_specs object SigmaL CHrepL ParentL) fptrsL mL (DFA_Inv SigmaL ParentL L_APD))
               (DFA_Inv SigmaL ParentL L_APD (l, gv, left))
               (malloc_token MTok Ews (snd csoL) left)
               (field_at Ews(*l_rsh*) object (DOT _ops) _ left).
(*    gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Lrep (l, gv, left)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsL csoL mL. entailer!. unfold FldAt, RdblFldAt.
      Exists mL (*l_rsh*); entailer!. }
    forward. deadvars!. clear - (* Msh Hmsh *)W.

    rewrite (DFA_fold_unfold SigmaR). Intros fptrsR csoR mR.
    first [unfold FldAt; Intros mL' | unfold RdblFldAt; Intros mR' r_rsh; rename H0 into R_rsh; rename H1 into Hr_rsh].
    apply JMeq_eq in H. subst (*mR'*)mL'. simpl. forward.

    sep_apply (extractEntryFromMT_sl ParentR defaultMethspec 1).
    Intros sh rr. apply JMeq_eq in H1; subst rr. unfold Znth; simpl.
    forward. forward.
    forward_call.
    sep_apply (insertEntryInMT_sl_N ParentR defaultMethspec 1 W dfaops fptrsR sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentR dfaops [_init; _final; _step] (DFA_ops_specs object SigmaR CHrepR ParentR) fptrsR mR (DFA_Inv SigmaR ParentR R_APD))
               (DFA_Inv SigmaR ParentR R_APD (r, gv, right))
               (malloc_token MTok Ews (snd csoR) right)
               (field_at Ews(*r_rsh*) object (DOT _ops) _ right).
(*  gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Rrep (r, gv, right)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsR csoR mR. entailer!. unfold FldAt, RdblFldAt.
      Exists mR (*r_rsh*); entailer!. }
    forward. deadvars!. (*clear - Msh Hmsh.*)

    replace (force_val (sem_cast tbool tbool (if Dfa_fin ParentR r then Vtrue else Vfalse)))
       with (if (Dfa_fin ParentR r) then Vtrue else Vfalse ).
    2: destruct (Dfa_fin ParentR r); reflexivity.
    replace (force_val (sem_cast_i2bool (if Dfa_fin ParentL l then Vtrue else Vfalse)))
       with (if (Dfa_fin ParentL l) then Vtrue else Vfalse ).
    2: destruct (Dfa_fin ParentL l); reflexivity.

    forward_if
     (PROP ( )
      LOCAL (temp _br (if Dfa_fin ParentR r then Vtrue else Vfalse);
             temp _bl (if Dfa_fin ParentL l then Vtrue else Vfalse);
             temp _t'3 (if Dfa_fin ParentL l then Vtrue else (if Dfa_fin ParentR r then Vtrue else Vfalse)))
      SEP (Rrep (r, gv, right); Lrep (l, gv, left);
           field_at Ews product (DOT _left) left v;
           field_at Ews product (DOT _right) right v(*;
           field_at msh product (DOT _ops) (gv _productops) v*))).
    + forward. entailer!. try rewrite H; try rewrite H1. reflexivity. 
    + forward. entailer!. try rewrite H; try rewrite H1. destruct (Dfa_fin ParentR r); reflexivity.
    + forward. clear (*- Msh Hmsh*). entailer!.
      - simpl. destruct (Dfa_fin ParentL l);
        destruct (Dfa_fin ParentR r); simpl; trivial.
      - simpl. Exists left right (*msh*). entailer!.
  Qed.

  Lemma verif_step_aux: semax_body productVprog nil f_product_step (_product_step, step_spec object ch CHrep _ (*ProductAPD*)Product_Inv).
  Proof. unfold step_spec.
    start_function. destruct H as [CHL CHR].
    unfold Product_Inv. destruct s as [l r]; Intros left right (*msh. rename H into Msh. rename H0 into Hmsh*).

    rewrite (DFA_fold_unfold SigmaL). Intros fptrsL csoL mL.
    first [unfold FldAt; Intros mL' | unfold RdblFldAt; Intros mL' l_rsh; rename H0 into L_rsh; rename H1 into Hl_rsh].
    apply JMeq_eq in H. subst mL'. simpl. forward.
    
    assert (W: 0 <= 2 < Zlength [_init; _final; _step]). { unfold Zlength; simpl; rep_lia. }
    sep_apply (extractEntryFromMT_sl ParentL defaultMethspec 2).
    Intros sh ll. apply JMeq_eq in H1; subst ll. unfold Znth; simpl.
    forward. forward.
    forward_call (l, gv, left, c, fst d).
    sep_apply (insertEntryInMT_sl_N ParentL defaultMethspec 2 W dfaops fptrsL sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentL dfaops [_init; _final; _step] (DFA_ops_specs object SigmaL CHrepL ParentL) fptrsL mL (DFA_Inv SigmaL ParentL L_APD))
               (DFA_Inv SigmaL ParentL L_APD (Dfa_trans ParentL l (fst d), gv, left))
               (malloc_token MTok Ews (snd csoL) left)
               (field_at Ews(*l_rsh*) object (DOT _ops) _ left).
(*    gather_SEP 0 1 2 Msh 3.*)
    replace_SEP 0 (Lrep (Dfa_trans ParentL l (fst d), gv, left)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsL csoL mL. entailer!. unfold FldAt, RdblFldAt.
      Exists mL (*l_rsh*); entailer!. }
    deadvars!. clear - (*Msh Hmsh*) W CHR.

    rewrite (DFA_fold_unfold SigmaR). Intros fptrsR csoR mR. simpl.
    first [unfold FldAt; Intros mR' | unfold RdblFldAt; Intros mR' r_rsh; rename H0 into R_rsh; rename H1 into Hr_rsh].
    apply JMeq_eq in H. subst mR'. simpl. forward.

    sep_apply (extractEntryFromMT_sl ParentR defaultMethspec 2).
    Intros sh rr. apply JMeq_eq in H1; subst rr. unfold Znth; simpl.
    forward. forward.
    forward_call (r, gv, right, c, snd d).
    sep_apply (insertEntryInMT_sl_N ParentR defaultMethspec 2 W dfaops fptrsR sh _  (JMeq_refl _)).

    gather_SEP (MTable ParentR dfaops [_init; _final; _step] (DFA_ops_specs object SigmaR CHrepR ParentR) fptrsR mR (DFA_Inv SigmaR ParentR R_APD))
               (DFA_Inv SigmaR ParentR R_APD (Dfa_trans ParentR r (snd d), gv, right))
               (malloc_token MTok Ews (snd csoR) right)
               (field_at Ews (*r_rsh*) object (DOT _ops) _ right).
(*    gather_SEP 0 1 2 3.*)
    replace_SEP 0 (Rrep (Dfa_trans ParentR r (snd d), gv, right)).
    { rewrite DFA_fold_unfold. entailer!.
      Exists fptrsR csoR mR. entailer!. unfold FldAt, RdblFldAt.
      Exists mR (*r_rsh*); entailer!. }
    clear (*- Msh Hmsh*). entailer!. simpl. Exists left right (*msh*). entailer!.
  Qed.

  Definition ProductGV (gv:globals): mpred :=
           EX sh1 sh2 sh3, !!(readable_share sh1 /\ readable_share sh2 /\ readable_share sh3) &&
                               (field_at sh1 dfaops (DOT _init)  (gv _product_init)  (gv _productops) *
                                field_at sh2 dfaops (DOT _final) (gv _product_final) (gv _productops) *
                                field_at sh3 dfaops (DOT _step)  (gv _product_step)  (gv _productops)).

  Lemma Product_Inv_isptrnull agv v: Product_Inv (agv, v) |-- !! is_pointer_or_null v.
  Proof. simpl. destruct agv as [[l r] gv]; Intros left right. entailer!. Qed.

  Definition ProductAPD:= Build_DFA_APD ch ProductDFA ProductGV Product_Inv (*Product_Inv_isptrnull*).

  Definition mkProductspec := (@makeProduct_spec SigmaL SigmaR ParentL ParentR CHrepL CHrepR L_APD R_APD M object dfaops _ops _init _final _step CompSpecs ProductAPD).

  Definition Product_internalspecs := [(_product_init, init_spec object ch _ Product_Inv);
                                       (_product_final, final_spec object ch _ Product_Inv);
                                       (_product_step, step_spec object ch CHrep _ Product_Inv);
                                       (_mkProduct, mkProductspec)]. 

  Definition Product_Gprog : funspecs := Product_internalspecs ++ [(_surely_malloc, surely_malloc_numericspec M)].
  (* SM_ASI free surelyMalloc.*)

  Lemma align_compatible_product_object {b i} (AL: align_compatible product (Vptr b i)):
        align_compatible object (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { subst i0. inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

  Lemma field_compatible_product_object_ops p: field_compatible product [StructField _ops] p ->
        field_compatible object [StructField _ops] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_product_object; trivial.
    - split; simpl; auto. left; trivial.
  Qed.

  Lemma ProductMT_GV' gv:
        ProductMT (gv _product_init) (gv _product_final) (gv _product_step)
                 (gv _productops) Product_Inv
        |-- ProductGV gv.
  Proof. unfold ProductMT. simpl.
    Transparent funcptrs. unfold MTable. Opaque funcptrs. simpl.
    Intros Ish gi Fsh gf Ssh gs. Exists Ish Fsh Ssh. entailer!.
    do 3 sep_apply func_ptr'_emp. cancel.
    apply JMeq_eq in H. subst gi. 
    apply JMeq_eq in H1. subst gf. 
    apply JMeq_eq in H3. subst gs. cancel.
  Qed.

  Lemma body_makeProducts: semax_body productVprog Product_Gprog f_mkProduct (_mkProduct, mkProductspec).
  Proof.
  unfold mkProductspec, makeProduct_spec, ProductAPD, ProductGV.
    (*redundant function_pointers.*)
  start_function. rename H into HWW.
  forward_call surelymalloc_spec_sub (product, gv). { simpl. rep_lia. }
  Intros s.
  make_func_ptr _product_init.
  make_func_ptr _product_final.
  make_func_ptr _product_step.
  simpl. Intros ish fsh ssh. rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh.
  sep_apply(make_ProductMT ish fsh ssh).
  forward. forward. forward. forward.
  sep_apply ProductMT_duplicable. sep_apply ProductMT_GV'. unfold ProductGV.
  Exists s. entailer!.
  rewrite (DFA_fold_unfold ch CHrep ProductDFA). 
  Exists [gv _product_init; gv _product_final; gv _product_step] (CompSpecs, product).
  simpl. Exists l r. cancel. Exists (gv _productops).

  cancel.
  rewrite sepcon_comm. apply sepcon_derives.
  2: apply derives_refl. entailer!.
  unfold_data_at (data_at Ews product _ s). 
  unfold FldAt,RdblFldAt.
  (*Exists (snd (slice.cleave Ews)) (gv _productops) (fst (slice.cleave Ews)).
  rewrite <- (field_at_share_join (fst (slice.cleave Ews)) (snd (slice.cleave Ews))
                                  Ews product [ StructField _ops])
  by apply slice.cleave_join.*)
  Exists (gv _productops).
  entailer!.
  (*{ split; split. apply slice.cleave_readable1; auto. apply sepalg.join_sub_refl.
    apply slice.cleave_readable2; auto. apply sepalg.join_sub_refl. }*)
  clear - H2. apply field_compatible_product_object_ops in H2.
  unfold field_at. simpl; entailer!.
  Qed.

  Lemma verif_init: semax_body productVprog Product_Gprog f_product_init (_product_init, init_spec object ch _ Product_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body productVprog Product_Gprog f_product_final (_product_final, final_spec object ch _ Product_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body productVprog Product_Gprog f_product_step (_product_step, step_spec object ch CHrep _ Product_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.

  Lemma Product_GP gv (GV: globals_ok gv):
        InitGPred (Vardefs productQP) gv |-- (*data_at Ews dfaops (gv _product_init, (gv _product_final, gv _product_step)) (gv _productops).*)
        ProductGV gv.
  Proof. unfold Vardefs. unfold InitGPred, ProductGV. Exists Ews Ews Ews.
    assert (X: filter isGvar (Maps.PTree.elements (QP.prog_defs productQP)) = [(_productops, Gvar v_productops)]) by reflexivity.
    rewrite X; clear X.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer.
    (*unfold_data_at (data_at Ews dfaops _ (gv _productops)).*)
    erewrite <- 3 mapsto_field_at; try reflexivity; auto with field_compatible.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _product_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _product_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _product_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  (*In addition to the constructor spec we export init and step, since these methods are inherited in IntersectionInherit.
    This is not described in the coference submission*)
  Definition Product_VSU: @VSU NullExtension.Espec nil [(_surely_malloc, surely_malloc_numericspec M)] productQP 
             [(_mkProduct,mkProductspec);
              (_product_init, init_spec object ch _ Product_Inv);
              (_product_step, step_spec object ch CHrep _ Product_Inv)
             ]
             ProductGV.
  Proof.
     mkVSU prog Product_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal body_makeProducts.
     - apply Product_GP. 
  Qed.

End Product_VSU.