Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import ASI_SingleDFA.
Require Import TrueFalseDFA.

Require Import SingleDFA.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition singleQP:= ltac:(QPprog prog).
Definition singleVprog := QPvarspecs singleQP.

Section Single_VSU.

Definition SingleGV: GVPred (Dfa_state TrueFalseDFA):= fun d gv =>
           data_at Ews tbool (Val.of_bool d) (gv _DFA_state).

  Definition SingleTF_APD:= Build_SingleDFA_APD _ TrueFalseDFA SingleGV.

  Lemma create_singleTF: VSU_initializer prog (SingleGV false).
  Proof. unfold VSU_initializer; intros gv GV. unfold Vardefs. unfold InitGPred, SingleGV.
    simpl filter; unfold Maps.PTree.prev.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer!.
    assert (FC: field_compatible tbool [] (gv _DFA_state)) by auto with field_compatible.
    assert (MD:forall p (Hp:field_compatible tbool [] p),
            mapsto_zeros 1 Ews p |-- data_at Ews tbool Vfalse p).
    { specialize (mapsto_zero_data_at_zero tbool Ews).
      rewrite zero_val_eq. simpl. simpl; intros MD p Hp.
      apply MD; clear MD; auto. }
    sep_apply MD; clear MD. cancel.
  Qed.

  Notation SingleTF_specs:= (SingleDFA_ops_specs bool CHrep TrueFalseDFA).
  Definition SingleTF_internalspecs:funspecs:= 
     combine [_DFA_init; _DFA_final; _DFA_step] (map (fun x => x SingleGV) SingleTF_specs).

  Definition SingleTF_Gprog : funspecs := SingleTF_internalspecs ++ [].

  Lemma verif_init: semax_body singleVprog SingleTF_Gprog f_DFA_init (_DFA_init, single_init_spec _ TrueFalseDFA SingleGV).
  Proof. start_function. unfold SingleGV. forward. unfold SingleGV. entailer!. Qed.

  Lemma verif_final: semax_body singleVprog SingleTF_Gprog f_DFA_final (_DFA_final, single_final_spec _ TrueFalseDFA SingleGV).
  Proof. start_function. unfold SingleGV. forward. 1: destruct s; entailer!. forward. Qed.

  Lemma verif_step: semax_body singleVprog SingleTF_Gprog f_DFA_step (_DFA_step, single_step_spec _ CHrep TrueFalseDFA SingleGV).
  Proof. start_function. unfold SingleGV. forward. unfold SingleGV, Val.of_bool; entailer!.
     clear - H; destruct H.
     destruct d. 
     + specialize (H (eq_refl _)). inv H. apply derives_refl.
     + simpl. remember (Int.eq c (Int.repr 48)) as b; destruct b; simpl. 2: apply derives_refl.
       exfalso; symmetry in Heqb; apply int_eq_e in Heqb; intuition.
   Qed.

  Definition SingleTF_VSU: @VSU NullExtension.Espec nil nil singleQP 
             SingleTF_internalspecs
             (SingleGV false).
  Proof.
     mkVSU prog SingleTF_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - apply create_singleTF.
  Qed.

End Single_VSU.