Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import ObjectBase.
Require Import ASI_StaticDFA.
Require Import TrueFalseDFA.
Require Import ASI_StaticTrueFalse.

Require Import StaticTrueFalse.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition st_truefalseQP:= ltac:(QPprog prog).
Definition st_truefalseVprog := QPvarspecs st_truefalseQP.
Definition staticops := (Tstruct _staticOps noattr).

Notation val := (Values.val).

Section StaticTrueFalse_VSU.

Definition StaticTrueFalseGV (gv : globals) :=
           (EX shi shf shs (*sh*), !!(readable_share shi /\ readable_share shf /\ readable_share shs (*/\ readable_share sh*)) &&
                               (field_at shi staticops (DOT _st_init)  (gv _st_tf_init)  (gv _st_tf_ops) *
                                field_at shf staticops (DOT _st_final) (gv _st_tf_final) (gv _st_tf_ops) *
                                field_at shs staticops (DOT _st_step)  (gv _st_tf_step)  (gv _st_tf_ops)) *
                                data_at Ews tbool Vfalse (gv _truefalse))%logic.

  Lemma create_TrueFalse: VSU_initializer prog StaticTrueFalseGV.
  Proof. unfold VSU_initializer; intros gv GV. unfold Vardefs. unfold InitGPred, StaticTrueFalseGV. Exists Ews Ews Ews.
    simpl filter; unfold Maps.PTree.prev.
(*    assert (X: filter isGvar (Maps.PTree.elements (QP.prog_defs (QPprog prog))) =
            [ (_st_truefalse_ops, Gvar v_st_truefalse_ops); (_truefalse, Gvar v_truefalse)]) by  reflexivity.
    rewrite X; clear X. *)
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer. 
    erewrite <- (*4*)3 mapsto_field_at; try reflexivity; auto with field_compatible.
    assert (FC: field_compatible tbool [] (gv _truefalse)) by auto with field_compatible.
    assert (MD:forall p (Hp:field_compatible tbool [] p),
            mapsto_zeros 1 Ews p (*(gv 156%positive(*_truefalse*)*)
              |-- data_at Ews tbool Vfalse p(*(gv _truefalse)*)).
    { specialize (mapsto_zero_data_at_zero tbool Ews).
      rewrite zero_val_eq. simpl. simpl; intros MD p Hp.
      apply MD; clear MD; auto. }
    sep_apply MD; clear MD. cancel.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _st_tf_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _st_tf_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _st_tf_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  Definition StaticTrueFalseInv: GPred (Dfa_state TrueFalseDFA) :=
    fun '(d, gv, v) => (*optional: !!(v= gv _st_tf_ops) &&*) data_at Ews tbool (Val.of_bool d) (gv _truefalse).

  Definition static_truefalse_specs:= SDFA_ops_specs bool CHrep TrueFalseDFA.

  Definition StaticTrueFalseMT (i f s:val): val -> GPred (Dfa_state TrueFalseDFA) -> mpred :=
    MTable _ staticops [_st_init; _st_final; _st_step] static_truefalse_specs [i;f;s].

  Lemma make_StaticTrueFalseMT:
  forall sh1 sh2 sh3 i f s m, readable_share sh1 -> readable_share sh2 -> readable_share sh3 ->
  (func_ptr' (st_init_spec _ TrueFalseDFA StaticTrueFalseInv) i *
   func_ptr' (st_final_spec _ TrueFalseDFA StaticTrueFalseInv) f *
   func_ptr' (st_step_spec _ CHrep TrueFalseDFA StaticTrueFalseInv) s *
   field_at sh1 staticops [StructField _st_init] i m *
   field_at sh2 staticops [StructField _st_final] f m *
   field_at sh3 staticops [StructField _st_step] s m
  |-- StaticTrueFalseMT i f s m StaticTrueFalseInv)%logic.
  Proof. intros. unfold StaticTrueFalseMT. simpl. Exists sh1 i sh2 f sh3 s. entailer!. Qed.

  Definition StaticTrueFalseAPD : DFA_APD bool TrueFalseDFA :=
             {| DFA_GV := StaticTrueFalseGV; DFA_Inv := StaticTrueFalseInv |}.

  Notation mkStaticTFspec := (makeStaticTrueFalse_spec StaticTrueFalseAPD staticops _st_init _st_final _st_step CompSpecs).

  Definition StaticTrueFalse_internalspecs:funspecs:= [(_st_tf_init, st_init_spec _ _ StaticTrueFalseInv);
                                       (_st_tf_final, st_final_spec _ _ StaticTrueFalseInv);
                                       (_st_tf_step, st_step_spec _ CHrep _ StaticTrueFalseInv);
                                       (_make_StaticTrueFalse, mkStaticTFspec)].

  Definition StaticTrueFalse_Gprog : funspecs := StaticTrueFalse_internalspecs ++ [].

Opaque funcptrs.

  Lemma body_makeStaticTrueFalse: semax_body st_truefalseVprog StaticTrueFalse_internalspecs
        f_make_StaticTrueFalse  (_make_StaticTrueFalse,mkStaticTFspec).
  Proof.
  unfold mkStaticTFspec, makeStaticTrueFalse_spec, makeStaticDFA_spec. function_pointers.
  start_function.
  make_func_ptr _st_tf_init.
  make_func_ptr _st_tf_final.
  make_func_ptr _st_tf_step.
  unfold StaticTrueFalseAPD, StaticTrueFalseGV. simpl.
  Intros ish fsh ssh. rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh.
  sep_apply(make_StaticTrueFalseMT ish fsh ssh (gv _st_tf_init) (gv _st_tf_final) (gv _st_tf_step) (gv _st_tf_ops) Hish Hfsh Hssh).
  forward.
  Exists (gv _st_tf_ops). rewrite SDFA_fold_unfold. simpl.
    Exists [gv _st_tf_init; gv _st_tf_final; gv _st_tf_step].
    simpl.
    entailer!.
    unfold StaticTrueFalseMT, static_truefalse_specs, SDFA_ops_specs. cancel.
  Qed.

  Lemma verif_init_aux: semax_body st_truefalseVprog nil f_st_tf_init (_st_tf_init, st_init_spec _ _ StaticTrueFalseInv).
  Proof. start_function. unfold StaticTrueFalseInv; Intros. forward. unfold StaticTrueFalseInv. entailer!. Qed.

  Lemma verif_final_aux: semax_body st_truefalseVprog nil f_st_tf_final (_st_tf_final, st_final_spec _ _ (*TrueFalseAPD*)StaticTrueFalseInv).
  Proof.
     start_function. unfold StaticTrueFalseInv; Intros; subst.
     forward. 1: solve[ destruct s; entailer! ].
     forward; try (simpl; entailer!).
  Qed.

  Lemma verif_step_aux: semax_body st_truefalseVprog nil f_st_tf_step (_st_tf_step, st_step_spec _ CHrep _ (*TrueFalseAPD*)StaticTrueFalseInv).
  Proof. unfold st_step_spec.
     start_function. unfold StaticTrueFalseInv; Intros.
     forward. unfold StaticTrueFalseInv. 
     entailer!. unfold Val.of_bool. simpl. clear - H; destruct H.
     remember (Int.eq c (Int.repr 48)) as b; symmetry in Heqb; destruct b; simpl.
     + apply int_eq_e in Heqb; subst c. rewrite H0; trivial. apply derives_refl.
     + destruct d; [ | apply derives_refl].
       specialize (H (eq_refl _)); inv H. rewrite Int.eq_true in Heqb; discriminate.
  Qed.

  Lemma verif_init: semax_body st_truefalseVprog StaticTrueFalse_internalspecs f_st_tf_init (_st_tf_init, st_init_spec _ _ StaticTrueFalseInv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body st_truefalseVprog StaticTrueFalse_internalspecs f_st_tf_final (_st_tf_final, st_final_spec _ _ StaticTrueFalseInv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body st_truefalseVprog StaticTrueFalse_internalspecs f_st_tf_step (_st_tf_step, st_step_spec _ CHrep _ StaticTrueFalseInv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.
 
  Definition StaticTrueFalse_VSU: @VSU NullExtension.Espec nil nil st_truefalseQP 
             [(_make_StaticTrueFalse, mkStaticTFspec)]
             StaticTrueFalseGV.
  Proof.
     mkVSU prog StaticTrueFalse_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal body_makeStaticTrueFalse.
     - apply create_TrueFalse.
  Qed.
End StaticTrueFalse_VSU.