Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
(*Require Export ASI_surelyMalloc.*)
Require Import ObjectBase.
Require Import ASI_StaticDFA.
Require Import WeekdayDFA.
Require Import ASI_StaticWeekdays. 

Require Import StaticWeekdays.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition st_weekdaysQP:= ltac:(QPprog prog).
Definition st_weekdaysVprog := QPvarspecs st_weekdaysQP.

Definition staticops := (Tstruct _staticOps noattr).

Notation val := (Values.val).

Section StaticWeekdays_VSU.

Definition StaticWeekdaysGV (d: Dfa_state WeekdayDFA) (gv : globals) :=
           (EX shi shf shs, !!(readable_share shi /\ readable_share shf /\ readable_share shs) &&
                               (field_at shi staticops (DOT _st_init)  (gv _st_weekday_init)  (gv _st_weekday_ops) *
                                field_at shf staticops (DOT _st_final) (gv _st_weekday_final) (gv _st_weekday_ops) *
                                field_at shs staticops (DOT _st_step)  (gv _st_weekday_step)  (gv _st_weekday_ops)) *
                                data_at Ews tint (Vint (Int.repr (proj1_sig d))) (gv _day))%logic.

  Lemma create_StaticWeekdaysGV: VSU_initializer prog (StaticWeekdaysGV Monday).
  Proof. unfold VSU_initializer; intros gv GV.
    unfold Vardefs. unfold InitGPred, StaticWeekdaysGV. Exists Ews Ews Ews.
    simpl filter; unfold Maps.PTree.prev.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer.
    erewrite <- (*4*)3 mapsto_field_at; try reflexivity; auto with field_compatible.
    assert (FC: field_compatible tint [] (gv _day)) by auto with field_compatible.
    assert (MD:forall p (Hp:field_compatible tint [] p),
            mapsto_zeros 4 Ews p
              |-- data_at Ews tint (Vint (Int.repr 0)) p).
    { specialize (mapsto_zero_data_at_zero tint Ews).
      rewrite zero_val_tint. simpl; intros MD p Hp.
      apply MD; clear MD; auto. }
    sep_apply MD; clear MD. cancel.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _st_weekday_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _st_weekday_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _st_weekday_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  Definition StaticWeekdayInv: GPred (Dfa_state WeekdayDFA) :=
    fun '(d,gv,v) => (*optional: !!(v = gv _st_weekday_ops) && *)data_at Ews tint (Vint (Int.repr (proj1_sig d))) (gv _day).

  Definition static_weekday_specs:= SDFA_ops_specs ch CHrep WeekdayDFA.

  Definition StaticWeekdayMT (i f s:val): val -> GPred (Dfa_state WeekdayDFA) -> mpred :=
    MTable _ staticops [_st_init; _st_final; _st_step] static_weekday_specs [i;f;s]. 

  Lemma make_StaticWeekdayMT:
  forall sh1 sh2 sh3 i f s m, readable_share sh1 -> readable_share sh2 -> readable_share sh3 ->
  (func_ptr' (st_init_spec ch WeekdayDFA StaticWeekdayInv) i *
   func_ptr' (st_final_spec ch WeekdayDFA StaticWeekdayInv) f *
   func_ptr' (st_step_spec ch CHrep WeekdayDFA StaticWeekdayInv) s *
   field_at sh1 staticops [StructField _st_init] i m *
   field_at sh2 staticops [StructField _st_final] f m *
   field_at sh3 staticops [StructField _st_step] s m
  |-- StaticWeekdayMT i f s m StaticWeekdayInv)%logic.
  Proof. intros. unfold StaticWeekdayMT. simpl. Exists sh1 i sh2 f sh3 s. entailer!. Qed.

  Definition StaticWeekdayAPD:= Build_DFA_APD ch WeekdayDFA (StaticWeekdaysGV Monday) StaticWeekdayInv.

  Notation mkStaticWDspec := (makeStaticWeekdays_spec StaticWeekdayAPD staticops _st_init _st_final _st_step CompSpecs).

  Definition StaticWeekday_internalspecs:funspecs:= [(_st_weekday_init, st_init_spec ch _ StaticWeekdayInv);
                                       (_st_weekday_final, st_final_spec ch _ StaticWeekdayInv);
                                       (_st_weekday_step, st_step_spec ch CHrep _ StaticWeekdayInv);
                                       (_make_StaticWeekdays, mkStaticWDspec)].

  Definition StaticWeekday_Gprog : funspecs := StaticWeekday_internalspecs ++ [].

Opaque funcptrs.

  Lemma body_makeStaticWeekdays: semax_body st_weekdaysVprog StaticWeekday_internalspecs
        f_make_StaticWeekdays  (_make_StaticWeekdays,mkStaticWDspec).
  Proof.
  unfold mkStaticWDspec, makeStaticWeekdays_spec, makeStaticDFA_spec. function_pointers.
  start_function. (* rename H into HWW.*)
  make_func_ptr _st_weekday_init.
  make_func_ptr _st_weekday_final.
  make_func_ptr _st_weekday_step.
  unfold StaticWeekdayAPD, StaticWeekdaysGV. simpl.
  Intros ish fsh ssh. rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh. (*rename H2 into Hsh.*)
  sep_apply(make_StaticWeekdayMT ish fsh ssh (gv _st_weekday_init) (gv _st_weekday_final) (gv _st_weekday_step) (gv _st_weekday_ops) Hish Hfsh Hssh).
  forward.
  Exists (gv _st_weekday_ops). apply andp_right. entailer!.
  rewrite SDFA_fold_unfold. simpl.
    Exists [gv _st_weekday_init; gv _st_weekday_final; gv _st_weekday_step].
    simpl. entailer!.
    unfold StaticWeekdayMT, static_weekday_specs, SDFA_ops_specs. cancel. 
  Qed.

  Lemma verif_init_aux: semax_body st_weekdaysVprog nil f_st_weekday_init (_st_weekday_init, st_init_spec ch _ StaticWeekdayInv).
  Proof.
    start_function. unfold StaticWeekdayInv; Intros. forward. unfold StaticWeekdayInv. entailer!. Qed.

  Lemma verif_final_aux: semax_body st_weekdaysVprog nil f_st_weekday_final (_st_weekday_final, st_final_spec ch _ (*WeekdayAPD*)StaticWeekdayInv).
  Proof.
     start_function. unfold StaticWeekdayInv; Intros. forward. forward. simpl. entailer!. destruct s as [s HS].
     unfold Val.of_bool. simpl in *.
     destruct (Int.eq_dec (Int.repr s) (Int.repr 6)).
     + apply repr_inj_signed in e. subst s; trivial. rep_lia. rep_lia.
     + rewrite Int.eq_false; trivial. simpl.
       remember (s =? 6) as b; symmetry in Heqb; destruct b; trivial.
       apply Z.eqb_eq in Heqb; subst s. elim n; trivial.
  Qed.
  Lemma verif_step_aux: semax_body st_weekdaysVprog nil f_st_weekday_step (_st_weekday_step, st_step_spec ch CHrep _ (*WeekdayAPD*)StaticWeekdayInv).
  Proof. unfold st_step_spec.
     start_function. unfold StaticWeekdayInv; Intros. subst c.
     assert (Hn: Z.of_nat (Byte.to_nat "n" ) = 110) by reflexivity.
     forward_if.
     + forward. unfold StaticWeekdayInv; entailer!.
       apply derives_refl'. f_equal. f_equal. f_equal.
       remember (110 =? proj1_sig d) as b;symmetry in Heqb; destruct b; simpl; trivial.
       exfalso. apply Z.eqb_eq in Heqb. rewrite Heqb in H; congruence.
       destruct d; simpl in *; rewrite Heqb; trivial.
     + forward. destruct s as [s HS]; simpl.
       apply repr_inj_signed in H; [ simpl in H | destruct d; simpl; rep_lia | rep_lia]. rewrite H.
       forward_if; try (do 2 forward; unfold StaticWeekdayInv; simpl; rewrite H; simpl; entailer! ).
       rewrite 2 Int.unsigned_repr in * by rep_lia; lia. 
  Qed.

  Lemma verif_init: semax_body st_weekdaysVprog StaticWeekday_internalspecs f_st_weekday_init (_st_weekday_init, st_init_spec ch _ StaticWeekdayInv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body st_weekdaysVprog StaticWeekday_internalspecs f_st_weekday_final (_st_weekday_final, st_final_spec ch _ StaticWeekdayInv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body st_weekdaysVprog StaticWeekday_internalspecs f_st_weekday_step (_st_weekday_step, st_step_spec ch CHrep _ StaticWeekdayInv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.
 
  Definition StaticWeekday_VSU: @VSU NullExtension.Espec nil nil st_weekdaysQP 
             [(_make_StaticWeekdays, mkStaticWDspec)]
             (StaticWeekdaysGV Monday).
  Proof.
     mkVSU prog StaticWeekday_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal body_makeStaticWeekdays.
     - apply create_StaticWeekdaysGV.
  Qed.

End StaticWeekdays_VSU.