Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Export ASI_surelyMalloc.
Require Import ObjectBase.
Require Import ASI_DFA.
Require Import ASI_EDFA.
Require Import WeekdayDFA.
Require Import ASI_Weekdays. (*for CHrep -- maybe duplicate in ASI_T_Weekdays?*)
Require Import ASI_T_Weekdays.

Require Import T_Weekdays.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition tweekdaysQP:= ltac:(QPprog prog).
Definition tweekdaysVprog := QPvarspecs tweekdaysQP.
Notation object := (Tstruct _object noattr).
Notation tweekdays := (Tstruct _t_weekdays noattr).
Notation tdfaops := (Tstruct _t_dfaOps noattr).
Notation val := (Values.val).

Section TWeekdays_VSU.
  Variable M: MallocFreeAPD.

  Definition TWeekday_Inv: GPred (Dfa_state TWeekdayDFA) := fun x =>
   match x with (d, gv, v) =>
   withspacer Ews (sizeof size_t + sizeof tint) (2 * sizeof size_t)
     (fun u => (field_at Ews tweekdays [StructField _day] (Vint (Int.repr (proj1_sig d))) u )%logic) v
   end.

  Definition tweekday_specs:= EDFA_ops_specs object ch CHrep TWeekdayDFA.

  Definition TWeekdayMT (i f s t:val): val -> GPred (Dfa_state TWeekdayDFA) -> mpred :=
    MTable (Dfa_state TWeekdayDFA) (*M*) tdfaops [_init; _final; _step; _twosteps] tweekday_specs [i;f;s;t].

  Lemma TWeekdayMT_duplicable i f s t m P: TWeekdayMT i f s t m P |-- TWeekdayMT i f s t m P * TWeekdayMT i f s t m P.
  Proof. apply funcptrs_duplicable. Qed.

  Lemma make_TWeekdayMT:
  forall sh1 sh2 sh3 sh4 i f s t m,
    readable_share sh1 -> readable_share sh2 ->
    readable_share sh3 ->readable_share sh4 ->
  (func_ptr' (init_spec object ch TWeekdayDFA TWeekday_Inv) i *
   func_ptr' (final_spec object ch TWeekdayDFA TWeekday_Inv) f *
   func_ptr' (step_spec object ch CHrep TWeekdayDFA TWeekday_Inv) s *
   func_ptr' (twosteps_spec object ch CHrep TWeekdayDFA TWeekday_Inv) t *
   field_at sh1 tdfaops [StructField _init] i m *
   field_at sh2 tdfaops [StructField _final] f m *
   field_at sh3 tdfaops [StructField _step] s m *
   field_at sh4 tdfaops [StructField _twosteps] t m
   (*data_at Ews methods (g,s) m*)
  |-- TWeekdayMT i f s t m TWeekday_Inv)%logic.
  Proof. intros. unfold TWeekdayMT. simpl. Exists sh1 i sh2 f sh3 s sh4 t. entailer!. Qed.

  Lemma verif_init_aux: semax_body tweekdaysVprog nil f_t_weekday_init (_t_weekday_init, init_spec object ch _ TWeekday_Inv).
  Proof.
    start_function. unfold TWeekday_Inv; unfold withspacer; simpl; Intros (*msh. rename H into Msh; rename H0 into Hmsh*).
    forward. forward. (* simpl. unfold withspacer; simpl. Exists msh; entailer!. *)
  Qed. 
  Lemma verif_final_aux: semax_body tweekdaysVprog nil f_t_weekday_final (_t_weekday_final, final_spec object ch _ TWeekday_Inv).
  Proof.
    start_function. unfold TWeekday_Inv, withspacer; simpl; Intros (*msh. rename H into Msh; rename H0 into Hmsh*).
    forward. destruct s as [x X]. simpl. forward. entailer!.
    + unfold Val.of_bool.
      destruct (Int.eq_dec (Int.repr x) (Int.repr 6)).
      - apply repr_inj_signed in e. subst x; trivial. rep_lia. rep_lia.
      - rewrite Int.eq_false; trivial. simpl.
        remember (x =? 6) as b; symmetry in Heqb; destruct b; trivial.
        apply Z.eqb_eq in Heqb; subst x. elim n; trivial.
    (*+ unfold TWeekday_Inv, withspacer; simpl. Exists msh; entailer!.*)
  Qed.
  Lemma verif_step_aux: semax_body tweekdaysVprog nil f_t_weekday_step (_t_weekday_step, step_spec object ch CHrep _ TWeekday_Inv).
  Proof. unfold step_spec.
     start_function. rename H into Hcd.
     unfold TWeekday_Inv, withspacer; simpl; Intros(* msh. rename H into Msh; rename H0 into Hmsh*).
     assert (Hn: Z.of_nat (Byte.to_nat "n" ) = 110) by reflexivity.
     forward_if.
     + forward. unfold TWeekday_Inv, withspacer; simpl. (*Exists msh; *)entailer!.
       clear (*Msh Hmsh*) H1 H3 (*H5*).
       apply derives_refl'. f_equal. f_equal. f_equal.
       apply negb_true_iff in H. simpl in H.
       remember (110 =? proj1_sig d) as b;symmetry in Heqb; destruct b; simpl; trivial.
       - exfalso. apply Z.eqb_eq in Heqb. rewrite <- Heqb, Int.eq_true in H. congruence.
       - simpl in Hn. destruct (proj1_sig d); trivial.
         clear - Heqb. rewrite Pos2Z.inj_eqb, Heqb; trivial.
     + forward. destruct s as [s HS]; simpl. forward_if;
       try (do 2 forward; apply negb_false_iff in H; apply int_eq_e in H;
         apply repr_inj_signed in H; [ simpl in H | destruct d; simpl; rep_lia | rep_lia];
         unfold TWeekday_Inv, withspacer; simpl; rewrite H; (*Exists msh;*) entailer!).
       rewrite 2 Int.unsigned_repr in * by rep_lia; lia. 
  Qed.

  Definition TWeekdayGV (gv:globals): mpred :=
     EX sh1 sh2 sh3 sh4, !!(readable_share sh1 /\ readable_share sh2 /\ readable_share sh3 /\ readable_share sh4) &&
                           (field_at sh1 tdfaops (DOT _init)  (gv _t_weekday_init)  (gv _t_weekdayops) *
                           field_at sh2 tdfaops (DOT _final) (gv _t_weekday_final) (gv _t_weekdayops) *
                           field_at sh3 tdfaops (DOT _step)  (gv _t_weekday_step)  (gv _t_weekdayops) *
                           field_at sh4 tdfaops (DOT _twosteps)  (gv _t_weekday_twosteps)  (gv _t_weekdayops)).

  Lemma TWeekday_Inv_isptrnull a v: TWeekday_Inv (a, v) |-- !! is_pointer_or_null v.
  Proof. unfold TWeekday_Inv, withspacer; destruct a; simpl; entailer!. Qed.

  Definition TWeekdayAPD:= Build_EDFA_APD ch TWeekdayDFA TWeekdayGV TWeekday_Inv (*TWeekday_Inv_isptrnull*).

  Notation mkTWDspec := (make_TWeekdays_spec TWeekdayAPD M object tdfaops _ops _init _final _step _twosteps CompSpecs).

  Definition TWeekday_internalspecs := [(_t_weekday_init, init_spec object ch _ TWeekday_Inv);
                                       (_t_weekday_final, final_spec object ch _ TWeekday_Inv);
                                       (_t_weekday_step, step_spec object ch CHrep _ TWeekday_Inv);
                                       (_t_weekday_twosteps, twosteps_spec object ch CHrep TWeekdayDFA TWeekday_Inv);
                                       (_make_TWeekdays, mkTWDspec)].

  Definition TWeekday_Gprog : funspecs := TWeekday_internalspecs ++ [(_surely_malloc, surely_malloc_numericspec M)].
  (* SM_ASI free surelyMalloc.*)

  Lemma align_compatible_tweekdays_object {b i} (AL: align_compatible tweekdays (Vptr b i)):
        align_compatible object (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

  Lemma field_compatible_tweekdays_object_ops p: field_compatible tweekdays [StructField _ops] p ->
        field_compatible object [StructField _ops] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tweekdays_object; trivial.
    - split; simpl; auto. left; trivial.
  Qed.

Opaque funcptrs.

  Lemma body_makeTWeekdays: semax_body tweekdaysVprog TWeekday_Gprog f_make_TWeekdays  (_make_TWeekdays,mkTWDspec).
  Proof.
  unfold mkTWDspec, make_TWeekdays_spec, makeEDFA_spec, TWeekdayGV. function_pointers.
  start_function. rename H into HWW.
  forward_call surelymalloc_spec_sub (tweekdays, gv). { simpl. rep_lia. }
  Intros s.
  make_func_ptr _t_weekday_init.
  make_func_ptr _t_weekday_final.
  make_func_ptr _t_weekday_step.
  make_func_ptr _t_weekday_twosteps.
  unfold TWeekdayAPD, TWeekdayGV. simpl.  Intros ish fsh ssh tsh.
  rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh. rename H2 into Htsh.
 
  sep_apply(make_TWeekdayMT ish fsh ssh tsh (gv _t_weekday_init) (gv _t_weekday_final)
                             (gv _t_weekday_step) (gv _t_weekday_twosteps) (gv _t_weekdayops) Hish Hfsh Hssh Htsh).
  intros.
  forward. forward.
  sep_apply TWeekdayMT_duplicable. forward.
  Exists s. entailer!.
  rewrite 2 sepcon_assoc. apply sepcon_derives.
  { unfold EDFA_GV. simpl. unfold TWeekdayMT. simpl.
    Transparent funcptrs. unfold MTable. Opaque funcptrs. simpl.
    Intros Ish gi Fsh gf Ssh gs Tsh gt. Exists Ish Fsh Ssh Tsh. entailer!.
    do 4 sep_apply func_ptr'_emp. cancel.
    apply JMeq_eq in H2. subst gi. 
    apply JMeq_eq in H4. subst gf. 
    apply JMeq_eq in H6. subst gs. 
    apply JMeq_eq in H8. subst gt. cancel. }
  rewrite EDFA_fold_unfold. 
  Exists [gv _t_weekday_init; gv _t_weekday_final; gv _t_weekday_step; gv _t_weekday_twosteps].
    Exists (CompSpecs, tweekdays) (*Weekday_Inv*) (gv _t_weekdayops). simpl.
    entailer!.
    rewrite sepcon_comm. apply sepcon_derives. 2: apply derives_refl.
    unfold withspacer; simpl.
    unfold_data_at (data_at Ews tweekdays _ s). cancel.
    unfold FldAt, RdblFldAt.

    (*rewrite <- (field_at_share_join (fst (slice.cleave Ews)) (snd (slice.cleave Ews))
                                  Ews tweekdays [ StructField _ops])
    by apply slice.cleave_join.
    Exists (snd (slice.cleave Ews)) (gv _t_weekdayops) (fst (slice.cleave Ews)). entailer!.
    { split3. apply slice.cleave_readable2; auto. apply sepalg.join_sub_refl.
      split. apply slice.cleave_readable1; auto. apply sepalg.join_sub_refl. }*)
    Exists  (gv _t_weekdayops); entailer!.

    apply field_compatible_tweekdays_object_ops in H2.
    unfold field_at. simpl; entailer!. apply derives_refl.
  Qed.

  Lemma verif_twosteps: semax_body tweekdaysVprog TWeekday_Gprog f_t_weekday_twosteps (_t_weekday_twosteps, twosteps_spec object ch CHrep _ TWeekday_Inv).
  Proof. unfold twosteps_spec.
    start_function.
    assert (Hd1: Int.zero_ext 8 (Int.repr (proj1_sig d1)) = Int.repr (proj1_sig d1)).
    { clear. destruct d1 as [d1 D1]; simpl in D1. rewrite zero_ext_inrange. trivial.
      rewrite Int.unsigned_repr; simpl; rep_lia. }
    assert (Hd2: Int.zero_ext 8 (Int.repr (proj1_sig d2)) = Int.repr (proj1_sig d2)).
    { clear. destruct d2 as [d2 D2]; simpl in D2. rewrite zero_ext_inrange. trivial.
      rewrite Int.unsigned_repr; simpl; rep_lia. }
    forward_call (s, gv, v, c1, d1).
    { entailer!. simpl. rewrite Hd1; trivial. } 
    forward_call (Dfa_trans TWeekdayDFA s d1, gv, v, c2, d2).
    { entailer!. simpl. rewrite Hd2; trivial. }
    forward.
  Qed.

  Lemma verif_init: semax_body tweekdaysVprog TWeekday_Gprog f_t_weekday_init (_t_weekday_init, init_spec object ch _ TWeekday_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body tweekdaysVprog TWeekday_Gprog f_t_weekday_final (_t_weekday_final, final_spec object ch _ TWeekday_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body tweekdaysVprog TWeekday_Gprog f_t_weekday_step (_t_weekday_step, step_spec object ch CHrep _ TWeekday_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.

  Lemma TWeekday_GP gv (GV: globals_ok gv):
        InitGPred (Vardefs tweekdaysQP) gv |-- TWeekdayGV gv.
  Proof. unfold Vardefs. unfold InitGPred, TWeekdayGV. Exists Ews Ews Ews Ews.
    assert (X: filter isGvar (Maps.PTree.elements (QP.prog_defs tweekdaysQP)) = [(_t_weekdayops, Gvar v_t_weekdayops)]) by reflexivity.
    rewrite X; clear X.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer.
    erewrite <- 4 mapsto_field_at; try reflexivity; auto with field_compatible.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives; [ | apply sepcon_derives]]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _t_weekday_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _t_weekday_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _t_weekday_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _t_weekday_twosteps) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  Definition TWeekday_VSU: @VSU NullExtension.Espec nil [(_surely_malloc, surely_malloc_numericspec M)] tweekdaysQP 
                                [(_make_TWeekdays,mkTWDspec)] TWeekdayGV.
  Proof.
     mkVSU prog TWeekday_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal verif_twosteps.
     - solve_SF_internal body_makeTWeekdays.
     - apply TWeekday_GP.
  Qed.

(*DFA predicate using tdfaops: works but is perhaps not ideal*)
Notation TwoDFA_t := (@DFA ch CHrep TWeekdayDFA (EApd2DApd TWeekdayAPD) M object tdfaops _ops _init _final _step CompSpecs).

Definition makeTWeekday_asDFA_spec_t: funspec :=
 WITH gv: globals
 PRE [ ]
    PROP (WA+WORD < 100) PARAMS() GLOBALS (gv) (*The condition WA+WORD  100 is true for 64 and 32 bit (and conservative...)*)
    SEP (mem_mgr M gv; EDFA_GV _ TWeekdayAPD gv)
 POST [ tptr object ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; EDFA_GV _ TWeekdayAPD gv;
          TwoDFA_t (Dfa_s TWeekdayDFA, gv, p)).

Lemma makeTwoWeekday_spec_subsumption_t: funspec_sub mkTWDspec makeTWeekday_asDFA_spec_t.
Proof. 
  do_funspec_sub. Exists w emp. entailer!.
  intros. Exists (eval_id ret_temp x); entailer!.
  sep_apply EDFA2DFA. apply derives_refl.
Qed. 

(*Version of DFA that uses dfaOps *)
Notation dfaops := (Tstruct _dfaOps noattr).
Notation TwoDFA := (@DFA ch CHrep TWeekdayDFA (EApd2DApd TWeekdayAPD) M object dfaops _ops _init _final _step CompSpecs).

Definition makeTWeekday_asDFA_spec: funspec :=
 WITH gv: globals
 PRE [ ]
    PROP (WA+WORD < 100) PARAMS() GLOBALS (gv) (*The condition WA+WORD  100 is true for 64 and 32 bit (and conservative...)*)
    SEP (mem_mgr M gv; EDFA_GV _ TWeekdayAPD gv)
 POST [ tptr object ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; EDFA_GV _ TWeekdayAPD gv;
          TwoDFA (Dfa_s TWeekdayDFA, gv, p)).

  Lemma align_compatible_tdfaops_dfaops {b i} (AL: align_compatible tdfaops (Vptr b i)):
        align_compatible dfaops (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

Lemma FC_step m (FC: field_compatible tdfaops (DOT _step) m):field_compatible dfaops (DOT _step) m.
Proof. intros.
    destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
    destruct m; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tdfaops_dfaops; trivial.
    - split; simpl; auto. do 2 right; left; trivial.
  Qed.
Lemma FC_init m (FC: field_compatible tdfaops (DOT _step) m):field_compatible dfaops (DOT _init) m. 
Proof. intros.
    destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
    destruct m; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tdfaops_dfaops; trivial.
    - split; simpl; auto. left; trivial.
  Qed.
Lemma FC_final m (FC: field_compatible tdfaops (DOT _step) m):field_compatible dfaops (DOT _final) m.
Proof. intros.
    destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
    destruct m; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tdfaops_dfaops; trivial.
    - split; simpl; auto. right; left; trivial.
  Qed.

Lemma DFA_tdfaops_dfaops ch CHrep A APD x:
      DFA ch CHrep A APD M object tdfaops _ops _init _final _step CompSpecs x
  |-- DFA ch CHrep A APD M object dfaops _ops _init _final _step CompSpecs x.
Proof.
  rewrite 2 DFA_fold_unfold. Intros fptrs cso m. Exists fptrs cso m. cancel.
  entailer!.
Transparent funcptrs. simpl. Opaque funcptrs.
destruct fptrs; [trivial | Intros ish iv]. Exists ish iv. entailer!.
destruct fptrs; [ eapply derives_trans; [ apply sepcon_FF_derives'; trivial | apply FF_left] | Intros fsh fv; Exists fsh fv; entailer!].
destruct fptrs; [ eapply derives_trans; [ apply sepcon_FF_derives'; trivial | apply FF_left] | Intros ssh sv; Exists ssh sv; entailer!].
unfold field_at; simpl; entailer!. split3.
apply FC_init; trivial.
apply FC_step; trivial.
apply FC_final; trivial.
Qed.

Lemma makeTwoWeekday_spec_subsumption: funspec_sub mkTWDspec makeTWeekday_asDFA_spec.
Proof. 
  do_funspec_sub. Exists w emp. entailer!.
  intros. Exists (eval_id ret_temp x); entailer!.
  sep_apply EDFA2DFA. simpl. apply DFA_tdfaops_dfaops.
Qed.

End TWeekdays_VSU.