Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Export ASI_surelyMalloc.
Require Import ObjectBase.
Require Import ASI_DFA.
Require Import TrueFalseDFA.
Require Import ASI_TrueFalse.

Require Import TrueFalse.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition truefalseQP:= ltac:(QPprog prog).
Definition truefalseVprog := QPvarspecs truefalseQP.
Notation object := (Tstruct _object noattr).
Notation tfrep := (Tstruct _tf_rep noattr).
Notation dfaops := (Tstruct _dfaOps noattr).

Section TrueFalse_VSU.
  Variable M: MallocFreeAPD.

  Definition TrueFalse_Inv: GPred (Dfa_state TrueFalseDFA) := fun x =>
   match x with (d, gv, v) =>
   withspacer Ews (sizeof size_t + sizeof tbool) (2 * sizeof size_t)
     (field_at Ews tfrep [StructField _tf] (Val.of_bool d)) v
   end.

  Definition tf_specs:= DFA_ops_specs object bool CHrep TrueFalseDFA.

  Definition TrueFalseMT (i f s:val): val -> GPred (Dfa_state TrueFalseDFA) -> mpred :=
    MTable (Dfa_state TrueFalseDFA) dfaops [_init; _final; _step] tf_specs [i;f;s].

  Lemma TrueFalseMT_duplicable i f s m P: TrueFalseMT i f s m P |-- TrueFalseMT i f s m P * TrueFalseMT i f s m P.
  Proof. apply funcptrs_duplicable. Qed.

  Lemma make_TrueFalseMT:
  forall sh1 sh2 sh3 i f s m, readable_share sh1 -> readable_share sh2 -> readable_share sh3 ->
  (func_ptr' (init_spec object bool TrueFalseDFA TrueFalse_Inv) i *
   func_ptr' (final_spec object bool TrueFalseDFA TrueFalse_Inv) f *
   func_ptr' (step_spec object bool CHrep TrueFalseDFA TrueFalse_Inv) s *
   field_at sh1 dfaops [StructField _init] i m *
   field_at sh2 dfaops [StructField _final] f m *
   field_at sh3 dfaops [StructField _step] s m
  |-- TrueFalseMT i f s m TrueFalse_Inv)%logic.
  Proof. intros. unfold TrueFalseMT. simpl. Exists sh1 i sh2 f sh3 s. entailer!. Qed.

  Lemma verif_init_aux: semax_body truefalseVprog nil f_tf_init (_tf_init, init_spec object bool _ TrueFalse_Inv).
  Proof.
    start_function. unfold TrueFalse_Inv; unfold withspacer; simpl; Intros.
    forward. entailer!; simpl. unfold withspacer; simpl. cancel.
  Qed. 
  Lemma verif_final_aux: semax_body truefalseVprog nil f_tf_final (_tf_final, final_spec object bool _ TrueFalse_Inv).
  Proof.
    start_function. unfold TrueFalse_Inv, withspacer; simpl; Intros.
    forward. destruct s; entailer!. forward.
  Qed.
  Lemma verif_step_aux: semax_body truefalseVprog nil f_tf_step (_tf_step, step_spec object bool CHrep _ TrueFalse_Inv).
  Proof. unfold step_spec.
     start_function. rename H into Hcd.
     unfold TrueFalse_Inv, withspacer; simpl; Intros.
     forward. entailer!. clear - Hcd. simpl. unfold withspacer, Val.of_bool; simpl; cancel.
     apply derives_refl'. f_equal. destruct Hcd.
     remember (Int.eq c (Int.repr 48)) as b; symmetry in Heqb; destruct b; simpl.
     + apply int_eq_e in Heqb; subst c. rewrite H0; trivial.
     + destruct d; [ | reflexivity].
       specialize (H (eq_refl _)); inv H. rewrite Int.eq_true in Heqb; discriminate.
  Qed.

  Definition TrueFalseGV (gv:globals): mpred :=
           EX sh1 sh2 sh3, !!(readable_share sh1 /\ readable_share sh2 /\ readable_share sh3) &&
                               (field_at sh1 dfaops (DOT _init)  (gv _tf_init)  (gv _tfops) *
                                field_at sh2 dfaops (DOT _final) (gv _tf_final) (gv _tfops) *
                                field_at sh3 dfaops (DOT _step)  (gv _tf_step)  (gv _tfops)).

  Lemma TrueFalse_Inv_isptrnull a v: TrueFalse_Inv (a, v) |-- !! is_pointer_or_null v.
  Proof. unfold TrueFalse_Inv, withspacer; destruct a; simpl; entailer!. Qed.

  Definition TrueFalseAPD:= Build_DFA_APD bool TrueFalseDFA TrueFalseGV TrueFalse_Inv. (*TrueFalse_Inv_isptrnull*)

  Notation mkTFspec := (makeTrueFalse_spec TrueFalseAPD M object dfaops _ops _init _final _step CompSpecs).

  Definition TrueFalse_internalspecs := [(_tf_init, init_spec object bool _ TrueFalse_Inv);
                                       (_tf_final, final_spec object bool _ TrueFalse_Inv);
                                       (_tf_step, step_spec object bool CHrep _ TrueFalse_Inv);
                                       (_make_TrueFalse, mkTFspec)]. 

  Definition TrueFalse_Gprog : funspecs := TrueFalse_internalspecs ++ [(_surely_malloc, surely_malloc_numericspec M)].
  (* SM_ASI free surelyMalloc.*)

  Lemma align_compatible_tfrep {b i} (AL: align_compatible tfrep (Vptr b i)):
        align_compatible object (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

  Lemma field_compatible_tfrep_ops p: field_compatible tfrep [StructField _ops] p ->
        field_compatible object [StructField _ops] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_tfrep; trivial.
    - split; simpl; auto. left; trivial.
  Qed.

Opaque funcptrs.

  Lemma body_makeTrueFalse: semax_body truefalseVprog TrueFalse_Gprog f_make_TrueFalse  (_make_TrueFalse,mkTFspec).
  Proof.
  unfold mkTFspec, makeTrueFalse_spec, makeDFA_spec, TrueFalseGV. function_pointers.
  start_function. rename H into HWW.
  forward_call surelymalloc_spec_sub (tfrep, gv). { simpl. rep_lia. }
  Intros s.
  make_func_ptr _tf_init.
  make_func_ptr _tf_final.
  make_func_ptr _tf_step.
  unfold TrueFalseAPD, TrueFalseGV. simpl. Intros ish fsh ssh. rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh.
 
  sep_apply(make_TrueFalseMT ish fsh ssh (gv _tf_init) (gv _tf_final) (gv _tf_step) (gv _tfops) Hish Hfsh Hssh).
  forward. forward.
  sep_apply TrueFalseMT_duplicable. forward.
  Exists s. entailer!.
  rewrite 2 sepcon_assoc. apply sepcon_derives.
  { unfold DFA_GV. simpl. unfold TrueFalseMT. simpl.
    Transparent funcptrs. unfold MTable. Opaque funcptrs. simpl.
    Intros Ish gi Fsh gf Ssh gs. Exists Ish Fsh Ssh. entailer!.
    do 3 sep_apply func_ptr'_emp. cancel.
    apply JMeq_eq in H2. subst gi. 
    apply JMeq_eq in H4. subst gf. 
    apply JMeq_eq in H6. subst gs. cancel. }
  rewrite DFA_fold_unfold. 
  Exists [gv _tf_init; gv _tf_final; gv _tf_step].
    Exists (CompSpecs, tfrep) (gv _tfops). simpl.
    entailer!.
    rewrite sepcon_comm. apply sepcon_derives. 2: apply derives_refl.
    unfold withspacer; simpl.
    unfold_data_at (data_at Ews tfrep _ s). cancel.
    unfold FldAt. simpl. Exists (gv _tfops); entailer!.
   apply field_compatible_tfrep_ops in H2.
    unfold field_at. simpl; entailer!.
  Qed.

  Lemma verif_init: semax_body truefalseVprog TrueFalse_Gprog f_tf_init (_tf_init, init_spec object bool _ TrueFalse_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body truefalseVprog TrueFalse_Gprog f_tf_final (_tf_final, final_spec object bool _ TrueFalse_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body truefalseVprog TrueFalse_Gprog f_tf_step (_tf_step, step_spec object bool CHrep _ TrueFalse_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.

  Lemma TrueFalse_GP gv (GV: globals_ok gv):
        InitGPred (Vardefs truefalseQP) gv |-- TrueFalseGV gv.
  Proof. unfold Vardefs. unfold InitGPred, TrueFalseGV. Exists Ews Ews Ews.
    assert (X: filter isGvar (Maps.PTree.elements (QP.prog_defs truefalseQP)) = [(_tfops, Gvar v_tfops)]) by reflexivity.
    rewrite X; clear X.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer.
    erewrite <- 3 mapsto_field_at; try reflexivity; auto with field_compatible.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _tf_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _tf_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _tf_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  (*For now, we only export the constructor. If later "subclasses" inherit init, final, or step, well need to export these here, too*)
  Definition TrueFalse_VSU: @VSU NullExtension.Espec nil [(_surely_malloc, surely_malloc_numericspec M)] truefalseQP 
             [ (_make_TrueFalse,mkTFspec)]
             TrueFalseGV.
  Proof.
     mkVSU prog TrueFalse_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal body_makeTrueFalse.
     - apply TrueFalse_GP.
  Qed.

End TrueFalse_VSU.