Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Export ASI_surelyMalloc.
Require Import ObjectBase.
Require Import ASI_DFA.
Require Import WeekdayDFA.
Require Import ASI_Weekdays.

Require Import Weekdays.

Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition weekdaysQP:= ltac:(QPprog prog).
Definition weekdaysVprog := QPvarspecs weekdaysQP.
Notation object := (Tstruct _object noattr).
Notation weekdays := (Tstruct _weekdays noattr).
Notation dfaops := (Tstruct _dfaOps noattr).

Section Weekdays_VSU.
  Variable M: MallocFreeAPD.

  Definition Weekday_Inv: GPred (Dfa_state WeekdayDFA) := fun x =>
   match x with (d, gv, v) =>
   withspacer Ews (sizeof size_t + sizeof tint) (2 * sizeof size_t)
     (field_at Ews weekdays [StructField _day] (Vint (Int.repr (proj1_sig d)))) v
   end.

  Definition weekday_specs:= DFA_ops_specs object ch CHrep WeekdayDFA.

  Definition WeekdayMT (i f s:val): val -> GPred (Dfa_state WeekdayDFA) -> mpred :=
    MTable (Dfa_state WeekdayDFA) dfaops [_init; _final; _step] weekday_specs [i;f;s].

  Lemma WeekdayMT_duplicable i f s m P: WeekdayMT i f s m P |-- WeekdayMT i f s m P * WeekdayMT i f s m P.
  Proof. apply funcptrs_duplicable. Qed.

  Lemma make_WeekdayMT:
  forall sh1 sh2 sh3 i f s m, readable_share sh1 -> readable_share sh2 -> readable_share sh3 ->
  (func_ptr' (init_spec object ch WeekdayDFA Weekday_Inv) i *
   func_ptr' (final_spec object ch WeekdayDFA Weekday_Inv) f *
   func_ptr' (step_spec object ch CHrep WeekdayDFA Weekday_Inv) s *
   field_at sh1 dfaops [StructField _init] i m *
   field_at sh2 dfaops [StructField _final] f m *
   field_at sh3 dfaops [StructField _step] s m
  |-- WeekdayMT i f s m Weekday_Inv)%logic.
  Proof. intros. unfold WeekdayMT. simpl. Exists sh1 i sh2 f sh3 s. entailer!. Qed.

  Lemma verif_init_aux: semax_body weekdaysVprog nil f_weekday_init (_weekday_init, init_spec object ch _ Weekday_Inv).
  Proof.
    start_function. unfold Weekday_Inv; unfold withspacer; simpl; Intros.
    forward. forward. 
  Qed. 
  Lemma verif_final_aux: semax_body weekdaysVprog nil f_weekday_final (_weekday_final, final_spec object ch _ (*WeekdayAPD*)Weekday_Inv).
  Proof.
    start_function. unfold Weekday_Inv, withspacer; simpl; Intros.
    forward. destruct s as [x X]. simpl. forward. entailer!.
    unfold Val.of_bool.
      destruct (Int.eq_dec (Int.repr x) (Int.repr 6)).
      - apply repr_inj_signed in e. subst x; trivial. rep_lia. rep_lia.
      - rewrite Int.eq_false; trivial. simpl.
        remember (x =? 6) as b; symmetry in Heqb; destruct b; trivial.
        apply Z.eqb_eq in Heqb; subst x. elim n; trivial.
  Qed.
  Lemma verif_step_aux: semax_body weekdaysVprog nil f_weekday_step (_weekday_step, step_spec object ch CHrep _ (*WeekdayAPD*)Weekday_Inv).
  Proof. unfold step_spec.
     start_function. rename H into Hcd.
     unfold Weekday_Inv, withspacer; simpl; Intros.
     assert (Hn: Z.of_nat (Byte.to_nat "n" ) = 110) by reflexivity.
     forward_if.
     + forward. unfold Weekday_Inv, withspacer; simpl. entailer!.
       clear H1 H3.
       apply derives_refl'. f_equal. f_equal. f_equal.
       apply negb_true_iff in H. simpl in H.
       remember (110 =? proj1_sig d) as b;symmetry in Heqb; destruct b; simpl; trivial.
       - exfalso. apply Z.eqb_eq in Heqb. rewrite <- Heqb, Int.eq_true in H. congruence.
       - simpl in Hn. destruct (proj1_sig d); trivial.
         clear - Heqb. rewrite Pos2Z.inj_eqb, Heqb; trivial.
     + forward. destruct s as [s HS]; simpl. forward_if;
       try (do 2 forward; apply negb_false_iff in H; apply int_eq_e in H;
         apply repr_inj_signed in H; [ simpl in H | destruct d; simpl; rep_lia | rep_lia];
         unfold Weekday_Inv, withspacer; simpl; rewrite H; entailer!).
       rewrite 2 Int.unsigned_repr in * by rep_lia; lia. 
  Qed. 

  Definition WeekdayGV (gv:globals): mpred :=
           EX sh1 sh2 sh3, !!(readable_share sh1 /\ readable_share sh2 /\ readable_share sh3) &&
                               (field_at sh1 dfaops (DOT _init)  (gv _weekday_init)  (gv _weekdayops) *
                                field_at sh2 dfaops (DOT _final) (gv _weekday_final) (gv _weekdayops) *
                                field_at sh3 dfaops (DOT _step)  (gv _weekday_step)  (gv _weekdayops)).

  Lemma Weekday_Inv_isptrnull a v: Weekday_Inv (a, v) |-- !! is_pointer_or_null v.
  Proof. unfold Weekday_Inv, withspacer; destruct a; simpl; entailer!. Qed.

  Definition WeekdayAPD:= Build_DFA_APD ch WeekdayDFA WeekdayGV Weekday_Inv. (*Weekday_Inv_isptrnull*)

  Notation mkWDspec := (makeWeekdays_spec WeekdayAPD M object dfaops _ops _init _final _step CompSpecs).

  Definition Weekday_internalspecs := [(_weekday_init, init_spec object ch _ Weekday_Inv);
                                       (_weekday_final, final_spec object ch _ Weekday_Inv);
                                       (_weekday_step, step_spec object ch CHrep _ Weekday_Inv);
                                       (_make_Weekdays, mkWDspec)]. 

  Definition Weekday_Gprog : funspecs := Weekday_internalspecs ++ [(_surely_malloc, surely_malloc_numericspec M)].
  (* SM_ASI free surelyMalloc.*)

  Lemma align_compatible_weekdays_object {b i} (AL: align_compatible weekdays (Vptr b i)):
        align_compatible object (Vptr b i).
  Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity | reflexivity |].
      simpl co_members in *; intros. specialize (H4 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H4 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H4. inv H. apply H0. }
      clear H1. congruence.
  Qed.

  Lemma field_compatible_weekdays_object_ops p: field_compatible weekdays [StructField _ops] p ->
        field_compatible object [StructField _ops] p.
  Proof. intros.
    destruct H as [Pt [COMPL [SZ [AL LNF]]]].
    destruct p; inv Pt. split3; try reflexivity.
    split3.
    - red; simpl. red in SZ; simpl in SZ. lia.
    - clear - AL. apply align_compatible_weekdays_object; trivial.
    - split; simpl; auto. left; trivial.
  Qed.

Opaque funcptrs.

  Lemma body_makeWeekdays: semax_body weekdaysVprog Weekday_Gprog f_make_Weekdays  (_make_Weekdays,mkWDspec).
  Proof.
  unfold mkWDspec, makeWeekdays_spec, makeDFA_spec, WeekdayGV. function_pointers.
  start_function. rename H into HWW.
  forward_call surelymalloc_spec_sub (weekdays, gv). { simpl. rep_lia. }
  Intros s.
  make_func_ptr _weekday_init.
  make_func_ptr _weekday_final.
  make_func_ptr _weekday_step.
  unfold WeekdayAPD, WeekdayGV. simpl.  Intros ish fsh ssh. rename H into Hish. rename H0 into Hfsh. rename H1 into Hssh.
 
  sep_apply(make_WeekdayMT ish fsh ssh (gv _weekday_init) (gv _weekday_final) (gv _weekday_step) (gv _weekdayops) Hish Hfsh Hssh).
  forward. forward.
  sep_apply WeekdayMT_duplicable. forward.
  Exists s. entailer!.
  rewrite 2 sepcon_assoc. apply sepcon_derives.
  { unfold DFA_GV. simpl. unfold WeekdayMT. simpl.
    Transparent funcptrs. unfold MTable. Opaque funcptrs. simpl.
    Intros Ish gi Fsh gf Ssh gs. Exists Ish Fsh Ssh. entailer!.
    do 3 sep_apply func_ptr'_emp. cancel.
    apply JMeq_eq in H2. subst gi. 
    apply JMeq_eq in H4. subst gf. 
    apply JMeq_eq in H6. subst gs. cancel. }
  rewrite DFA_fold_unfold. 
  Exists [gv _weekday_init; gv _weekday_final; gv _weekday_step].
    Exists (CompSpecs, weekdays) (gv _weekdayops). simpl.
    entailer!.
    rewrite sepcon_comm. apply sepcon_derives. 2: apply derives_refl.
    unfold withspacer; simpl.
    unfold_data_at (data_at Ews weekdays _ s). cancel.
    unfold FldAt. simpl. Exists (gv _weekdayops); entailer!.
   apply field_compatible_weekdays_object_ops in H2.
    unfold field_at. simpl; entailer!.
  Qed.

  Lemma verif_init: semax_body weekdaysVprog Weekday_Gprog f_weekday_init (_weekday_init, init_spec object ch _ Weekday_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_init_aux. LNR_tac. Qed.

  Lemma verif_final: semax_body weekdaysVprog Weekday_Gprog f_weekday_final (_weekday_final, final_spec object ch _ Weekday_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_final_aux. LNR_tac. Qed.

  Lemma verif_step: semax_body weekdaysVprog Weekday_Gprog f_weekday_step (_weekday_step, step_spec object ch CHrep _ Weekday_Inv).
  Proof. apply semax_body_subsumespec_GprogNil. apply verif_step_aux. LNR_tac. Qed.

  Lemma Weekday_GP gv (GV: globals_ok gv):
        InitGPred (Vardefs weekdaysQP) gv |-- WeekdayGV gv.
  Proof. unfold Vardefs. unfold InitGPred, WeekdayGV. Exists Ews Ews Ews.
    assert (X: filter isGvar (Maps.PTree.elements (QP.prog_defs weekdaysQP)) = [(_weekdayops, Gvar v_weekdayops)]) by reflexivity.
    rewrite X; clear X.
    simpl; Intros. rewrite sepcon_emp.
    unfold globvar2pred; simpl; entailer.
    erewrite <- 3 mapsto_field_at; try reflexivity; auto with field_compatible.
    rewrite ! sepcon_assoc. 
    apply sepcon_derives; [ | apply sepcon_derives]; simpl.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _weekday_init) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _weekday_final) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
    - unfold field_address. simpl. rewrite if_true by (auto with field_compatible).
      destruct (GV  _weekday_step) as [G | G].
      + destruct G as [g G]; rewrite G. entailer.
      + rewrite G. entailer.
  Qed.

  (*For now, we only export the constructor. If later "subclasses" inherit init, final, or step, well need to export these here, too*)
  Definition Weekday_VSU: @VSU NullExtension.Espec nil [(_surely_malloc, surely_malloc_numericspec M)] weekdaysQP 
             [(_make_Weekdays,mkWDspec)]
             WeekdayGV.
  Proof.
     mkVSU prog Weekday_internalspecs.
     - solve_SF_internal verif_init.
     - solve_SF_internal verif_final.
     - solve_SF_internal verif_step.
     - solve_SF_internal body_makeWeekdays.
     - apply Weekday_GP.
  Qed.

End Weekdays_VSU.